#include "backend-vdcs.h"
#include <ds/elapsed.h>
#include <ds/json.h>
#include <ds/locale/locale.h>
#include <ds/log.h>
#include <ds/memory.h>
#include <boost/algorithm/string.hpp>
#include <boost/serialization/strong_typedef.hpp>
#include "ds485types.h"
#include "event.h"
#include "handler/system_states.h"
#include "http_client.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance.h"
#include "model/state.h"
#include "propertysystem_common_paths.h"

static ds::LogChannel dsLogChannel("backendVdcs");

namespace dss {

std::ostream& operator<<(std::ostream& ostream, const BackendVdcId& x) {
  return ostream << static_cast<const std::string&>(x);
}

std::ostream& operator<<(std::ostream& ostream, const BackendVdcDeviceId& x) {
  return ostream << static_cast<const std::string&>(x);
}

static BackendVdcId HEKATRON{"hekatron"};

class BackendVdcsImpl DS_FINAL : public BackendVdcs,
                                 private PropertyListener,
                                 public EventInterpreterPlugin,
                                 public StructureModifyingBusInterface {
public:
  struct VdcDevice {
    VdsdSpec_t spec;
  };

  struct Vdc {
    VdcSpec spec;
    std::map<BackendVdcDeviceId, VdcDevice> devices;
    devid_t lastShortAddress;

    Vdc() : spec{}, devices{}, lastShortAddress{} {}
  };

  BackendVdcsImpl(
      Apartment& apartment, PropertySystem& propertySystem, EventInterpreter& eventInterpreter, ds::Conf& conf)
      : EventInterpreterPlugin("BackendVdcs", &eventInterpreter),
        m_apartment(apartment),
        m_propertySystem(propertySystem),
        m_conf(conf) {
    DS_NOTICE("BackendVdcs");
    m_apartment.getBusInterface()->add(*this);
    m_enabledNode = m_propertySystem.getProperty(pp_websvc_mshub_active);
    m_enabledNode->addListener(this);

    eventInterpreter.addPlugin(this);
    subscribe();

    update();
  }

  ~BackendVdcsImpl() { DS_NOTICE("~BackendVdcs"); }

  void propertyChanged(PropertyNodePtr, PropertyNodePtr node) DS_OVERRIDE {
    // TODO(soon): thread synchronization
    DS_DEBUG_ENTER(node->getName());
    if (node == m_enabledNode) {
      update();
    }
  }

  void subscribe() DS_OVERRIDE {
    getEventInterpreter().subscribe(boost::make_shared<EventSubscription>(*this, EventName::StateChange));
  }

  void handleEvent(Event& event, const EventSubscription&) DS_OVERRIDE {
    // TODO(soon): thread synchronization
    DS_DEBUG_ENTER(event);

    // notify hekatron vdc on active -> inactive || active -> activeMuted transition
    //
    // TODO(someday): Give cloud backendVdcs option to subscribe
    // to events and move this logic there.
    try {
      if (m_vdcs.find(HEKATRON) != m_vdcs.end() && event.getName() == EventName::StateChange &&
          event.getPropertyByName("statename") == StateName::Fire) {
        const auto& oldValue = event.getPropertyByName("oldvalue");
        if (oldValue == "1") {
          DS_INFO("Notify hekatron mute fire.", event);
          http(HttpRequest{url : ds::str("/v1/cloudvdc/vdcs/hekatron/action/muteFire"), type : POST});
        }
      }
    } catch (const std::exception& e) {
      DS_ERROR("Failed to notify hekatron fire mute", e);
    }
  }

  void notify(const std::string& path, ResourceNotification&& notification) DS_OVERRIDE {
    // TODO(soon): thread synchronization
    DS_INFO_ENTER(path);
    std::vector<std::string> pathParts;
    boost::split(pathParts, path, boost::is_any_of("/"));
    DS_REQUIRE(pathParts.size() > 0);
    pathParts.erase(pathParts.begin());

    if (pathParts.size() >= 2 && pathParts.at(0) == "vdcs") {
      const auto& vdcId = BackendVdcId{pathParts.at(1)};
      if (pathParts.size() == 2) {
        notifyVdc(vdcId, std::move(notification));
        return;
      } else if (pathParts.size() >= 4 && pathParts.at(2) == "devices") {
        const auto& vdcDeviceId = BackendVdcDeviceId{pathParts.at(3)};
        if (pathParts.size() == 4) {
          notifyVdcDevice(vdcId, vdcDeviceId, std::move(notification));
          return;
        }
      }
    }
    DS_FAIL_REQUIRE("Unsupported notify path.", path);
  }

  void notifyVdc(const BackendVdcId& vdcId, ResourceNotification&& notification) {
    DS_DEBUG_ENTER(vdcId);
    if (auto put = boost::get<PutResource>(&notification)) {
      putVdc(vdcId, ds::fromJson<VdcSpec>(put->value));
      return;
    } else if (boost::get<DeleteResource>(&notification)) {
      deleteVdc(vdcId);
      return;
    } else if (boost::get<PatchResource>(&notification)) {
      DS_FAIL_REQUIRE("PATCH vdc not supported (yet).");
    }
    DS_FAIL_REQUIRE("Unsupported notify.", vdcId);
  }

  void notifyVdcDevice(const BackendVdcId& vdcId, const BackendVdcDeviceId& id, ResourceNotification&& notification) {
    DS_DEBUG_ENTER(vdcId, id);
    if (auto put = boost::get<PutResource>(&notification)) {
      putVdcDevice(vdcId, id, ds::fromJson<VdsdSpec_t>(put->value));
      return;
    } else if (boost::get<DeleteResource>(&notification)) {
      deleteVdcDevice(vdcId, id);
      return;
    } else if (boost::get<PatchResource>(&notification)) {
      DS_FAIL_REQUIRE("PATCH vdc device not supported (yet).");
    }
    DS_FAIL_REQUIRE("Unsupported notify.", vdcId, id);
  }

  void putVdc(const BackendVdcId& id, VdcSpec&& specIn) {
    DS_DEBUG_ENTER(id, specIn);
    DS_NOTICE("Put vdc", id, specIn);
    specIn.implementationId = static_cast<const std::string&>(id);
    auto& vdc = m_vdcs[id];
    vdc.spec = std::move(specIn);
    const auto& spec = vdc.spec;
    auto dsm = m_apartment.allocateDSMeter(str2dsuid(spec.dSUID));
    dsm->setBusMemberType(BusMember_backendVdc);
    dsm->setCapability_HasDevices(true);
    dsm->setCapability_HasMetering(spec.capabilities.metering);
    dsm->setCapability_HasBlinking(spec.capabilities.identification);
    dsm->setCapability_HasDynamicDefinitions(spec.capabilities.dynamicDefinitions);
    dsm->setHardwareVersion(spec.hardwareVersion);
    dsm->setHardwareName(spec.model);
    dsm->setSoftwareVersion(spec.modelVersion);
    dsm->setVdcModelUID(spec.modelUID);
    dsm->setVdcConfigURL(spec.configURL);
    dsm->setVdcHardwareGuid(spec.hardwareGuid);
    dsm->setVdcHardwareModelGuid(spec.hardwareModelGuid);
    dsm->setVdcImplementationId(spec.implementationId);
    dsm->setVdcVendorGuid(spec.vendorGuid);
    dsm->setVdcOemGuid(spec.oemGuid);
    dsm->setVdcOemModelGuid(spec.oemModelGuid);
    dsm->setName(spec.name);
    dsm->setAuthorized(spec.authorized);
    dsm->setIsPresent(true);
    dsm->setIsConnected(true);
    dsm->setIsValid(true);
  }

  void deleteVdc(const BackendVdcId& id) {
    DS_DEBUG_ENTER(id);
    DS_NOTICE("Delete vdc", id);
    const auto& vdcIt = m_vdcs.find(id);
    if (vdcIt != m_vdcs.end()) {
      const auto& vdc = vdcIt->second;
      const auto& spec = vdc.spec;
      if (const auto& dsm = m_apartment.tryGetDSMeterByDSID(str2dsuid(spec.dSUID))) {
        dsm->setIsPresent(false);
      }
      m_vdcs.erase(vdcIt);
    } else {
      DS_FAIL_REQUIRE("Vdc not found.", id);
    }
  }

  Vdc& vdc(const BackendVdcId& vdcId) {
    const auto& it = m_vdcs.find(vdcId);
    DS_REQUIRE(it != m_vdcs.end(), "Unknown vdc.", vdcId);
    return it->second;
  }

  void putVdcDevice(const BackendVdcId& vdcId, const BackendVdcDeviceId& id, VdsdSpec_t&& specIn) {
    DS_DEBUG_ENTER(vdcId, id, specIn);
    auto& vdc = this->vdc(vdcId);
    auto& device = vdc.devices[id];
    m_deviceDsuidToVdc.erase(device.spec.dSUID);
    device.spec = specIn;
    m_deviceDsuidToVdc[device.spec.dSUID] = vdcId;
    const auto& spec = device.spec;
    DS_NOTICE("Put vdc device", vdcId, id, spec);
    const auto& dsm = m_apartment.getDSMeterByDSID(str2dsuid(vdc.spec.dSUID));
    auto modelDevice = m_apartment.allocateDevice(str2dsuid(spec.dSUID));
    bool vdcDataQueryNeeded = modelDevice->getVdsdSpec().oemModelGuid.empty() && !specIn.oemModelGuid.empty();
    modelDevice->setVdsdSpec(std::move(specIn));
    DeviceReference modelDeviceReference(modelDevice, &m_apartment);
    modelDevice->setVdcDevice(false); // vdc device means something else
    if (modelDevice->getShortAddress() == ShortAddressStaleDevice) {
      modelDevice->setShortAddress(vdc.lastShortAddress++);
    }
    modelDevice->setDSMeter(dsm);
    if (modelDevice->getName().empty()) { // dss is data owner
      modelDevice->setName(spec.name);
    }
    modelDevice->addToGroup(spec.primaryGroup);
    modelDevice->setIsPresent(spec.active);
    modelDevice->setIsConnected(spec.active);
    modelDevice->setButtonInputMode(ButtonInputMode::DEACTIVATED);

    // Always valid. Read is atomic in backendVdc
    modelDevice->setIsValid(true);
    dsm->addDevice(modelDeviceReference);

    // Binary inputs.
    //
    // For simplicity, all binary input arrays need to be in the same order
    // and ordered by dSIndex
    //
    // Binary input settings and descriptions are applied only once for simplicy.
    // Device->setBinaryInputs has side effects when called again
    auto binaryInputsSize = spec.binaryInputSettings.size();
    if (modelDevice->getBinaryInputCount() != binaryInputsSize) {
      std::vector<DeviceBinaryInputSpec_t> deviceBinaryInputs;
      for (uint i = 0; i < binaryInputsSize; i++) {
        deviceBinaryInputs.push_back(DeviceBinaryInputSpec_t{});
        auto& x = deviceBinaryInputs.back();
        x.TargetGroup = spec.binaryInputSettings[i].second.group;
        x.InputType = static_cast<BinaryInputType>(spec.binaryInputDescriptions[i].second.sensorFunction);
        x.InputID = static_cast<BinaryInputId>(15); // Avoids magic numbers in property tree.
      }
      modelDevice->setBinaryInputs(deviceBinaryInputs);
      DS_REQUIRE(binaryInputsSize == spec.binaryInputDescriptions.size());
      for (size_t i = 0; i < binaryInputsSize; i++) {
        const auto& description = spec.binaryInputDescriptions.at(i);
        DS_REQUIRE(
            spec.binaryInputSettings.at(i).first == description.first, "setting and description ids don't match 1:1");
        DS_REQUIRE(description.second.dSIndex == int(i), "descriptions must be in dSIndex order");
        modelDevice->setBinaryInputsInfo(description.second.dSIndex, description.first, description.second.name);
      }
    }
    DS_REQUIRE(binaryInputsSize == spec.binaryInputStates.size());
    for (size_t i = 0; i < binaryInputsSize; i++) {
      const auto& state = spec.binaryInputStates.at(i);
      DS_REQUIRE(spec.binaryInputSettings.at(i).first == state.first, "setting and states ids don't match 1:1");
      const auto& stateValue = static_cast<BinaryInputStateValue>(state.second.value);
      modelDevice->handleBinaryInputEvent(i, stateValue);
    }

    if (vdcDataQueryNeeded) {
      TaskProcessor& tp(m_apartment.getModelMaintenance()->getTaskProcessor());
      tp.addEvent(boost::make_shared<ModelMaintenance::VdcDataQuery>(modelDevice));
    }

    // Hekatron device presence enables mute fire support within dss.
    if (vdcId == HEKATRON) {
      const auto& origin = coJSON; // coSystem would be ignored by system_states code
      m_apartment.getNonScriptState(StateName::FireMuteEnabled)->setState(origin, State_Active);
    }
  }

  void deleteVdcDevice(const BackendVdcId& vdcId, const BackendVdcDeviceId& id) {
    DS_DEBUG_ENTER(vdcId, id);
    DS_NOTICE("Delete vdc device", vdcId, id);
    auto& vdc = this->vdc(vdcId);
    auto& devices = vdc.devices;
    const auto& deviceIt = devices.find(id);
    if (deviceIt != devices.end()) {
      const auto& device = deviceIt->second;
      m_deviceDsuidToVdc.erase(device.spec.dSUID);
      if (const auto& modelDevice = m_apartment.tryGetDeviceByDSID(str2dsuid(device.spec.dSUID))) {
        modelDevice->setIsPresent(false);
      }
      devices.erase(deviceIt);
    } else {
      DS_FAIL_REQUIRE("Vdc device not found.", vdcId, id);
    }
  }

  void update() {
    DS_DEBUG_ENTER("");
    updateVdcs();
    BOOST_FOREACH (const auto& vdcEntry, m_vdcs) { updateVdcDevices(vdcEntry.first); }
  }

  void updateVdcs() {
    DS_DEBUG_ENTER("");
    try {
      auto&& vdcs =
          ds::fromJson<std::vector<std::pair<BackendVdcId, VdcSpec>>>(http(HttpRequest{url : "/v1/cloudvdc/vdcs"}));
      for (auto begin = vdcs.begin(), end = vdcs.end(); begin != end; ++begin) {
        auto& vdc = *begin;
        const auto vdcId = vdc.first;
        try {
          putVdc(vdcId, std::move(vdc.second));
        } catch (const std::exception& e) {
          DS_ERROR("Failed to put vdc", vdcId, e);
        }
      }
    } catch (const std::exception& e) {
      DS_ERROR("Failed to update vdcs", e);
    }
  }

  void updateVdcDevices(const BackendVdcId& vdcId) {
    DS_DEBUG_ENTER(vdcId);
    try {
      auto&& devices = ds::fromJson<std::vector<VdsdSpec_t>>(
          http(HttpRequest{url : ds::str("/v1/cloudvdc/vdcs/", vdcId, "/devices")}));
      for (auto begin = devices.begin(), end = devices.end(); begin != end; ++begin) {
        auto& device = *begin;
        const auto id = device.dSUID;
        try {
          putVdcDevice(vdcId, BackendVdcDeviceId{device.dSUID}, std::move(device));
        } catch (const std::exception& e) {
          DS_ERROR("Failed to put vdc device", vdcId, id, e);
        }
      }
    } catch (const std::exception& e) {
      DS_ERROR("Failed to update vdc devices", vdcId, e);
    }
  }

  StructureModifyingBusInterface& structureModifyingBusInterface() DS_OVERRIDE { return *this; }

  BackendVdcId vdcIdForDeviceDsuid(const dsuid_t& dsuid) const DS_OVERRIDE {
    const auto& it = m_deviceDsuidToVdc.find(ds::str(dsuid));
    DS_REQUIRE(it != m_deviceDsuidToVdc.end(), "Unknown vdc device.", dsuid);
    return it->second;
  }

  std::string vdcDsuidForId(const BackendVdcId& vdcId) const DS_OVERRIDE {
    const auto& it = m_vdcs.find(vdcId);
    if (it != m_vdcs.end()) {
      return it->second.spec.dSUID;
    }
    DS_FAIL_REQUIRE("Unknown vdc", vdcId);
  }

  ds::JsonFragment http(HttpRequest&& x, const boost::optional<std::locale>& locale = boost::none) const DS_OVERRIDE {
    DS_DEBUG_ENTER(x);
    // PC build does not have remote connectivity available.
    // Ds-conf options `dssBackendVdcsDssId` and `dssBackendVdcsOspToken`
    // allow to mock values from remote connectivity add-on
    const auto& dssId =
        m_conf.tryRead<std::string>("dssBackendVdcsDssId").value_or(m_propertySystem.getStringValue(pp_sysinfo_dsid));
    const auto& ospToken = m_conf.tryRead<std::string>("dssBackendVdcsOspToken")
                               .value_or(m_propertySystem.getStringValue(pp_websvc_rc_osptoken));
    x.url = ds::str(baseUrl(), x.url);
    x.headers.insert({"x-ds-dss-id", dssId});
    x.headers.insert({"x-ds-osp-token", ospToken});
    if (locale) {
      x.headers.insert({"Accept-Language", ds::locale::language(*locale)});
    }

    HttpClient httpClient;
    std::string result;

    ds::Elapsed elapsed;
    auto code = httpClient.request(x, &result);
    DS_NOTICE("Http request", x.url, code, result, elapsed);
    DS_REQUIRE(code / 100 == 2, "Http request failed.", x.url, code, result);
    return ds::fromJson<ds::JsonFragment>(result);
  }

private:
  Apartment& m_apartment;
  PropertySystem& m_propertySystem;
  ds::Conf& m_conf;
  PropertyNodePtr m_enabledNode;

  std::map<BackendVdcId, Vdc> m_vdcs;
  std::map<std::string, BackendVdcId> m_deviceDsuidToVdc;

  std::string baseUrl() const { return m_propertySystem.getStringValue(pp_websvc_private_url); }

  std::string name() const DS_OVERRIDE { return dsLogChannel.name(); }

  //
  // StructureModifyingBusInterface
  //
  void setZoneID(const dsuid_t&, const devid_t, const int) DS_OVERRIDE {
    // zones are not propagated
  }
  void createZone(const dsuid_t&, const int) DS_OVERRIDE {
    // zones are not propagated
  }
  void removeZone(const dsuid_t&, const int) DS_OVERRIDE {
    // zones are not propagated
  }
  void addToGroup(const dsuid_t&, const int, const int) DS_OVERRIDE { DS_FAIL_REQUIRE("Failed."); }
  void removeFromGroup(const dsuid_t&, const int, const int) DS_OVERRIDE { DS_FAIL_REQUIRE("Failed."); }
  void removeDeviceFromDSMeter(const dsuid_t&, const int) DS_OVERRIDE { DS_FAIL_REQUIRE("Failed."); }
  void removeDeviceFromDSMeters(const dsuid_t& dsuid) DS_OVERRIDE {
    http(HttpRequest{
      url : ds::str("/v1/cloudvdc/vdcs/", vdcIdForDeviceDsuid(dsuid), "/devices/", dsuid),
      type : DELETE,
    });
  }
  void sceneSetName(uint16_t, uint8_t, uint8_t, const std::string&) DS_OVERRIDE {}
  void deviceSetName(dsuid_t, devid_t, const std::string&) DS_OVERRIDE {} // dss is data owner
  void meterSetName(dsuid_t, const std::string&) DS_OVERRIDE { DS_FAIL_REQUIRE("Failed."); }
  void createGroup(uint16_t, uint8_t, ApplicationType, uint32_t, const std::string&) DS_OVERRIDE {}
  void removeGroup(uint16_t, uint8_t) DS_OVERRIDE {}
  void groupSetName(uint16_t, uint8_t, const std::string&) DS_OVERRIDE {}
  void createCluster(uint8_t, ApplicationType, uint32_t, const std::string&) DS_OVERRIDE {}
  void removeCluster(uint8_t) DS_OVERRIDE {}
  void groupSetApplication(uint16_t, uint8_t, ApplicationType, uint32_t) DS_OVERRIDE {}
  void clusterSetName(uint8_t, const std::string&) DS_OVERRIDE {}
  void clusterSetApplication(uint8_t, ApplicationType, uint32_t) DS_OVERRIDE {}
  void clusterSetProperties(uint8_t, uint16_t _location, uint16_t _floor, uint16_t _protectionClass) DS_OVERRIDE {}
  void clusterSetLockedScenes(uint8_t, const std::vector<int>&) DS_OVERRIDE {}
  void clusterSetConfigurationLock(uint8_t, bool) DS_OVERRIDE {}
  void clusterSetOperationLock(uint8_t, bool) DS_OVERRIDE {}
  void setButtonSetsLocalPriority(const dsuid_t&, const devid_t, bool) DS_OVERRIDE { DS_FAIL_REQUIRE("Failed."); }
  void setButtonCallsPresent(const dsuid_t&, const devid_t, bool) DS_OVERRIDE { DS_FAIL_REQUIRE("Failed."); }
  void setZoneHeatingConfig(const dsuid_t&, const uint16_t, const ZoneHeatingConfigSpec_t) DS_OVERRIDE {
    DS_FAIL_REQUIRE("Failed.");
  }
  void setZoneHeatingOperationModes(const dsuid_t&, const uint16_t, const ZoneHeatingOperationModeSpec_t) DS_OVERRIDE {
    DS_FAIL_REQUIRE("Failed.");
  }
  void setZoneSensor(const uint16_t, SensorType, const dsuid_t&) DS_OVERRIDE {}
  void resetZoneSensor(const uint16_t, SensorType) DS_OVERRIDE {}
  void setCircuitPowerStateConfig(const dsuid_t&, const int, const int, const int) DS_OVERRIDE {
    DS_FAIL_REQUIRE("Failed.");
  }

  void setProperty(const dsuid_t&, const ::google::protobuf::RepeatedPtrField<::vdcapi::PropertyElement>&) DS_OVERRIDE {
    DS_FAIL_REQUIRE("Failed.");
  }
  vdcapi::Message getProperty(
      const dsuid_t&, const ::google::protobuf::RepeatedPtrField<::vdcapi::PropertyElement>&) DS_OVERRIDE {
    DS_FAIL_REQUIRE("Failed.");
  }

  void setFireMuteEnabled(dsuid_t dsm, bool) DS_OVERRIDE {}
  void removeVdc(const dsuid_t& vdsmID, const dsuid_t& vdcID) DS_OVERRIDE { }
};

BackendVdcs::~BackendVdcs(){};

std::unique_ptr<BackendVdcs> makeBackendVdcs(
    Apartment& apartment, PropertySystem& propertySystem, EventInterpreter& eventInterpreter, ds::Conf& conf) {
  return ds::make_unique<BackendVdcsImpl>(apartment, propertySystem, eventInterpreter, conf);
}

} // namespace dss
