// This program is a code generator plugin for `capnp compile` which generates C++ code.

#include <capnp/schema.capnp.h>
#include <capnp/serialize.h>
#include <kj/debug.h>
#include <kj/io.h>
#include <kj/string-tree.h>
#include <kj/tuple.h>
#include <kj/vector.h>
#include <capnp/schema-loader.h>
#include <capnp/dynamic.h>
#include <kj/miniposix.h>
#include <unordered_map>
#include <unordered_set>
#include <map>
#include <set>
#include <kj/main.h>
#include <algorithm>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#ifndef VERSION
#define VERSION "(unknown)"
#endif

namespace capnp {
namespace {

static constexpr uint64_t NAMESPACE_ANNOTATION_ID = 0xb9c6f99ebf805f2cull;
static constexpr uint64_t NAME_ANNOTATION_ID = 0xf264a779fef191ceull;

struct OrderByName {
  template <typename T>
  inline bool operator()(const T& a, const T& b) const {
    return a.getProto().getName() < b.getProto().getName();
  }
};

template <typename MemberList>
kj::Array<uint> makeMembersByName(MemberList&& members) {
  auto sorted = KJ_MAP(member, members) { return member; };
  std::sort(sorted.begin(), sorted.end(), OrderByName());
  return KJ_MAP(member, sorted) { return member.getIndex(); };
}

kj::StringPtr baseName(kj::StringPtr path) {
  KJ_IF_MAYBE(slashPos, path.findLast('/')) {
    return path.slice(*slashPos + 1);
  } else {
    return path;
  }
}

kj::String safeIdentifier(kj::StringPtr identifier) {
  // Given a desired identifier name, munge it to make it safe for use in generated code.
  //
  // If the identifier is a keyword, this adds an underscore to the end.

  static const std::set<kj::StringPtr> keywords({
    "alignas", "alignof", "and", "and_eq", "asm", "auto", "bitand", "bitor", "bool", "break",
    "case", "catch", "char", "char16_t", "char32_t", "class", "compl", "const", "constexpr",
    "const_cast", "continue", "decltype", "default", "delete", "do", "double", "dynamic_cast",
    "else", "enum", "explicit", "export", "extern", "false", "float", "for", "friend", "goto",
    "if", "inline", "int", "long", "mutable", "namespace", "new", "noexcept", "not", "not_eq",
    "nullptr", "operator", "or", "or_eq", "private", "protected", "public", "register",
    "reinterpret_cast", "return", "short", "signed", "sizeof", "static", "static_assert",
    "static_cast", "struct", "switch", "template", "this", "thread_local", "throw", "true",
    "try", "typedef", "typeid", "typename", "union", "unsigned", "using", "virtual", "void",
    "volatile", "wchar_t", "while", "xor", "xor_eq"
  });

  if (keywords.count(identifier) > 0) {
    return kj::str(identifier, '_');
  } else {
    return kj::heapString(identifier);
  }
}

// =======================================================================================

class CppTypeName {
  // Used to build a C++ type name string. This is complicated in the presence of templates,
  // because we must add the "typename" and "template" disambiguator keywords as needed.

public:
  inline CppTypeName(): isArgDependent(false), needsTypename(false),
                        hasInterfaces_(false), hasDisambiguatedTemplate_(false) {}
  CppTypeName(CppTypeName&& other) = default;
  CppTypeName(const CppTypeName& other)
      : name(kj::strTree(other.name.flatten())),
        isArgDependent(other.isArgDependent),
        needsTypename(other.needsTypename),
        hasInterfaces_(other.hasInterfaces_),
        hasDisambiguatedTemplate_(other.hasDisambiguatedTemplate_) {}

  CppTypeName& operator=(CppTypeName&& other) = default;
  CppTypeName& operator=(const CppTypeName& other) {
    name = kj::strTree(other.name.flatten());
    isArgDependent = other.isArgDependent;
    needsTypename = other.needsTypename;
    return *this;
  }

  static CppTypeName makeRoot() {
    return CppTypeName(kj::strTree(" "), false);
  }

  static CppTypeName makeNamespace(kj::StringPtr name) {
    return CppTypeName(kj::strTree(" ::", name), false);
  }

  static CppTypeName makeTemplateParam(kj::StringPtr name) {
    return CppTypeName(kj::strTree(name), true);
  }

  static CppTypeName makePrimitive(kj::StringPtr name) {
    return CppTypeName(kj::strTree(name), false);
  }

  void addMemberType(kj::StringPtr innerName) {
    // Append "::innerName" to refer to a member.

    name = kj::strTree(kj::mv(name), "::", innerName);
    needsTypename = isArgDependent;
  }

  void addMemberValue(kj::StringPtr innerName) {
    // Append "::innerName" to refer to a member.

    name = kj::strTree(kj::mv(name), "::", innerName);
    needsTypename = false;
  }

  bool hasDisambiguatedTemplate() {
    return hasDisambiguatedTemplate_;
  }

  void addMemberTemplate(kj::StringPtr innerName, kj::Array<CppTypeName>&& params) {
    // Append "::innerName<params, ...>".
    //
    // If necessary, add the "template" disambiguation keyword in front of `innerName`.

    bool parentIsArgDependent = isArgDependent;
    needsTypename = parentIsArgDependent;
    hasDisambiguatedTemplate_ = hasDisambiguatedTemplate_ || parentIsArgDependent;

    name = kj::strTree(kj::mv(name),
        parentIsArgDependent ? "::template " : "::",
        innerName, '<',
        kj::StringTree(KJ_MAP(p, params) {
          if (p.isArgDependent) isArgDependent = true;
          if (p.hasInterfaces_) hasInterfaces_ = true;
          if (p.hasDisambiguatedTemplate_) hasDisambiguatedTemplate_ = true;
          return kj::strTree(kj::mv(p));
        }, ", "),
        '>');
  }

  void setHasInterfaces() {
    hasInterfaces_ = true;
  }

  bool hasInterfaces() {
    return hasInterfaces_;
  }

  kj::StringTree strNoTypename() const & { return name.flatten(); }
  kj::StringTree strNoTypename() && { return kj::mv(name); }
  // Stringify but never prefix with `typename`. Use in contexts where `typename` is implicit.

private:
  kj::StringTree name;

  bool isArgDependent;
  // Does the name contain any template-argument-dependent types?

  bool needsTypename;
  // Does the name require a prefix of "typename"?

  bool hasInterfaces_;
  // Does this type name refer to any interface types? If so it may need to be #ifdefed out in
  // lite mode.

  bool hasDisambiguatedTemplate_;
  // Whether the type name contains a template type that had to be disambiguated using the
  // "template" keyword, e.g. "Foo<T>::template Bar<U>".
  //
  // TODO(msvc): We only track this because MSVC seems to get confused by it in some weird cases.

  inline CppTypeName(kj::StringTree&& name, bool isArgDependent)
      : name(kj::mv(name)), isArgDependent(isArgDependent), needsTypename(false),
        hasInterfaces_(false), hasDisambiguatedTemplate_(false) {}

  friend kj::StringTree KJ_STRINGIFY(CppTypeName&& typeName);
  friend kj::String KJ_STRINGIFY(const CppTypeName& typeName);
};

kj::StringTree KJ_STRINGIFY(CppTypeName&& typeName) {
  if (typeName.needsTypename) {
    return kj::strTree("typename ", kj::mv(typeName.name));
  } else {
    return kj::mv(typeName.name);
  }
}
kj::String KJ_STRINGIFY(const CppTypeName& typeName) {
  if (typeName.needsTypename) {
    return kj::str("typename ", typeName.name);
  } else {
    return typeName.name.flatten();
  }
}

// =======================================================================================

class CapnpcCppMain {
public:
  CapnpcCppMain(kj::ProcessContext& context): context(context) {}

  kj::MainFunc getMain() {
    return kj::MainBuilder(context, "Cap'n Proto C++ plugin version " VERSION,
          "This is a Cap'n Proto compiler plugin which generates C++ code. "
          "It is meant to be run using the Cap'n Proto compiler, e.g.:\n"
          "    capnp compile -oc++ foo.capnp")
        .callAfterParsing(KJ_BIND_METHOD(*this, run))
        .build();
  }

private:
  kj::ProcessContext& context;
  SchemaLoader schemaLoader;
  std::unordered_set<uint64_t> usedImports;
  bool hasInterfaces = false;

  CppTypeName cppFullName(Schema schema, kj::Maybe<InterfaceSchema::Method> method) {
    return cppFullName(schema, schema, method);
  }

  CppTypeName cppFullName(Schema schema, Schema brand, kj::Maybe<InterfaceSchema::Method> method) {
    auto node = schema.getProto();

    if (node.getScopeId() == 0) {
      // This is the file-level scope. Search for the namespace annotation.
      KJ_REQUIRE(node.isFile(),
          "Non-file had scopeId zero; perhaps it's a method param / result struct?");
      usedImports.insert(node.getId());
      KJ_IF_MAYBE(ns, annotationValue(node, NAMESPACE_ANNOTATION_ID)) {
        return CppTypeName::makeNamespace(ns->getText());
      } else {
        return CppTypeName::makeRoot();
      }
    } else {
      // This is a named type.

      // Figure out what name to use.
      Schema parent = schemaLoader.get(node.getScopeId());
      kj::StringPtr unqualifiedName;
      kj::String ownUnqualifiedName;
      KJ_IF_MAYBE(annotatedName, annotationValue(node, NAME_ANNOTATION_ID)) {
        // The node's name has been overridden for C++ by an annotation.
        unqualifiedName = annotatedName->getText();
      } else {
        // Search among the parent's nested nodes to for this node, in order to determine its name.
        auto parentProto = parent.getProto();
        for (auto nested: parentProto.getNestedNodes()) {
          if (nested.getId() == node.getId()) {
            unqualifiedName = nested.getName();
            break;
          }
        }
        if (unqualifiedName == nullptr) {
          // Hmm, maybe it's a group node?
          if (parentProto.isStruct()) {
            for (auto field: parentProto.getStruct().getFields()) {
              if (field.isGroup() && field.getGroup().getTypeId() == node.getId()) {
                ownUnqualifiedName = toTitleCase(protoName(field));
                unqualifiedName = ownUnqualifiedName;
                break;
              }
            }
          }
        }
        KJ_REQUIRE(unqualifiedName != nullptr,
            "A schema Node's supposed scope did not contain the node as a NestedNode.");
      }

      auto result = cppFullName(parent, brand, method);

      // Construct the generic arguments.
      auto params = node.getParameters();
      if (params.size() > 0) {
        auto args = brand.getBrandArgumentsAtScope(node.getId());

#if 0
        // Figure out exactly how many params are not bound to AnyPointer.
        // TODO(msvc): In a few obscure cases, MSVC does not like empty template pramater lists,
        //   even if all parameters have defaults. So, we give in and explicitly list all
        //   parameters in our generated code for now. Try again later.
        uint paramCount = 0;
        for (uint i: kj::indices(params)) {
          auto arg = args[i];
          if (arg.which() != schema::Type::ANY_POINTER || arg.getBrandParameter() != nullptr ||
              (arg.getImplicitParameter() != nullptr && method != nullptr)) {
            paramCount = i + 1;
          }
        }
#else
        uint paramCount = params.size();
#endif

        result.addMemberTemplate(unqualifiedName,
            KJ_MAP(i, kj::range(0u, paramCount)) {
              return typeName(args[i], method);
            });
      } else {
        result.addMemberType(unqualifiedName);
      }

      return result;
    }
  }

  kj::String toUpperCase(kj::StringPtr name) {
    kj::Vector<char> result(name.size() + 4);

    for (char c: name) {
      if ('a' <= c && c <= 'z') {
        result.add(c - 'a' + 'A');
      } else if (result.size() > 0 && 'A' <= c && c <= 'Z') {
        result.add('_');
        result.add(c);
      } else {
        result.add(c);
      }
    }

    if (result.size() == 4 && memcmp(result.begin(), "NULL", 4) == 0) {
      // NULL probably collides with a macro.
      result.add('_');
    }

    result.add('\0');

    return kj::String(result.releaseAsArray());
  }

  kj::String toTitleCase(kj::StringPtr name) {
    kj::String result = kj::heapString(name);
    if ('a' <= result[0] && result[0] <= 'z') {
      result[0] = result[0] - 'a' + 'A';
    }
    return kj::mv(result);
  }

  CppTypeName typeName(Type type, kj::Maybe<InterfaceSchema::Method> method) {
    switch (type.which()) {
      case schema::Type::VOID: return CppTypeName::makePrimitive(" ::capnp::Void");

      case schema::Type::BOOL: return CppTypeName::makePrimitive("bool");
      case schema::Type::INT8: return CppTypeName::makePrimitive(" ::int8_t");
      case schema::Type::INT16: return CppTypeName::makePrimitive(" ::int16_t");
      case schema::Type::INT32: return CppTypeName::makePrimitive(" ::int32_t");
      case schema::Type::INT64: return CppTypeName::makePrimitive(" ::int64_t");
      case schema::Type::UINT8: return CppTypeName::makePrimitive(" ::uint8_t");
      case schema::Type::UINT16: return CppTypeName::makePrimitive(" ::uint16_t");
      case schema::Type::UINT32: return CppTypeName::makePrimitive(" ::uint32_t");
      case schema::Type::UINT64: return CppTypeName::makePrimitive(" ::uint64_t");
      case schema::Type::FLOAT32: return CppTypeName::makePrimitive("float");
      case schema::Type::FLOAT64: return CppTypeName::makePrimitive("double");

      case schema::Type::TEXT: return CppTypeName::makePrimitive(" ::std::string");
      case schema::Type::DATA: return CppTypeName::makePrimitive(" ::std::vector< ::uint8_t>");

      case schema::Type::ENUM:
        return cppFullName(type.asEnum(), method);
      case schema::Type::STRUCT:
        return cppFullName(type.asStruct(), method);
      case schema::Type::INTERFACE: {
        auto result = cppFullName(type.asInterface(), method);
        result.setHasInterfaces();
        return result;
      }

      case schema::Type::LIST: {
        CppTypeName result = CppTypeName::makeNamespace("std");
        auto params = kj::heapArrayBuilder<CppTypeName>(1);
        params.add(typeName(type.asList().getElementType(), method));
        result.addMemberTemplate("vector", params.finish());
        return result;
      }

      case schema::Type::ANY_POINTER:
        KJ_IF_MAYBE(param, type.getBrandParameter()) {
          return CppTypeName::makeTemplateParam(schemaLoader.get(param->scopeId).getProto()
              .getParameters()[param->index].getName());
        } else KJ_IF_MAYBE(param, type.getImplicitParameter()) {
          KJ_IF_MAYBE(m, method) {
            auto params = m->getProto().getImplicitParameters();
            KJ_REQUIRE(param->index < params.size());
            return CppTypeName::makeTemplateParam(params[param->index].getName());
          } else {
            return CppTypeName::makePrimitive(" ::capnp::AnyPointer");
          }
        } else {
          switch (type.whichAnyPointerKind()) {
            case schema::Type::AnyPointer::Unconstrained::ANY_KIND:
              return CppTypeName::makePrimitive(" ::capnp::AnyPointer");
            case schema::Type::AnyPointer::Unconstrained::STRUCT:
              return CppTypeName::makePrimitive(" ::capnp::AnyStruct");
            case schema::Type::AnyPointer::Unconstrained::LIST:
              return CppTypeName::makePrimitive(" ::capnp::AnyList");
            case schema::Type::AnyPointer::Unconstrained::CAPABILITY:
              hasInterfaces = true;  // Probably need to #include <capnp/capability.h>.
              return CppTypeName::makePrimitive(" ::capnp::Capability");
          }
          KJ_UNREACHABLE;
        }
    }
    KJ_UNREACHABLE;
  }

  template <typename P>
  kj::Maybe<schema::Value::Reader> annotationValue(P proto, uint64_t annotationId) {
    for (auto annotation: proto.getAnnotations()) {
      if (annotation.getId() == annotationId) {
        return annotation.getValue();
      }
    }
    return nullptr;
  }

  template <typename P>
  kj::StringPtr protoName(P proto) {
    KJ_IF_MAYBE(name, annotationValue(proto, NAME_ANNOTATION_ID)) {
      return name->getText();
    } else {
      return proto.getName();
    }
  }

  kj::StringTree literalValue(Type type, schema::Value::Reader value) {
    switch (value.which()) {
      case schema::Value::VOID: return kj::strTree(" ::capnp::VOID");
      case schema::Value::BOOL: return kj::strTree(value.getBool() ? "true" : "false");
      case schema::Value::INT8: return kj::strTree(value.getInt8());
      case schema::Value::INT16: return kj::strTree(value.getInt16());
      case schema::Value::INT32: return kj::strTree(value.getInt32());
      case schema::Value::INT64: return kj::strTree(value.getInt64(), "ll");
      case schema::Value::UINT8: return kj::strTree(value.getUint8(), "u");
      case schema::Value::UINT16: return kj::strTree(value.getUint16(), "u");
      case schema::Value::UINT32: return kj::strTree(value.getUint32(), "u");
      case schema::Value::UINT64: return kj::strTree(value.getUint64(), "llu");
      case schema::Value::FLOAT32: {
        auto text = kj::str(value.getFloat32());
        if (text.findFirst('.') == nullptr &&
            text.findFirst('e') == nullptr &&
            text.findFirst('E') == nullptr) {
          text = kj::str(text, ".0");
        }
        return kj::strTree(kj::mv(text), "f");
      }
      case schema::Value::FLOAT64: return kj::strTree(value.getFloat64());
      case schema::Value::ENUM: {
        EnumSchema schema = type.asEnum();
        if (value.getEnum() < schema.getEnumerants().size()) {
          return kj::strTree(
              cppFullName(schema, nullptr), "::",
              toUpperCase(protoName(schema.getEnumerants()[value.getEnum()].getProto())));
        } else {
          return kj::strTree("static_cast<", cppFullName(schema, nullptr),
                             ">(", value.getEnum(), ")");
        }
      }

      case schema::Value::TEXT:
      case schema::Value::DATA:
      case schema::Value::STRUCT:
      case schema::Value::INTERFACE:
      case schema::Value::LIST:
      case schema::Value::ANY_POINTER:
        KJ_FAIL_REQUIRE("literalValue() can only be used on primitive types.");
    }
    KJ_UNREACHABLE;
  }

  struct NodeText {
    kj::StringTree decl;
    kj::StringTree def;
    kj::StringTree inlineDef;
    kj::StringTree sourceDef;
    kj::StringTree structTypedef;
  };

  NodeText makeNodeText(kj::StringPtr namespace_, kj::StringPtr scope, kj::StringPtr scope_,
                        kj::StringPtr name, Schema schema) {

    auto proto = schema.getProto();
    KJ_IF_MAYBE(annotatedName, annotationValue(proto, NAME_ANNOTATION_ID)) {
      name = annotatedName->getText();
    }
    auto subScope = kj::str(scope, name, "::");
    auto subScope_ = kj::str(scope_, name, "_");

    // Compute nested nodes, including groups.
    kj::Vector<NodeText> nestedTexts(proto.getNestedNodes().size());
    for (auto nested: proto.getNestedNodes()) {
      nestedTexts.add(makeNodeText(
          namespace_, subScope, subScope_, nested.getName(), schemaLoader.getUnbound(nested.getId())));
    };

    if (proto.isStruct()) {
      for (auto field: proto.getStruct().getFields()) {
        if (field.isGroup()) {
          nestedTexts.add(makeNodeText(
              namespace_, subScope, subScope_, toTitleCase(protoName(field)),
              schemaLoader.getUnbound(field.getGroup().getTypeId())));
        }
      }
    }

    auto top = makeNodeTextWithoutNested(
        namespace_, scope, scope_, name, schema,
        KJ_MAP(x, nestedTexts) { return kj::mv(x.structTypedef); }
    );

    return NodeText {
      kj::strTree(
          KJ_MAP(n, nestedTexts) { return kj::mv(n.decl); },
          kj::mv(top.decl)),

      kj::strTree(
          KJ_MAP(n, nestedTexts) { return kj::mv(n.def); },
          kj::mv(top.def)),

      kj::strTree(
          KJ_MAP(n, nestedTexts) { return kj::mv(n.inlineDef); },
          kj::mv(top.inlineDef)),

      kj::strTree(
          KJ_MAP(n, nestedTexts) { return kj::mv(n.sourceDef); },
          kj::mv(top.sourceDef)),

      kj::mv(top.structTypedef)
  };
  }

  NodeText makeNodeTextWithoutNested(kj::StringPtr namespace_, kj::StringPtr scope, kj::StringPtr scope_,
                                     kj::StringPtr name, Schema schema, kj::Array<kj::StringTree> structTypedefs) {
    auto proto = schema.getProto();
    KJ_IF_MAYBE(annotatedName, annotationValue(proto, NAME_ANNOTATION_ID)) {
      name = annotatedName->getText();
    }
    switch (proto.which()) {
      case schema::Node::FILE:
        KJ_FAIL_REQUIRE("This method shouldn't be called on file nodes.");

      case schema::Node::STRUCT: {
        auto x = makeStructText(scope, scope_, name, schema.asStruct(), kj::mv(structTypedefs));
        return NodeText {kj::mv(x.decl), kj::mv(x.def), kj::mv(x.inlineDef), kj::mv(x.sourceDef),
          kj::strTree("  typedef ", scope_, name, " ", name, ";\n")
        };
      }

      case schema::Node::ENUM: {
        auto x = makeEnumText(scope, scope_, name, schema.asEnum());
        return NodeText {kj::mv(x.decl), kj::mv(x.def), kj::mv(x.inlineDef), kj::mv(x.sourceDef),
          kj::strTree("  typedef ", scope_, name, " ", name, ";\n")
        };
      }

      case schema::Node::INTERFACE: {
        return NodeText {};
      }
      case schema::Node::CONST: {
        return NodeText {};
      }
      case schema::Node::ANNOTATION: {
        return NodeText {};
      }
    }
    KJ_UNREACHABLE;
  }
  struct StructText {
    kj::StringTree decl;
    kj::StringTree def;
    kj::StringTree inlineDef;
    kj::StringTree sourceDef;
  };


  StructText makeStructText(kj::StringPtr scope, kj::StringPtr scope_, kj::StringPtr name,
                          StructSchema schema, kj::Array<kj::StringTree> structTypedefs) {
    auto proto = schema.getProto();
    KJ_IF_MAYBE(annotatedName, annotationValue(proto, NAME_ANNOTATION_ID)) {
      name = annotatedName->getText();
    }
    auto fullName_ = kj::str(scope_, name);
    auto subScope = kj::str(scope, name, "::");
    auto subScope_ = kj::str(fullName_, "_");
    auto fieldTexts = KJ_MAP(f, schema.getFields()) {
      return makeFieldText(subScope, subScope_, f);
    };

    int initializerIndex = 0;
    return StructText {
      //decl
      kj::strTree("struct ", fullName_, ";\n"),
      //def
      kj::strTree(
          "struct ", fullName_, " {\n",
           kj::mv(structTypedefs),
          "\n",
          KJ_MAP(x, fieldTexts) { return kj::mv(x.def); },
          "\n",
          "  ", fullName_, "();\n",
          "\n",
          "  void addFieldsToJson(::dss::JSONWriter&) const;\n",
          "  void setJson(::json_object*);\n",
          "  static ", fullName_, " fromJson(const char*);\n",
          "};\n",
          "std::ostream& operator<<(std::ostream&, const ", fullName_, "&);\n",
          "\n"
      ),
      //inlineDef
      kj::strTree(
          "inline ", fullName_, "::", fullName_, "() \n",
          KJ_MAP(x, fieldTexts) {
            return x.init.size()
                ? kj::strTree((initializerIndex++ == 0 ? "  : " : "  , "), kj::mv(x.init), "\n")
                : kj::strTree();
          },
          "{}\n",
          "\n"
      ),
      //def
      kj::strTree(
          "void ", fullName_, "::addFieldsToJson(::dss::JSONWriter& json) const {\n",
          KJ_MAP(x, fieldTexts) {
            return kj::mv(x.addToJson);
          },
          "}\n",
          "\n",
          "void ", fullName_, "::setJson(::json_object* object) {\n",
          "  DS_REQUIRE(json_object_is_type(object, json_type_object), \"Expecting json object.\");\n",
          "  json_object_object_foreach(object, dsCapnpKey, dsCapnpValue) {\n",
          KJ_MAP(x, fieldTexts) {
            return kj::mv(x.setJson);
          },
          "  }\n",
          "}\n",
          "\n",
          fullName_, " ", fullName_, "::fromJson(const char* text) {\n",
          "  struct json_tokener* tok = json_tokener_new();\n",
          "  DS_DEFER(json_tokener_free(tok));\n",
          "\n",
          "  json_object* object = json_tokener_parse_ex(tok, text, -1);\n",
          "  DS_REQUIRE(object, json_tokener_error_desc(json_tokener_get_error(tok)));\n",
          "  DS_DEFER(json_object_put(object));\n",
          "\n",
          "  ", fullName_, " out;\n",
          "  out.setJson(object);\n",
          "  return out;\n",
          "}\n",
          "\n",
          "std::ostream& operator<<(std::ostream& ostream, const ", fullName_, "& x) {\n",
          "  JSONWriter json(JSONWriter::jsonNoneResult);\n",
          "  x.addFieldsToJson(json);\n"
          "  return ostream << json.successJSON();\n"
          "}\n",
          "\n"
      )
    };
  }

  struct FieldText {
    kj::StringTree def;
    kj::StringTree init;
    kj::StringTree addToJson;
    kj::StringTree setJson;
  };

  FieldText makeFieldText(kj::StringPtr scope, kj::StringPtr scope_, StructSchema::Field field) {
    auto proto = field.getProto();
    auto baseName = protoName(proto);
    auto name = safeIdentifier(baseName);
    auto typeSchema = field.getType();
    CppTypeName type = typeName(typeSchema, nullptr);
    auto slot = proto.getSlot();
    auto defaultBody = slot.getDefaultValue();

    kj::StringPtr enumDefaultName;
    if (typeSchema.which() == schema::Type::ENUM) {
      enumDefaultName = protoName(typeSchema.asEnum().getEnumerants()[defaultBody.getEnum()].getProto());
    }

    kj::String defaultText;
    [&] {
      kj::StringTree s;
      switch (typeSchema.which()) {
        case schema::Type::VOID: return;
        case schema::Type::BOOL: if (auto x = defaultBody.getBool()) { s = kj::strTree(x); } break;
        case schema::Type::INT8: if (auto x = defaultBody.getInt8()) { s = kj::strTree(x); } break;
        case schema::Type::INT16: if (auto x = defaultBody.getInt16()) { s = kj::strTree(x); } break;
        case schema::Type::INT32: if (auto x = defaultBody.getInt32()) { s = kj::strTree(x); } break;
        case schema::Type::INT64: if (auto x = defaultBody.getInt64()) { s = kj::strTree(x, "ll"); } break;
        case schema::Type::UINT8: if (auto x = defaultBody.getUint8()) { s = kj::strTree(x, "u"); } break;
        case schema::Type::UINT16: if (auto x = defaultBody.getUint16()) { s = kj::strTree(x, "u"); } break;
        case schema::Type::UINT32: if (auto x = defaultBody.getUint32()) { s = kj::strTree(x, "u"); } break;
        case schema::Type::UINT64: if (auto x = defaultBody.getUint64()) { s = kj::strTree(x, "ull"); } break;
        case schema::Type::FLOAT32: if (auto x = defaultBody.getFloat32()) { s = kj::strTree(x); } break;
        case schema::Type::FLOAT64: if (auto x = defaultBody.getFloat64()) { s = kj::strTree(x); } break;
        case schema::Type::TEXT: {
          auto x = defaultBody.getText();
          if (x.size()) {
            s = kj::strTree('"', x, '"');
          }
          return;
        }
        case schema::Type::DATA: KJ_REQUIRE(defaultBody.getData().size() == 0, ""); return;
        case schema::Type::ENUM: if (defaultBody.getEnum()) { s = kj::strTree(type, "::", toUpperCase(enumDefaultName)); } break;
        case schema::Type::STRUCT: return;
        case schema::Type::LIST: return;
        case schema::Type::INTERFACE: return;
        case schema::Type::ANY_POINTER: return;
      };
      defaultText = kj::str("(", kj::mv(s), ")");
    }();

    return FieldText {
      kj::strTree("  ", type, " ", name, ";\n"),
      (defaultText.size() == 0 ? kj::strTree() :  kj::strTree(name, kj::mv(defaultText))),
      //addToJson
      kj::strTree(
        "  json.add(\"", name, "\");\n",
        [&] {
          auto add = [&] { return kj::strTree("  json.add(", name, ");\n"); };
          switch (typeSchema.which()) {
            case schema::Type::VOID: return kj::strTree("  json.addNull();\n");
            case schema::Type::BOOL: return add();
            case schema::Type::INT8: return add();
            case schema::Type::INT16: return add();
            case schema::Type::INT32: return add();
            case schema::Type::INT64: return add();
            case schema::Type::UINT8: return add();
            case schema::Type::UINT16: return add();
            case schema::Type::UINT32: return add();
            case schema::Type::UINT64: return add();
            case schema::Type::FLOAT32: return add();
            case schema::Type::FLOAT64: return add();
            case schema::Type::TEXT: return add();
            case schema::Type::DATA: return kj::strTree("  static_assert(false, \"DATA type not implemented yet.\");\n");
            case schema::Type::ENUM: return kj::strTree("  json.add(ds::tryFromEnum<const char *>(", name, ").value_or(\"", enumDefaultName, "\"));\n");
            case schema::Type::STRUCT: return kj::strTree(
              "  json.startObject();\n",
              "  ", name, ".addFieldsToJson(json);\n",
              "  json.endObject();\n");
            case schema::Type::LIST: {
              auto&& elementTypeSchema = typeSchema.asList().getElementType();
              return kj::strTree(
                "  json.startArray();\n",
                "  BOOST_FOREACH(auto&& _x, ", name, ") {\n",
                (elementTypeSchema.which() == schema::Type::STRUCT
                  ? "    json.startObject(); _x.addFieldsToJson(json); json.endObject();\n"
                  : "    json.add(_x);\n"),
                "  }\n",
                "  json.endArray();\n");
            }
            case schema::Type::INTERFACE: return kj::strTree("  static_assert(false, \"INTERFACE type not implemented yet.\");\n");
            case schema::Type::ANY_POINTER: return kj::strTree("  static_assert(false, \"ANY_POINTER type not implemented yet.\");\n");
          };
          return kj::strTree();
        }()),
      // setJson
      [&] {
        auto intCase = [&] {
          return kj::strTree(
            "  if (strcmp(dsCapnpKey, \"", name, "\") == 0) {\n",
            "    DS_CONTEXT(\"/", name, "\");\n",
            "    DS_REQUIRE(json_object_is_type(dsCapnpValue, json_type_double) ||\n"
            "        json_object_is_type(dsCapnpValue, json_type_int), \"Expecting json number\");\n",
            "    ", name, " = json_object_get_int64(dsCapnpValue);\n",
            "  }\n");
        };
        auto doubleCase = [&] {
          return kj::strTree(
            "  if (strcmp(dsCapnpKey, \"", name, "\") == 0) {\n",
            "    DS_CONTEXT(\"/", name, "\");\n",
            "    DS_REQUIRE(json_object_is_type(dsCapnpValue, json_type_double) ||\n"
            "        json_object_is_type(dsCapnpValue, json_type_int), \"Expecting json number\");\n",
            "    ", name, " = json_object_get_double(dsCapnpValue);\n",
            "  }\n");
        };

        switch (typeSchema.which()) {
          case schema::Type::VOID: return kj::strTree();
          case schema::Type::BOOL: return kj::strTree(
            "  if (strcmp(dsCapnpKey, \"", name, "\") == 0) {\n",
            "    DS_CONTEXT(\"/", name, "\");\n",
            "    DS_REQUIRE(json_object_is_type(dsCapnpValue, json_type_boolean), \"Expecting json boolean\");\n",
            "    ", name, " = json_object_get_boolean(dsCapnpValue);\n",
            "  }\n");
          case schema::Type::INT8: return intCase();
          case schema::Type::INT16: return intCase();
          case schema::Type::INT32: return intCase();
          case schema::Type::INT64: return intCase();
          case schema::Type::UINT8: return intCase();
          case schema::Type::UINT16: return intCase();
          case schema::Type::UINT32: return intCase();
          case schema::Type::UINT64: return intCase();
          case schema::Type::FLOAT32: return doubleCase();
          case schema::Type::FLOAT64: return doubleCase();
          case schema::Type::TEXT: return kj::strTree(
            "  if (strcmp(dsCapnpKey, \"", name, "\") == 0) {\n",
            "    DS_CONTEXT(\"/", name, "\");\n",
            "    DS_REQUIRE(json_object_is_type(dsCapnpValue, json_type_string), \"Expecting json string\");\n",
            "    ", name, " = json_object_get_string(dsCapnpValue);\n",
            "  }\n");
          case schema::Type::DATA: return kj::strTree();
          case schema::Type::ENUM: return kj::strTree(
            "  if (strcmp(dsCapnpKey, \"", name, "\") == 0) {\n",
            "    DS_CONTEXT(\"/", name, "\");\n",
            "    DS_REQUIRE(json_object_is_type(dsCapnpValue, json_type_string), \"Expecting json string\");\n",
            "    auto&& _dsCapnpStringValue = json_object_get_string(dsCapnpValue);\n",
            "    if (auto x = ds::tryToEnum<", type, ">(_dsCapnpStringValue)) {\n",
            "      ", name, " = *x;\n",
            "    } else {\n",
            "      DS_FAIL_REQUIRE(\"Unrecognized enum value\", _dsCapnpStringValue);\n",
            "    }\n",
            "  }\n");
          case schema::Type::STRUCT: return kj::strTree(
            "  if (strcmp(dsCapnpKey, \"", name, "\") == 0) {\n",
            "    DS_CONTEXT(\"/", name, "\");\n",
            "    ", name, ".setJson(dsCapnpValue);\n",
            "  }\n");
          case schema::Type::LIST: return kj::strTree();
          case schema::Type::INTERFACE: return kj::strTree();
          case schema::Type::ANY_POINTER: return kj::strTree();
        };
        return kj::strTree();
      }()
    };
  }

  struct EnumText {
    kj::StringTree decl;
    kj::StringTree def;
    kj::StringTree inlineDef;
    kj::StringTree sourceDef;
  };

  EnumText makeEnumText(kj::StringPtr scope, kj::StringPtr scope_, kj::StringPtr name, EnumSchema schema) {
    auto proto = schema.getProto();
    KJ_IF_MAYBE(annotatedName, annotationValue(proto, NAME_ANNOTATION_ID)) {
      name = annotatedName->getText();
    }
    auto fullName = kj::str(scope_, name);
    auto enumerants = schema.asEnum().getEnumerants();

    return EnumText {
      kj::strTree(
        "enum class ", fullName, " : uint16_t {\n",
        KJ_MAP(e, enumerants) {
          return kj::strTree("  ", toUpperCase(protoName(e.getProto())), ",\n");
        },
        "};\n"
      ),
      kj::strTree(
        "ds::Enum<", fullName, "> dsEnum(", fullName, ") __attribute__((const));\n",
        "std::ostream& operator<<(std::ostream&, ", fullName, ");\n",
        "\n"
      ),
      kj::strTree(),
      kj::strTree(
        "ds::Enum<", fullName, "> dsEnum(", fullName, ") {\n",
        "  static constexpr std::pair<", fullName, ", const char *> x[] = {\n",
        KJ_MAP(e, enumerants) {
          return kj::strTree("    {", fullName, "::", toUpperCase(protoName(e.getProto())), ", \"", protoName(e.getProto()), "\"},\n");
        },
        "  };\n",
        "  return {gsl::make_span(x), \"", fullName, "\"};\n",
        "}\n",
        "std::ostream& operator<<(std::ostream& ostream, ", fullName, " x) { return ds::ostream(ostream, x); }\n",
        "\n"
      )
    };
  }

  // -----------------------------------------------------------------

  struct FileText {
    kj::StringTree header;
    kj::StringTree source;
  };

  FileText makeFileText(Schema schema,
                        schema::CodeGeneratorRequest::RequestedFile::Reader request) {
    usedImports.clear();

    auto node = schema.getProto();
    auto displayName = node.getDisplayName();

    kj::Vector<kj::ArrayPtr<const char>> namespaceParts;
    kj::String namespacePrefix;

    for (auto annotation: node.getAnnotations()) {
      if (annotation.getId() == NAMESPACE_ANNOTATION_ID) {
        kj::StringPtr ns = annotation.getValue().getText();
        kj::StringPtr ns2 = ns;
        namespacePrefix = kj::str("::", ns);

        for (;;) {
          KJ_IF_MAYBE(colonPos, ns.findFirst(':')) {
            namespaceParts.add(ns.slice(0, *colonPos));
            ns = ns.slice(*colonPos);
            if (!ns.startsWith("::")) {
              context.exitError(kj::str(displayName, ": invalid namespace spec: ", ns2));
            }
            ns = ns.slice(2);
          } else {
            namespaceParts.add(ns);
            break;
          }
        }

        break;
      }
    }

    auto nodeTexts = KJ_MAP(nested, node.getNestedNodes()) {
      return makeNodeText(namespacePrefix, "", "", nested.getName(),
                          schemaLoader.getUnbound(nested.getId()));
    };

    kj::String separator = kj::str("// ", kj::repeat('=', 87), "\n");

    kj::Vector<kj::StringPtr> includes;
    for (auto import: request.getImports()) {
      if (usedImports.count(import.getId()) > 0) {
        includes.add(import.getName());
      }
    }

    kj::StringTree sourceDef = kj::strTree(
        KJ_MAP(n, nodeTexts) { return kj::mv(n.sourceDef); });

    return FileText {
      kj::strTree(
          "// Generated by Cap'n Proto compiler, DO NOT EDIT\n"
          "// source: ", baseName(displayName), "\n",
          "\n",
          "#pragma once\n",
          "\n",
          "#include <array>\n",
          "#include <string>\n",
          "#include <vector>\n",
          "#include <cstdint>\n",
          "#include <iosfwd>\n",
          "#include <utility>\n",
          "#include <ds/enum.h>\n",
          "\n",
          "extern \"C\" {\n"
          "typedef struct json_object json_object;\n",
          "}\n",
          "namespace dss {\n"
          "class JSONWriter;\n"
          "}\n",
          "\n",
          KJ_MAP(path, includes) {
            if (path.startsWith("/")) {
              return kj::strTree("#include <", path.slice(1), ".h>\n");
            } else {
              return kj::strTree("#include \"", path, ".h\"\n");
            }
          },
          "\n",

          KJ_MAP(n, namespaceParts) { return kj::strTree("namespace ", n, " {\n"); }, "\n",
          KJ_MAP(n, nodeTexts) { return kj::mv(n.decl); },
          "\n",
          separator, "\n",
          KJ_MAP(n, nodeTexts) { return kj::mv(n.def); },
          separator, "\n",
          KJ_MAP(n, nodeTexts) { return kj::mv(n.inlineDef); },
          KJ_MAP(n, namespaceParts) { return kj::strTree("}  // namespace\n"); }, "\n"
      ),

      kj::strTree(
          "// Generated by Cap'n Proto compiler, DO NOT EDIT\n"
          "// source: ", baseName(displayName), "\n",
          "\n",
          "#include \"", baseName(displayName), ".dss.h\"\n",
          "#include <boost/foreach.hpp>\n",
          "#include <json-c/json.h>\n",
          "#include <ds/enum.h>\n",
          "#include <ds/log.h>\n",
          "#include \"web/webrequests.h\"\n",
          "\n",
          sourceDef.size() == 0 ? kj::strTree() : kj::strTree(
              "\n", separator, "\n",
              KJ_MAP(n, namespaceParts) { return kj::strTree("namespace ", n, " {\n"); },
              "\n",
              kj::mv(sourceDef),
              KJ_MAP(n, namespaceParts) { return kj::strTree("}  // namespace\n"); }, "\n")
        )
    };
  }

  // -----------------------------------------------------------------

  void makeDirectory(kj::StringPtr path) {
    KJ_IF_MAYBE(slashpos, path.findLast('/')) {
      // Make the parent dir.
      makeDirectory(kj::str(path.slice(0, *slashpos)));
    }

    if (kj::miniposix::mkdir(path.cStr(), 0777) < 0) {
      int error = errno;
      if (error != EEXIST) {
        KJ_FAIL_SYSCALL("mkdir(path)", error, path);
      }
    }
  }

  void writeFile(kj::StringPtr filename, const kj::StringTree& text) {
    if (!filename.startsWith("/")) {
      KJ_IF_MAYBE(slashpos, filename.findLast('/')) {
        // Make the parent dir.
        makeDirectory(kj::str(filename.slice(0, *slashpos)));
      }
    }

    int fd;
    KJ_SYSCALL(fd = open(filename.cStr(), O_CREAT | O_WRONLY | O_TRUNC, 0666), filename);
    kj::FdOutputStream out((kj::AutoCloseFd(fd)));

    text.visit(
        [&](kj::ArrayPtr<const char> text) {
          out.write(text.begin(), text.size());
        });
  }

  kj::MainBuilder::Validity run() {
    ReaderOptions options;
    options.traversalLimitInWords = 1 << 30;  // Don't limit.
    StreamFdMessageReader reader(STDIN_FILENO, options);
    auto request = reader.getRoot<schema::CodeGeneratorRequest>();

    for (auto node: request.getNodes()) {
      schemaLoader.load(node);
    }

    kj::FdOutputStream rawOut(STDOUT_FILENO);
    kj::BufferedOutputStreamWrapper out(rawOut);

    for (auto requestedFile: request.getRequestedFiles()) {
      auto schema = schemaLoader.get(requestedFile.getId());
      auto fileText = makeFileText(schema, requestedFile);

      writeFile(kj::str(schema.getProto().getDisplayName(), ".dss.h"), fileText.header);
      writeFile(kj::str(schema.getProto().getDisplayName(), ".dss.cpp"), fileText.source);
    }
    return true;
  }
};

}  // namespace
}  // namespace capnp

KJ_MAIN(capnp::CapnpcCppMain)
