/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "watchdog.h"
#include <ds/log.h>
#include <atomic>
#include <boost/asio.hpp>
#include <boost/thread/condition_variable.hpp>
#include <thread>

static ds::LogChannel dsLogChannel("dsAsioWatchdog");

namespace ds {
namespace asio {

struct LatencyWatchdog::Impl {
  boost::asio::io_service &m_ioService;
  boost::chrono::milliseconds m_failureLatency;
  volatile bool m_shutdown;
  boost::mutex m_mutex;
  boost::condition_variable m_conditionVariable;
  std::thread m_thread;
  std::function<void()> m_fire;

  Impl(boost::asio::io_service &ioService, boost::chrono::milliseconds failureLatency)
      : m_ioService(ioService),
        m_failureLatency(failureLatency),
        m_shutdown(false),
        m_thread([this]() { run(); }),
        m_fire([this]() { defaultFire(); }) {
    DS_INFO_ENTER(m_failureLatency.count());
  }
  ~Impl() {
    DS_INFO_ENTER("");
    m_shutdown = true;
    m_conditionVariable.notify_one();
    m_thread.join();
  }

  void run() {
    DS_INFO_ENTER("");
    boost::unique_lock<boost::mutex> lock(m_mutex);
    boost::chrono::steady_clock clock;
    while (true) {
      auto timePoint = clock.now();
      std::atomic_bool pong(false);
      DS_INFO("Watchdog ping");
      m_ioService.dispatch([&]() {
        DS_INFO("Watchdog pong");
        pong = true;
      });
      m_conditionVariable.wait_for(lock, m_failureLatency, [&]() {
        auto duration = clock.now() - timePoint;
        auto durationMs = boost::chrono::duration_cast<boost::chrono::milliseconds>(duration).count();
        DS_INFO("Watchdog wait for ", m_shutdown, durationMs);
        return m_shutdown || (duration >= m_failureLatency);
      });
      if (m_shutdown) {
        break;
      }
      if (!pong) {
        DS_INFO("Fire.");
        m_fire();
      }
    }
  }

  void defaultFire() { DS_FAIL_ASSERT("LatencyWatchdog failed.", m_failureLatency.count()); }
};

LatencyWatchdog::LatencyWatchdog(boost::asio::io_service &ioService, boost::chrono::milliseconds failureLatency)
    : m_impl(new Impl(ioService, failureLatency)) {}

LatencyWatchdog::~LatencyWatchdog() = default;

void LatencyWatchdog::setFire(std::function<void()> &&x) { m_impl->m_fire = std::move(x); }

FileDescriptorsWatchdog::FileDescriptorsWatchdog(boost::asio::io_service &ioService, boost::chrono::milliseconds period)
    : m_timer(ioService), m_period(period), m_fire([this](const std::exception &e) { defaultFire(e); }) {
  DS_INFO_ENTER(m_period.count());
  asyncLoop();
}

FileDescriptorsWatchdog::~FileDescriptorsWatchdog() = default;

void FileDescriptorsWatchdog::asyncLoop() {
  m_timer.randomlyExpiresFromNowPercentDown(m_period, 20, [this] {
    DS_INFO_ENTER();
    try {
      auto &ioService = m_timer.ioService();
      boost::asio::local::stream_protocol::socket socket1(ioService);
      boost::asio::local::stream_protocol::socket socket2(ioService);
      boost::asio::local::connect_pair(socket1, socket2);
    } catch (const std::exception &e) {
      DS_INFO("Fire.", e);
      m_fire(e);
    }
    asyncLoop();
  });
}

void FileDescriptorsWatchdog::defaultFire(const std::exception &e) {
  DS_FAIL_ASSERT("FileDescriptorsWatchdog failed.", e);
}

} // namespace sqlite3
} // namespace ds
