/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <boost/asio/io_service.hpp>
#include <boost/chrono/chrono.hpp>
#include <functional>
#include <memory>
#include "timer.h"

namespace ds {
namespace asio {

/// Software watchdog that guards that ioService event loop is not stalled.
///
/// Spawns seperate thread and regurarly pings ioService from it.
class LatencyWatchdog {
public:
  LatencyWatchdog(boost::asio::io_service &ioService, boost::chrono::milliseconds failureLatency);
  ~LatencyWatchdog();

  /// Set function called when watchog times out. Default function asserts
  ///
  /// WARNING: callback is called from non-ioService thread.
  void setFire(std::function<void()> &&);

private:
  struct Impl;
  const std::unique_ptr<Impl> m_impl;
};

/// Software watchdog that guards that the process is able
/// to allocate new file descriptors.
///
/// When file descriptors are leaking and process reaches the limit
/// bad things tend to happen.
/// The best recovery tends to be process restart.
class FileDescriptorsWatchdog {
public:
  FileDescriptorsWatchdog(boost::asio::io_service &ioService, boost::chrono::milliseconds period);
  ~FileDescriptorsWatchdog();

  /// Set function called when watchog fires. Default asserts.
  void setFire(std::function<void(const std::exception &)> &&x) { m_fire = std::move(x); }

private:
  Timer m_timer;
  boost::chrono::milliseconds m_period;
  std::function<void(const std::exception &)> m_fire;

  void asyncLoop();
  void defaultFire(const std::exception &e);
};

} // namespace asio
} // namespace ds
