/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "common.h"
#include <cxxabi.h>
#include <execinfo.h>
#include <algorithm>
#include "gsl.h"
#include "log.h"

namespace ds {

StackTrace StackTrace::capture(size_t ignoreCount) {
  ignoreCount += 2; // this stack frame
  void* buffer[32];
  unsigned size = ::backtrace(buffer, ds::arraySize(buffer));
  auto values = gsl::make_span(buffer, buffer + size);
  values = values.subspan(std::min(ignoreCount, values.size()));
  return StackTrace{std::vector<void*>(values.begin(), values.end())};
}

std::ostream& operator<<(std::ostream& stream, const StackTrace& stackTrace) {
  // Based on https://panthema.net/2008/0901-stacktrace-demangled/

  const auto& values = stackTrace.values;
  if (values.empty()) {
    return stream << "  <empty bactrace, possibly corrupt>\n";
  }

  char** symbols = backtrace_symbols(values.data(), values.size());
  DS_DEFER(free(symbols));

  // allocate string which will be filled with the demangled function name
  size_t demangledNameSize = 256;
  char* demangledName = static_cast<char*>(malloc(demangledNameSize));
  DS_DEFER(free(demangledName));

  for (size_t i = 0; i < values.size(); i++) {
    auto&& symbol = symbols[i];
    char *beginName = 0, *beginOffset = 0, *endOffset = 0;

    // find parentheses and +address offset surrounding the mangled name:
    // ./module(function+0x15c) [0x8048a6d]
    for (char* p = symbol; *p; ++p) {
      if (*p == '(')
        beginName = p;
      else if (*p == '+')
        beginOffset = p;
      else if (*p == ')' && beginOffset) {
        endOffset = p;
        break;
      }
    }

    if (beginName && beginOffset && endOffset && beginName < beginOffset) {
      *beginName++ = '\0';
      *beginOffset++ = '\0';
      *endOffset = '\0';
      // mangled name is now in [beginName, beginOffset) and caller
      // offset in [beginOffset, endOffset). now apply __cxa_demangle():

      int status;
      stream << "  " << symbol << " : ";
      if (auto x = abi::__cxa_demangle(beginName, demangledName, &demangledNameSize, &status)) {
        stream << x;
      } else { // demangling failed
        stream << beginName;
      }
      stream << " +" << beginOffset << '\n';
    } else {
      // couldn't parse the line? print the whole line.
      stream << "  " << symbol << '\n';
    }
  }

  return stream;
}

} // namespace ds
