/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

// Requirements:
// - readonly key - value storage. key is string matching [a-z-]+, value is json value
// - possible to override at runtime in device. This override is preserved during update.
// - possible to override for feed
// - default value is in source code
//
// Config files:
// - {sysconfdir}/ds-conf/{key} holds runtime override - /etc/ds-conf/{key}
// - {datadir}/ds-conf/{key} holds feed override - /usr/share/ds-conf/{key}
//

#include <gsl.h>
#include <boost/optional.hpp>
#include "json.h"
#include "log.h"
#include "system.h"

namespace ds {

class Conf {
public:
  ~Conf() = default;

  /// Try read conf file.
  ///
  /// @param system System abstraction
  /// @param key Configuration key
  ///
  template <typename T>
  boost::optional<T> tryRead(gsl::czstring key);

  /// Handy override returning default value if conf file was not possible to read.
  template <typename T>
  T readOr(gsl::czstring key, T&& defaultValue);

  // implementation details

  enum Location {
    SYSCONF, // in /etc/
    DATA,    // in /usr/share/data
  };
  virtual boost::optional<JsonFragment> readJsonIfExists(Location location, gsl::czstring key) = 0;
  virtual LogChannel& logChannel() const = 0;
};

std::unique_ptr<Conf> makeConf(System& system);

// Implementation details

template <typename T>
boost::optional<T> Conf::tryRead(gsl::czstring key) {
  try {
    if (const auto& json = readJsonIfExists(Location::SYSCONF, key)) {
      return fromJson<T>(*json);
    }
  } catch (const std::exception& e) {
    DS_CHANNEL_ERROR(logChannel(), "Failed to parse sysconf value.", key, e);
  }

  try {
    if (const auto& json = readJsonIfExists(Location::DATA, key)) {
      return fromJson<T>(*json);
    }
  } catch (const std::exception& e) {
    DS_CHANNEL_ERROR(logChannel(), "Failed to parse data value.", key, e);
  }
  return {};
}

template <typename T>
T Conf::readOr(gsl::czstring key, T&& defaultValue) {
  return tryRead<T>(key).value_or(std::forward<T>(defaultValue));
}

} // namespace ds
