/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "enum.h"
#include <ds/catch/catch.h>
#include <ds/str.h>
#include <array>
#include <boost/optional/optional_io.hpp>

static const char* TAGS = "[dsEnum][ds]";

namespace {

enum class Transport { CAR, PLANE };

ds::Enum<Transport> dsEnum(Transport) {
  static constexpr std::pair<Transport, const char*> x[] = {
      {Transport::CAR, "car"}, {Transport::PLANE, "plane"},
  };
  return {gsl::make_span(x), "Transport"};
}

std::ostream& operator<<(std::ostream& stream, Transport x) { return ds::ostream(stream, x); }

TEST_CASE("For enum with dsEnum", TAGS) {
  SECTION("tryFromEnum<const char*> returns matching string") {
    CHECK(ds::tryFromEnum<const char*>(Transport::CAR).value() == std::string("car"));
    CHECK(ds::tryFromEnum<const char*>(Transport::PLANE).value() == std::string("plane"));
  }

  SECTION("tryFromEnum<const char*> returns none on non-matching option") {
    CHECK(bool(ds::tryFromEnum<const char*>(static_cast<Transport>(100))) == false);
  }

  SECTION("tryFromEnum<std::string> returns matching string") {
    CHECK(ds::tryFromEnum<std::string>(Transport::CAR).value() == "car");
    CHECK(ds::tryFromEnum<std::string>(Transport::PLANE).value() == "plane");
  }

  SECTION("tryFromEnum<std::string> returns none on non-matching option") {
    CHECK(bool(ds::tryFromEnum<std::string>(static_cast<Transport>(100))) == false);
  }

  SECTION("fromEnum<std::string> returns matching string") {
    CHECK(ds::fromEnum<std::string>(Transport::CAR) == "car");
    CHECK(ds::fromEnum<std::string>(Transport::PLANE) == "plane");
  }

  SECTION("fromEnum<std::string> returns number on non-matching option") {
    CHECK(ds::fromEnum<std::string>(static_cast<Transport>(100)) == "100");
    CHECK(ds::fromEnum<std::string>(static_cast<Transport>(-100)) == "-100");
  }

  SECTION("tryToEnum returns matching value") {
    CHECK(ds::tryToEnum<Transport>("car").value() == Transport::CAR);
    CHECK(ds::tryToEnum<Transport>("plane").value() == Transport::PLANE);
  }

  SECTION("tryToEnum returns none on non-matching option") {
    CHECK(bool(ds::tryToEnum<Transport>("notAMatchingOption")) == false);
  }

  SECTION("tryToEnum accepts std::string") {
    CHECK(ds::tryToEnum<Transport>(std::string("car")).value() == Transport::CAR);
  }

  SECTION("toEnum returns matching value") {
    CHECK(ds::toEnum<Transport>("car") == Transport::CAR);
    CHECK(ds::toEnum<Transport>("plane") == Transport::PLANE);
    CHECK(ds::toEnum<Transport>(std::string("plane")) == Transport::PLANE);
    CHECK(ds::toEnum<Transport>("5") == static_cast<Transport>(5));
  }

  SECTION("toEnum returns default value for unknown options") {
    CHECK(ds::toEnum<Transport>("unknown") == Transport());
  }

  SECTION("requireEnum return enum for known options") {
    CHECK(ds::requireEnum<Transport>("car") == Transport::CAR);
    CHECK(ds::requireEnum<Transport>("plane") == Transport::PLANE);
    CHECK(ds::requireEnum<Transport>(std::string("car")) == Transport::CAR);
  }

  SECTION("requireEnum throws for unknown options") {
    CHECK_THROWS_FIND(
        ds::requireEnum<Transport>("notAMatchingOption"), "Unknown Transport enum option. value:notAMatchingOption");
    CHECK_THROWS_FIND(ds::requireEnum<Transport>(std::string("notAMatchingOption")),
        "Unknown Transport enum option. value:notAMatchingOption");
  }

  SECTION("tryToEnum accepts boost::optional<std::string>") {
    CHECK(ds::tryToEnum<Transport>(boost::optional<std::string>("car")).value() == Transport::CAR);
    CHECK(ds::tryToEnum<Transport>(boost::optional<std::string>("plane")).value() == Transport::PLANE);
    CHECK(ds::tryToEnum<Transport>(boost::optional<std::string>("unknown")) == boost::none);
    CHECK(ds::tryToEnum<Transport>(boost::optional<std::string>()) == boost::none);
  }

  SECTION("tryToEnum accepts int") {
    CHECK(ds::tryToEnum<Transport>(0).value() == Transport::CAR);
    CHECK(ds::tryToEnum<Transport>(1).value() == Transport::PLANE);
    CHECK(ds::tryToEnum<Transport>(2) == boost::none);
    CHECK(ds::tryToEnum<Transport>(-1) == boost::none);
  }

  SECTION("tryToEnum accepts boost::optional<int>") {
    CHECK(ds::tryToEnum<Transport>(boost::optional<int>(0)).value() == Transport::CAR);
    CHECK(ds::tryToEnum<Transport>(boost::optional<int>(1)).value() == Transport::PLANE);
    CHECK(ds::tryToEnum<Transport>(boost::optional<int>(2)) == boost::none);
    CHECK(ds::tryToEnum<Transport>(boost::optional<int>(-1)) == boost::none);
    CHECK(ds::tryToEnum<Transport>(boost::optional<int>()) == boost::none);
  }

  SECTION("ostream << writes matching values as string") {
    CHECK(ds::str(Transport::CAR) == "car");
    CHECK(ds::str(Transport::PLANE) == "plane");
    CHECK(ds::str(static_cast<Transport>(100)) == "100");
  }

  SECTION("isValidEnum == true for matching value ") { CHECK(ds::isValidEnum(Transport::CAR)); }

  SECTION("isValidEnum == false for non-matching value ") { CHECK(!ds::isValidEnum(static_cast<Transport>(100))); }
}

} // namespace
