/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once
#include <gsl.h>
#include <boost/optional.hpp>
#include <cstring>
#include <ostream>
#include "log.h"

namespace ds {

// Introduces reflection for enums and handy template conversion functions
// using this reflection.
//
// Example use:
//
// 1. define the enum
//
//   enum class Transport { CAR, PLANE };
//
// 2. provide enum to string mapping
//
//   ds::Enum<Transport> dsEnum(Transport) {
//     static constexpr std::pair<Transport, const char*> x[] = {
//             {Transport::CAR, "car"}, {Transport::PLANE, "plane"},
//     };
//     return {gsl::make_span(x), "ds::Transport"};
//   }
//   inline std::ostream& operator<<(std::ostream& stream, Transport x) { return ds::ostream(stream, x); }
//
// NOTE: `dsEnum` and `operator ostream <<` must be declared
// in the same namespace as enum to let it just work everywhere.
//
// The `dsEnum` argument value is ignored,
// the argument only triggers C++ argument dependent lookup.
//
// 3. use
//
//   ds::tryFromEnum<const char*>(Transport:CAR) -> boost::optional<const char*> as "car", "plane", none
//   ds::tryFromEnum<std::string>(Transport:CAR) -> boost::optional<std::string> as "car", "plane", none
//   ds::tryFromEnum<int>(Transport:CAR) -> boost::optional<int> as 0, 1, none
//
//   // Use when the code is determined to use the value whether the option is known or not.
//   ds::fromEnum<std::string>(Transport:CAR) -> std::string as "car", "plane", "2"
//
//   ds::tryToEnum<Transport>("car") -> boost::optional<Transport>
//   ds::tryToEnum<Transport>(1) -> boost::optional<Transport>
//
//   // Use when the code is determined to use the value whether the option is known or not.
//   ds::toEnum<Transport>("car") -> Transport::CAR
//   ds::toEnum<Transport>("5") -> static_cast<Transport>(5)
//   ds::toEnum<Transport>("unknown") -> Transport()
//
//   // Use when the code wants to fail on unknown option
//   ds::requireEnum<Transport>(1) -> Transport as Transport::CAR, Transport::PLANE or throws
//
//   ds::isValidEnum(Transport::CAR) -> bool
//
//   std::cout << Transport:CAR -> serializes as "car", "plane", "2"
//

template <typename T>
struct Enum {
  gsl::span<const std::pair<T, const char*>> options;
  gsl::czstring name; // for better error messages
};

template <typename T, typename = decltype(dsEnum(T()))>
boost::optional<T> tryToEnum(const char* x) {
  BOOST_FOREACH (auto&& option, dsEnum(T()).options) {
    if (strcmp(x, option.second) == 0) {
      return option.first;
    }
  }
  return boost::none;
}

template <typename T, typename = decltype(dsEnum(T()))>
boost::optional<T> tryToEnum(const std::string& x) {
  return tryToEnum<T>(x.c_str());
}

template <typename T, typename = decltype(dsEnum(T()))>
T toEnum(const char* x) {
  BOOST_FOREACH (auto&& option, dsEnum(T()).options) {
    if (strcmp(option.second, x) == 0) {
      return option.first;
    }
  }
  return static_cast<T>(atoi(x));
}

template <typename T, typename = decltype(dsEnum(T()))>
T toEnum(const std::string& x) {
  return toEnum<T>(x.c_str());
}

template <typename T, typename = decltype(dsEnum(T()))>
boost::optional<T> tryToEnum(int x) {
  BOOST_FOREACH (auto&& option, dsEnum(T()).options) {
    if (static_cast<int>(option.first) == x) {
      return option.first;
    }
  }
  return boost::none;
}

template <typename T, typename = decltype(dsEnum(T()))>
boost::optional<T> tryToEnum(const boost::optional<std::string>& x) {
  if (x) {
    return tryToEnum<T>(*x);
  }
  return boost::none;
}

template <typename T, typename = decltype(dsEnum(T()))>
boost::optional<T> tryToEnum(const boost::optional<int>& x) {
  if (x) {
    return tryToEnum<T>(*x);
  }
  return boost::none;
}

template <typename T, typename Arg, typename = decltype(dsEnum(T()))>
T requireEnum(Arg&& value) {
  if (auto out = tryToEnum<T>(std::forward<Arg>(value))) {
    return *out;
  }
  DS_FAIL_REQUIRE("Unknown", ds::hideArg(dsEnum(T()).name), "enum option.", value);
}

namespace detail {
template <typename R, typename T, typename = decltype(dsEnum(T()))>
struct Enum {};
template <typename T>
struct Enum<const char*, T, decltype(dsEnum(T()))> {
  static boost::optional<const char*> tryFrom(T x) {
    BOOST_FOREACH (auto&& option, dsEnum(T()).options) {
      if (option.first == x) {
        return option.second;
      }
    }
    return boost::none;
  }
};
template <typename T>
struct Enum<std::string, T, decltype(dsEnum(T()))> {
  static boost::optional<std::string> tryFrom(T x) {
    BOOST_FOREACH (auto&& option, dsEnum(T()).options) {
      if (option.first == x) {
        return std::string(option.second);
      }
    }
    return boost::none;
  }
  static std::string from(T x) {
    BOOST_FOREACH (auto&& option, dsEnum(T()).options) {
      if (option.first == x) {
        return std::string(option.second);
      }
    }
    return std::to_string(static_cast<int>(x));
  }
};
template <typename T>
struct Enum<int, T, decltype(dsEnum(T()))> {
  static boost::optional<int> tryFrom(T x) {
    BOOST_FOREACH (auto&& option, dsEnum(T()).options) {
      if (option.first == x) {
        return static_cast<int>(x);
      }
    }
    return boost::none;
  }
};
} // namespace detail

template <typename R, typename T, typename = decltype(dsEnum(T()))>
boost::optional<R> tryFromEnum(T x) {
  return detail::Enum<R, T, decltype(dsEnum(T()))>::tryFrom(x);
}

template <typename R, typename T, typename = decltype(dsEnum(T()))>
R fromEnum(T x) {
  return detail::Enum<R, T, decltype(dsEnum(T()))>::from(x);
}

template <typename T, typename = decltype(dsEnum(T()))>
bool isValidEnum(T x) {
  return bool(tryFromEnum<const char*>(x));
}

template <typename T, typename = decltype(dsEnum(T()))>
std::ostream& ostream(std::ostream& stream, T x) {
  return stream << fromEnum<std::string>(x);
}

} // namespace ds
