/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <ds/common.h>
#include <gsl.h>
#include <memory>
#include <stdexcept>
#include <string>

namespace ds {

template <typename T>
struct Enum;

/// Exception thrown in case of errors.
///
/// Code that does not care about special values
/// like `type`, `userMessage` shall catch `std::exception` directly.
///
/// Use macros in log.h to make instances.
/// They capture additional context like
/// file name, line number, stack trace, etc.
/// See DS_REQUIRE, DS_FAIL_REQUIRE, DS_FAIL_WITH_USER_MESSAGE
///
/// Inheriting std::runtime_error instead of std::exception
/// for backwards compatibility with dss::DSSException.
/// New code shall not depend on it, it would be against
/// the error handling policy.
class Exception : public std::runtime_error {
public:
  /// What kind of failure?
  /// How shall the caller recover?
  /// Fail? Retry? Retry later? Retry now?
  ///
  /// From kj::Exception::Type
  enum class Type {

    // Something went wrong. This is the usual error type.
    FAILED = 0,

    /// The call failed because of a temporary lack of resources.
    /// This could be space resources
    /// (out of memory, out of disk space) or time resources
    /// (request queue overflow, operation
    /// timed out).
    ///
    /// The operation might work if tried again,
    /// but it should NOT be repeated immediately as this
    /// may simply exacerbate the problem.
    OVERLOADED = 1,

    /// The call required communication over a connection that has been lost.
    /// The caller will need to try again.
    DISCONNECTED = 2,

    // The requested method is not implemented. The caller may wish to revert to a fallback
    // approach based on other methods.
    UNIMPLEMENTED = 3
  };

  /// Create exception with given technical message.
  /// @arg message Technicall message in English, returned by what() (@see what()).
  explicit Exception(const std::string& message);

  /// Create exception with given technical message.
  /// @arg message Technicall message in English, returned by what() (@see what()).
  explicit Exception(const char* message);

  Exception(Type type, const std::string& message);
  Exception(Type type, const char* message);
  Exception(const std::string& message, std::string userMessage);
  Exception(Type type, const std::string& message, std::string userMessage);

  /// What kind of failure? @see Type
  Type type() const;

  // Technical message in english not intended for user.
  // Inherited from std::runtime_error.
  // const char* what() const;

  /// Translated error message intended for user.
  /// Frontends shall present a generic error message to user if empty.
  /// Can be multiline with markdown formatting
  const std::string& userMessage() const;

private:
  struct D;
  std::shared_ptr<D> m_maybeD;
};

ds::Enum<Exception::Type> dsEnum(Exception::Type);
std::ostream& operator<<(std::ostream& stream, Exception::Type x);

} // namespace ds
