/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "json.h"
#include <ds/catch/catch.h>
#include <ds/str.h>
#include <array>
#include <boost/optional/optional_io.hpp>
#include <boost/serialization/strong_typedef.hpp>
#include <cmath>
#include <tuple>
static const char* TAGS = "[dsJson][ds]";

namespace {

enum class Transport { CAR, PLANE };

ds::Enum<Transport> dsEnum(Transport) {
  static constexpr std::pair<Transport, const char*> x[] = {
      {Transport::CAR, "car"}, {Transport::PLANE, "plane"},
  };
  return {gsl::make_span(x), "Transport"};
}
std::ostream& operator<<(std::ostream& stream, Transport x) { return ds::ostream(stream, x); }

struct X {
  int i;
  std::string s;
  Transport transport;
  ds::JsonFragment json = nullptr;
  boost::optional<int> oi;
  X(int i = 0, std::string s = std::string{}, Transport transport = Transport{},
      ds::JsonFragment json = ds::JsonFragment(), boost::optional<int> oi = boost::optional<int>())
      : i(i), s(std::move(s)), transport(transport), json(std::move(json)), oi(std::move(oi)) {}

  std::tuple<const int&, const std::string&, const Transport&, const ds::JsonFragment&, const boost::optional<int>&>
  tie() const {
    return std::tie(i, s, transport, json, oi);
  }
};
bool operator==(const X& a, const X& b) { return a.tie() == b.tie(); }

BOOST_STRONG_TYPEDEF(std::string, Id);

} // namespace

DS_REFLECT(X, (i)(s)(transport)(json)(oi));

namespace {
TEST_CASE("fromJson", TAGS) {
  SECTION("bool") {
    CHECK(ds::fromJson<bool>("false") == false);
    CHECK(ds::fromJson<bool>("true") == true);

    SECTION("accepts only json bool") {
      CHECK_THROWS(ds::fromJson<std::string>("7"));
      CHECK_THROWS(ds::fromJson<std::string>("7.5"));
      CHECK_THROWS(ds::fromJson<std::string>("a"));
      CHECK_THROWS(ds::fromJson<std::string>("{}"));
      CHECK_THROWS(ds::fromJson<std::string>("[]"));
    }
  }

  SECTION("number") {
    CHECK(ds::fromJson<int>("7") == 7);
    CHECK(ds::fromJson<int>("5.4") == 5);
    CHECK(ds::fromJson<int>("5.5") == 5);
    CHECK(ds::fromJson<double>("7.5") == 7.5);
    CHECK(ds::fromJson<double>("7") == 7);
    CHECK(std::isnan(ds::fromJson<double>("\"nan\"")));
    CHECK(std::isnan(ds::fromJson<double>("\"NaN\"")));
    CHECK(ds::fromJson<double>("\"inf\"") == std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"Inf\"") == std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"infinity\"") == std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"Infinity\"") == std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"-inf\"") == -std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"-Inf\"") == -std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"-infinity\"") == -std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"-Infinity\"") == -std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<unsigned char>("7") == 7);
    CHECK(ds::fromJson<signed char>("7") == 7);
    CHECK(ds::fromJson<int>("\"7\"") == 7);

    SECTION("fails on narrowing error") { CHECK_THROWS(ds::fromJson<unsigned char>("1000")); }

    SECTION("accepts only numbers in strings") {
      // TODO(someday): the code needs to be fixed
    }

    SECTION("accepts only json numbers || strings") {
      CHECK_THROWS(ds::fromJson<int>("false"));
      CHECK_THROWS(ds::fromJson<int>("true"));
      CHECK_THROWS(ds::fromJson<int>("{}"));
      CHECK_THROWS(ds::fromJson<int>("[]"));
    }
  }

  SECTION("std::string") {
    CHECK(ds::fromJson<std::string>(R"("")") == "");
    CHECK(ds::fromJson<std::string>(R"("a")") == "a");
    CHECK(ds::fromJson<std::string>(R"("hello")") == "hello");
    CHECK(ds::fromJson<std::string>("\"one\\ttwo\"") == "one\ttwo");

    SECTION("accepts only json strings") {
      CHECK_THROWS(ds::fromJson<std::string>("false"));
      CHECK_THROWS(ds::fromJson<std::string>("true"));
      CHECK_THROWS(ds::fromJson<std::string>("7"));
      CHECK_THROWS(ds::fromJson<std::string>("7.5"));
      CHECK_THROWS(ds::fromJson<std::string>("{}"));
      CHECK_THROWS(ds::fromJson<std::string>("[]"));
    }
  }

  SECTION("enum") {
    CHECK(ds::fromJson<Transport>(R"("car")") == Transport::CAR);
    CHECK(ds::fromJson<Transport>(R"("plane")") == Transport::PLANE);
    CHECK(ds::fromJson<Transport>(R"("unknown")") == static_cast<Transport>(0));

    SECTION("accepts only json strings") {
      CHECK_THROWS(ds::fromJson<Transport>("false"));
      CHECK_THROWS(ds::fromJson<Transport>("true"));
      CHECK_THROWS(ds::fromJson<Transport>("7"));
      CHECK_THROWS(ds::fromJson<Transport>("7.5"));
      CHECK_THROWS(ds::fromJson<Transport>("{}"));
      CHECK_THROWS(ds::fromJson<Transport>("[]"));
    }
  }

  SECTION("boost::optional") {
    // json library cannot parse "null" :(
    CHECK(ds::fromJson<boost::optional<int>>(ds::JsonFragment()) == boost::none);
    CHECK(ds::fromJson<boost::optional<int>>(R"(7)") == 7);
  }

  SECTION("std::vector") {
    CHECK(ds::fromJson<std::vector<int>>(R"([3,1.4,6])") == std::vector<int>({3, 1, 6}));
    CHECK(ds::fromJson<std::vector<std::string>>(R"(["a","b"])") == std::vector<std::string>({"a", "b"}));

    SECTION("accepts only json array") {
      CHECK_THROWS(ds::fromJson<Transport>("false"));
      CHECK_THROWS(ds::fromJson<Transport>("true"));
      CHECK_THROWS(ds::fromJson<Transport>("7"));
      CHECK_THROWS(ds::fromJson<Transport>("7.5"));
      CHECK_THROWS(ds::fromJson<Transport>("a"));
      CHECK_THROWS(ds::fromJson<Transport>("{}"));
    }
  }

  SECTION("std::vector<std::pair<std::string, T>>") {
    typedef std::vector<std::pair<std::string, int>> VectorStringInt;
    typedef std::vector<std::pair<std::string, std::string>> VectorStringString;
    CHECK(ds::fromJson<VectorStringInt>(R"({})") == VectorStringInt{});
    {
      VectorStringInt expected{{"b", 1}, {"a", -1}, {"", 7}};
      CHECK(ds::fromJson<VectorStringInt>(R"({"b":1, "a":-1,"":7})") == expected);
    }
    {
      VectorStringString expected{{"b", "B"}};
      CHECK(ds::fromJson<VectorStringString>(R"({"b":"B"})") == expected);
    }

    SECTION("accepts only json object") {
      CHECK_THROWS(ds::fromJson<Transport>("false"));
      CHECK_THROWS(ds::fromJson<Transport>("true"));
      CHECK_THROWS(ds::fromJson<Transport>("7"));
      CHECK_THROWS(ds::fromJson<Transport>("7.5"));
      CHECK_THROWS(ds::fromJson<Transport>("a"));
      CHECK_THROWS(ds::fromJson<Transport>("[]"));
    }
  }

  SECTION("std::vector<std::pair<Id, T>>") {
    typedef std::vector<std::pair<Id, int>> VectorIdInt;
    CHECK(ds::fromJson<VectorIdInt>(R"({})") == VectorIdInt{});
    {
      VectorIdInt expected{{Id{"b"}, 1}, {Id{"a"}, -1}, {Id{}, 7}};
      CHECK(ds::fromJson<VectorIdInt>(R"({"b":1, "a":-1,"":7})") == expected);
    }
  }

  SECTION("std::set<Enum>") {
    typedef std::set<Transport> SetEnum;
    CHECK(ds::fromJson<SetEnum>(R"({})") == SetEnum{});
    {
      SetEnum expected{Transport::CAR, Transport::PLANE};
      CHECK(ds::fromJson<SetEnum>(R"({"car":true, "plane":true, "":true})") == expected);
    }
  }

  SECTION("JsonFragment") {
    CHECK(ds::fromJson<int>(ds::fromJson<ds::JsonFragment>("7")) == 7);
    CHECK(ds::fromJson<bool>(ds::fromJson<ds::JsonFragment>("false")) == false);
    CHECK(ds::fromJson<std::string>(ds::fromJson<ds::JsonFragment>("\"hi\"")) == "hi");
  }

  SECTION("struct") {
    CHECK(ds::fromJson<X>(R"({})") == X());
    CHECK(ds::fromJson<X>(R"({})") == X(0, "", Transport{}));
    CHECK(ds::fromJson<X>(R"({"i":4})") == X(4, "", Transport{}));
    CHECK(ds::fromJson<X>(R"({"s":"a"})") == X(0, "a", Transport{}));
    CHECK(ds::fromJson<X>(R"({"transport":"car"})") == X(0, "", Transport::CAR));
    CHECK(ds::fromJson<X>(R"({"transport":"plane"})") == X(0, "", Transport::PLANE));
    CHECK(ds::fromJson<X>(R"({"s":"a","i":7})") == X(7, "a", Transport{}));
    CHECK(ds::fromJson<X>(ds::fromJson<X>(R"({"json":{"s":"a","i":7}})").json) == X(7, "a", Transport{}));
    CHECK(ds::fromJson<X>(R"({"s":null})") == X());
    SECTION("optional<T> can be used to distinguish missing field / null field and value default") {
      CHECK(ds::fromJson<X>(R"({})") == X());
      CHECK(ds::fromJson<X>(R"({"oi":null})") == X());
      CHECK(ds::fromJson<X>(R"({"oi":0})") == X({}, {}, {}, {}, 0));
      CHECK(ds::fromJson<X>(R"({"oi":7})") == X({}, {}, {}, {}, 7));
    }

    SECTION("accepts only json object") {
      CHECK_THROWS(ds::fromJson<Transport>("false"));
      CHECK_THROWS(ds::fromJson<Transport>("true"));
      CHECK_THROWS(ds::fromJson<Transport>("7"));
      CHECK_THROWS(ds::fromJson<Transport>("7.5"));
      CHECK_THROWS(ds::fromJson<Transport>("a"));
      CHECK_THROWS(ds::fromJson<Transport>("[]"));
    }
  }
}

TEST_CASE("toJson", TAGS) {
  SECTION("bool") {
    CHECK(ds::toJson(false) == "false");
    CHECK(ds::toJson(true) == "true");
  }

  SECTION("number") {
    CHECK(ds::toJson(7) == "7");
    CHECK(std::atof(ds::toJson(5.5).c_str()) == 5.5);
    CHECK(ds::toJson((unsigned char)7) == "7");
    CHECK(ds::toJson((signed char)7) == "7");
    CHECK(ds::toJson(7LL) == "\"7\"");
    CHECK(ds::toJson(7ULL) == "\"7\"");
    CHECK(ds::toJson(9007199254740991ULL) == "\"9007199254740991\"");
    CHECK(ds::toJson(-9223372036854775807LL) == "\"-9223372036854775807\"");
  }

  SECTION("string") {
    CHECK(ds::toJson("") == "\"\"");
    CHECK(ds::toJson("a") == "\"a\"");
    CHECK(ds::toJson("hello") == "\"hello\"");
    CHECK(ds::toJson("one\ttwo") == "\"one\\ttwo\"");
    CHECK(ds::toJson(std::string{}) == "\"\"");
    CHECK(ds::toJson(std::string{"hello"}) == "\"hello\"");
  }

  SECTION("enum") {
    CHECK(ds::toJson(Transport::CAR) == "\"car\"");
    CHECK(ds::toJson(Transport::PLANE) == "\"plane\"");
  }

  SECTION("boost::optional") {
    CHECK(ds::toJson(boost::optional<int>()) == R"(null)");
    CHECK(ds::toJson(boost::optional<int>(7)) == R"(7)");
    CHECK(ds::toJson(boost::optional<std::string>("a")) == R"("a")");
  }

  SECTION("std::vector") {
    CHECK(ds::toJson(std::vector<int>({})) == R"([])");
    CHECK(ds::toJson(std::vector<int>({3, 1, 6})) == R"([3,1,6])");
    CHECK(ds::toJson(std::vector<std::string>{"a", "b"}) == R"(["a","b"])");
  }

  SECTION("std::vector<std::pair<std::string, T>>") {
    CHECK(ds::toJson(std::vector<std::pair<std::string, int>>({})) == R"({})");
    CHECK(ds::toJson(std::vector<std::pair<std::string, int>>({{"b", 3}, {"a", 7}})) == R"({"b":3,"a":7})");
    CHECK(ds::toJson(std::vector<std::pair<std::string, std::string>>({{"b", "B"}, {"", ""}})) == R"({"b":"B","":""})");
  }

  SECTION("std::set<Enum>") {
    CHECK(ds::toJson(std::set<Transport>({})) == R"({})");
    CHECK(ds::toJson(std::set<Transport>({Transport::CAR, Transport::PLANE, static_cast<Transport>(-1)})) ==
          R"({"car":true,"plane":true})");
  }

  SECTION("jsonObject") {
    CHECK(ds::toJson(ds::fromJson<ds::JsonFragment>("7")) == "7");
    CHECK(ds::toJson(ds::fromJson<ds::JsonFragment>("false")) == "false");
    CHECK(ds::toJson(ds::fromJson<ds::JsonFragment>("\"hi\"")) == "\"hi\"");
  }

  SECTION("struct") {
    CHECK(ds::toJson(X({})) == R"({})");
    CHECK(ds::toJson(X(0, "", Transport{})) == R"({})");
    CHECK(ds::toJson(X(4, "", Transport{})) == R"({"i":4})");
    CHECK(ds::toJson(X(0, "a", Transport{})) == R"({"s":"a"})");
    CHECK(ds::toJson(X(0, "", Transport::CAR)) == R"({})");
    CHECK(ds::toJson(X(0, "", Transport::PLANE)) == R"({"transport":"plane"})");
    CHECK(ds::toJson(X(7, "a", Transport{})) == R"({"i":7,"s":"a"})");
    CHECK(ds::toJson(X(7, "a", Transport{}, ds::toJsonFragment(X{0, "a", Transport::PLANE}))) ==
          R"({"i":7,"s":"a","json":{"s":"a","transport":"plane"}})");

    SECTION("optional<T> can be used to distinguish missing field / null field and value default") {
      CHECK(ds::toJson(X({}, {}, {}, {}, {})) == R"({})");
      CHECK(ds::toJson(X({}, {}, {}, {}, 0)) == R"({"oi":0})");
      CHECK(ds::toJson(X({}, {}, {}, {}, 7)) == R"({"oi":7})");
    }
  }
}

TEST_CASE("toJsonFragment", TAGS) {
  // toJson test covers all details.
  CHECK(ds::toJson(ds::toJsonFragment(X{0, "a", Transport::PLANE})) == R"({"s":"a","transport":"plane"})");
}

TEST_CASE("JsonFragment can be stringified", TAGS) {
  CHECK(ds::str(ds::toJsonFragment(X{0, "a", Transport::PLANE})) == R"({"s":"a","transport":"plane"})");
}
} // namespace
