/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <gsl.h>
#include <json-c/json.h>
#include <set>
#include "common.h"
#include "enum.h"
#include "reflect.h"
#include "str.h"

///
/// Json parsing and serialization for types with reflection.
///
/// Use:
///
///   ds::toJson(std::vector<int>({4,3})) -> "([4,3])"
///
///   ds::fromJson<std::vector<int>>("[4,3]") -> std::vector<int>({4,3})
///
/// Supported types:
///
/// bool
///   * bool: true, false
///
/// [u|]int[8|16|32]_t
///   * number: 1,-10,0,
///   * JSON value will be a decimal number. Either numbers or strings are accepted.
///
/// [u|]int64_t
///   * string: "1", "10""
///   * JSON value will be a decimal string. Either numbers or strings are accepted.
///   * Compatible with protobuf, capnproto.
///   * It may be worthwhile not use [u]int64_t in json apis. Consider string?
///
/// std::string
///   * string: "Hello world!"
///
/// enum with reflection (ds/enum.h)
///   * camelCase string or number: "fooBar", 4
///   * Unknown enum option is serialized as number.
///   * Only known string options are accepted.
///
/// vector<T>
///   * array: [v, ...]
///
/// vector<std::pair<std::string, T>>
///   * object: {"key1" : value1, ...}
///
/// vector<std::pair<Key, T>> - Key must be convertible to and from std::string
///   * object: {"key1" : value1, ...}
///
/// vector<std::set<Enum>>
///   * object: {"key1" : true, ...}
///   * Unknown enum options are skipped. Keys with `false` value are skipped.
///
/// struct with reflection (ds/reflect.h)
///   * object: {"name": "John", "height": 1.82}
///   * Field names use "lowerCamelCase".
///   * Missing fields preserve default value
///   * Fields with default value are not serialized.
///   * Fields with boost::optional<T> type recognize missing value and default value.
///
/// JsonFragment
///   * any json fragment injected / accepted
///   * for places where type is not known at parsing / serialization time
///

namespace ds {

struct JsonFragmentDeleter {
  void operator()(json_object* x) { json_object_put(x); }
};

// Json object with ownership.
typedef std::unique_ptr<json_object, JsonFragmentDeleter> JsonFragment;

std::ostream& operator<<(std::ostream&, const JsonFragment&);

template <typename T, class Enable = void>
struct FromJson;

template <typename T>
T fromJson(const JsonFragment& json) {
  return FromJson<T>::call(json.get());
}

template <typename T>
T fromJson(gsl::czstring text);

template <typename T>
T fromJson(const std::string& text) {
  return fromJson<T>(text.c_str());
}

template <typename T, class Enable = void>
struct ToJsonFragment;

template <typename T>
JsonFragment toJsonFragment(const T& x) {
  return ToJsonFragment<typename std::decay<T>::type>::call(x);
}

template <typename T>
std::string toJson(const T& x) {
  return json_object_to_json_string_ext(toJsonFragment(x).get(), 0);
}

// implementation details - from json

template <typename T>
T fromJson(gsl::czstring text) {
  json_tokener* tok = json_tokener_new();
  DS_REQUIRE(tok);
  DS_DEFER(json_tokener_free(tok));

  // we don't use gsl::string_span on input,
  // because json-c is not able to parse bool and numbers
  // with explicit len instead of -1
  json_object* json = json_tokener_parse_ex(tok, text, -1);
  DS_REQUIRE(json, "Failed to parse json.", ds::hideArg(json_tokener_error_desc(json_tokener_get_error(tok))));
  DS_DEFER(json_object_put(json));

  return FromJson<T>::call(json);
}

template <>
struct FromJson<bool> {
  static bool call(json_object* json) {
    const auto& type = json_object_get_type(json);
    DS_REQUIRE(type == json_type_boolean, "Expecting json boolen.", type);
    return json_object_get_boolean(json);
  }
};

template <typename T>
struct FromJson<T, typename std::enable_if<std::is_integral<T>::value>::type> {
  static T call(json_object* json) {
    const auto& type = json_object_get_type(json);
    DS_REQUIRE(type == json_type_string || type == json_type_int || type == json_type_double,
        "Expecting json number / string.", type);
    auto x = json_object_get_int64(json);
    return ds::narrow<T>(x);
  }
};

template <typename T>
struct FromJson<T, typename std::enable_if<std::is_floating_point<T>::value>::type> {
  static T call(json_object* json) {
    const auto& type = json_object_get_type(json);
    DS_REQUIRE(type == json_type_string || type == json_type_int || type == json_type_double,
        "Expecting json number / string.", type);
    return json_object_get_double(json);
  }
};

template <>
struct FromJson<std::string> {
  static std::string call(json_object* json) {
    const auto& type = json_object_get_type(json);
    DS_REQUIRE(type == json_type_string, type);
    return json_object_get_string(json);
  }
};

template <typename T>
struct FromJson<T, typename std::enable_if<std::is_enum<T>::value>::type> {
  static T call(json_object* json) {
    const auto& jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_string, "Expecting json string.", jsonType);
    return ds::toEnum<T>(json_object_get_string(json));
  }
};

template <typename T>
struct FromJson<boost::optional<T>> {
  static boost::optional<T> call(json_object* json) {
    auto jsonType = json_object_get_type(json);
    if (jsonType == json_type_null) {
      return {};
    }
    return FromJson<T>::call(json);
  }
};

template <typename Item>
struct FromJson<std::vector<Item>> {
  static std::vector<Item> call(json_object* json) {
    auto jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_array, "Expecting json array.", jsonType);
    std::vector<Item> result{};
    const auto& size = json_object_array_length(json);
    result.reserve(size);
    for (int i = 0; i < size; i++) {
      DS_NON_OE_CORE_CONTEXT("/", ds::hideArg(i));
      result.push_back(FromJson<Item>::call(json_object_array_get_idx(json, i)));
    }
    return result;
  }
};

template <typename First, typename Second>
struct FromJson<std::vector<std::pair<First, Second>>> {
  static std::vector<std::pair<First, Second>> call(json_object* json) {
    auto jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_object, "Expecting json object.", jsonType);
    std::vector<std::pair<First, Second>> result{};
    json_object_object_foreach(json, key, value) {
      DS_NON_OE_CORE_CONTEXT("/", ds::hideArg(key));
      result.push_back({First{key}, FromJson<Second>::call(value)});
    }
    return result;
  }
};

template <typename T>
struct FromJson<std::set<T>> {
  static std::set<T> call(json_object* json) {
    static_assert(std::is_enum<T>::value, "Enum type required as key");
    auto jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_object, "Expecting json object.", jsonType);
    std::set<T> result{};
    json_object_object_foreach(json, key, value) {
      DS_NON_OE_CORE_CONTEXT("/", ds::hideArg(key));
      auto valueType = json_object_get_type(value);
      DS_REQUIRE(valueType == json_type_boolean, "Expecting json bool.", valueType);
      if (json_object_get_boolean(value)) {
        if (auto enumKey = ds::tryToEnum<T>(key)) {
          result.insert(*enumKey);
        } else {
          // Unknown enum keys are skipped.
        }
      }
    }
    return result;
  }
};

template <>
struct FromJson<JsonFragment> {
  static JsonFragment call(json_object* json) { return JsonFragment{json_object_get(json)}; }
};

template <typename T>
struct FromJson<T, typename std::enable_if<ds::reflector<T>::is_defined::value>::type> {
  struct Visitor : boost::noncopyable {
    T& x;
    json_object* json;
    Visitor(T& x, json_object* json) : x(x), json(json) {}

    template <typename MemberPtr, MemberPtr m>
    void operator()(const char* name) const {
      DS_NON_OE_CORE_CONTEXT("/", ds::hideArg(name));

      auto& field = x.*m;
      typedef typename std::decay<decltype(field)>::type Field;

      json_object* jsonField = DS_NULLPTR;
      if (json_object_object_get_ex(json, name, &jsonField)) {
        if (json_object_get_type(jsonField) != json_type_null) {
          field = FromJson<Field>::call(jsonField);
        }
      }
    }
  };

  static T call(json_object* json) {
    auto jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_object, "Expecting json object.", jsonType);
    T result{};
    ds::reflector<T>::visit(Visitor(result, json));
    return result;
  }
};

// implementation details - toJson

typedef std::unique_ptr<json_object, JsonFragmentDeleter> JsonFragment;

template <>
struct ToJsonFragment<bool> {
  static JsonFragment call(bool x) { return JsonFragment{json_object_new_boolean(x)}; }
};

#define DS_INTEGER_CASE(TYPE)                                                           \
  template <>                                                                           \
  struct ToJsonFragment<TYPE> {                                                         \
    static JsonFragment call(TYPE x) { return JsonFragment{json_object_new_int64(x)}; } \
  };
DS_INTEGER_CASE(unsigned char);
DS_INTEGER_CASE(signed char);
DS_INTEGER_CASE(unsigned short);
DS_INTEGER_CASE(short);
DS_INTEGER_CASE(unsigned int);
DS_INTEGER_CASE(int);
#undef DS_INTEGER_CASE

template <>
struct ToJsonFragment<unsigned long long> {
  static JsonFragment call(unsigned long long x) {
    return JsonFragment{json_object_new_string(std::to_string(x).c_str())};
  }
};

template <>
struct ToJsonFragment<long long> {
  static JsonFragment call(long long x) { return JsonFragment{json_object_new_string(std::to_string(x).c_str())}; }
};

template <typename T>
struct ToJsonFragment<T, typename std::enable_if<std::is_integral<T>::value>::type> {
  static JsonFragment call(T x) { return JsonFragment{json_object_new_int64(x)}; }
};

template <typename T>
struct ToJsonFragment<T, typename std::enable_if<std::is_floating_point<T>::value>::type> {
  static JsonFragment call(T x) { return JsonFragment{json_object_new_double(x)}; }
};
inline JsonFragment toJsonFragment(gsl::cstring_span x) { return toJsonFragment<gsl::cstring_span>(x); }

template <typename T>
struct ToJsonFragment<T, typename std::enable_if<std::is_convertible<T, gsl::cstring_span>::value>::type> {
  static JsonFragment call(gsl::cstring_span x) { return JsonFragment{json_object_new_string_len(x.data(), x.size())}; }
};

template <typename T>
struct ToJsonFragment<T, typename std::enable_if<std::is_enum<T>::value>::type> {
  static JsonFragment call(T x) { return JsonFragment{json_object_new_string(ds::fromEnum<std::string>(x).c_str())}; }
};

template <typename T>
struct ToJsonFragment<boost::optional<T>> {
  static JsonFragment call(const boost::optional<T>& x) {
    if (x) {
      return ToJsonFragment<T>::call(*x);
    } else {
      return NULL;
    }
  }
};

template <typename Item>
struct ToJsonFragment<std::vector<Item>> {
  static JsonFragment call(const std::vector<Item>& items) {
    auto result = JsonFragment{json_object_new_array()};

    int i = 0;
    DS_NON_OE_CORE_CONTEXT("/", ds::hideArg(i));
    BOOST_FOREACH (const auto& item, items) {
      json_object_array_add(result.get(), ToJsonFragment<Item>::call(item).release());
      i++;
    }
    return result;
  }
};

template <typename Second>
struct ToJsonFragment<std::vector<std::pair<std::string, Second>>> {
  static JsonFragment call(const std::vector<std::pair<std::string, Second>>& items) {
    auto result = JsonFragment{json_object_new_object()};

    BOOST_FOREACH (const auto& item, items) {
      DS_NON_OE_CORE_CONTEXT("/", ds::hideArg(item.first));
      json_object_object_add(result.get(), item.first.c_str(), ToJsonFragment<Second>::call(item.second).release());
    }
    return result;
  }
};

template <typename T>
struct ToJsonFragment<std::set<T>> {
  static_assert(std::is_enum<T>::value, "Enum type required as key");
  static JsonFragment call(const std::set<T>& items) {
    auto result = JsonFragment{json_object_new_object()};

    BOOST_FOREACH (const auto& item, items) {
      DS_NON_OE_CORE_CONTEXT("/", ds::hideArg(item));
      if (auto key = ds::tryFromEnum<std::string>(item)) {
        json_object_object_add(result.get(), key->c_str(), json_object_new_boolean(true));
      } else {
        // Unknown enum keys are skipped.
      }
    }
    return result;
  }
};

template <>
struct ToJsonFragment<JsonFragment> {
  static JsonFragment call(const JsonFragment& x) { return JsonFragment{json_object_get(x.get())}; }
};

template <typename T>
struct ToJsonFragment<T, typename std::enable_if<ds::reflector<T>::is_defined::value>::type> {
  struct Visitor : boost::noncopyable {
    const JsonFragment& json;
    const T& x;
    Visitor(const JsonFragment& json, const T& x) : json(json), x(x) {}

    template <typename MemberPtr, MemberPtr m>
    void operator()(const char* name) const {
      DS_NON_OE_CORE_CONTEXT("/", ds::hideArg(name));

      const auto& field = x.*m;
      typedef typename std::decay<decltype(field)>::type Field;

      if (isDefaultInstance(field)) {
        return;
      }
      json_object_object_add(json.get(), name, ToJsonFragment<Field>::call(field).release());
    }
  };

  static JsonFragment call(const T& x) {
    JsonFragment result{json_object_new_object()};
    ds::reflector<T>::visit(Visitor(result, x));
    return result;
  }
};

} // namespace ds
