#include "messages.h"
#include <ds/log.h>
#include <ds/memory.h>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/foreach.hpp>
#include <boost/locale/gnu_gettext.hpp>
#include <boost/locale/util.hpp>
#include <boost/range/algorithm/find_if.hpp>
#include <boost/range/algorithm_ext/push_back.hpp>

static ds::LogChannel dsLogChannel("dsLocaleMessages");

namespace ds {
namespace locale {

namespace {

class Domain {
public:
  Domain(MessagesInfo info, std::string name) : m_language(info.language), m_name(std::move(name)) {
    DS_INFO_ENTER(*this);
    std::vector<std::string> splitDomains;
    boost::split(splitDomains, m_name, boost::is_any_of(","));
    m_splitDomainsCount = splitDomains.size();

    info.domains.clear();
    boost::push_back(info.domains, splitDomains);

    // load target language .mo files
    m_gettext = boost::locale::gnu_gettext::create_messages_facet<char>(info);
    // Subclasses of std::locale::facet are owned by std::locale instances.
    m_gettextLocale = std::locale(std::locale::classic(), m_gettext);

    // load en language .mo files
    info.language = "en_US";
    info.country.clear();
    info.variant.clear();
    info.encoding = "UTF-8";
    m_enGettext = boost::locale::gnu_gettext::create_messages_facet<char>(info);
    m_enGettextLocale = std::locale(std::locale::classic(), m_enGettext);
  }
  Domain(const Domain &) = delete;

  ~Domain() { DS_INFO_ENTER(*this); }

  const char *get(const char *context, const char *msgId) const {
    for (int domainId = 0; domainId < m_splitDomainsCount; ++domainId) {
      if (auto s = m_gettext->get(domainId, context, msgId)) {
        return s;
      }
    }
    auto &&domain = m_name;
    auto &&language = m_language;
    if (context) {
      DS_INFO("Missing translation", msgId, context, domain, language);
    } else {
      DS_INFO("Missing translation", msgId, domain, language);
    }
    for (int domainId = 0; domainId < m_splitDomainsCount; ++domainId) {
      if (auto s = m_enGettext->get(domainId, context, msgId)) {
        return s;
      }
    }
    return DS_NULLPTR;
  }

  const char *get(const char *context, const char *singleMsgId, int n) const {
    for (int domainId = 0; domainId < m_splitDomainsCount; ++domainId) {
      if (auto s = m_gettext->get(domainId, context, singleMsgId, n)) {
        return s;
      }
    }
    auto &&domain = m_name;
    auto &&language = m_language;
    if (context) {
      DS_INFO("Missing translation", singleMsgId, context, domain, language);
    } else {
      DS_INFO("Missing translation", singleMsgId, domain, language);
    }
    for (int domainId = 0; domainId < m_splitDomainsCount; ++domainId) {
      if (auto s = m_enGettext->get(domainId, context, singleMsgId, n)) {
        return s;
      }
    }
    return DS_NULLPTR;
  }

  const std::string &language() const { return m_language; }
  const std::string &name() const { return m_name; }

private:
  std::string m_language;
  std::string m_name;
  int m_splitDomainsCount;
  boost::locale::message_format<char> *m_gettext;   // owned by m_locale
  boost::locale::message_format<char> *m_enGettext; // owned by m_enLocale
  std::locale m_gettextLocale;
  std::locale m_enGettextLocale;
};

std::ostream &operator<<(std::ostream &s, const Domain &x) {
  return s << "(name:" << x.name() << " language:" << x.language() << ")";
}

class MessagesImpl : public Messages {
public:
  MessagesImpl(const MessagesInfo &info) : m_info(info) {
    if (m_info.encoding.empty()) {
      m_info.encoding = "UTF-8"; // our target encoding is by default UTF8
    }
    if (m_info.paths.size() == 0) {
      // DS_LOCALEDIR is passed as -DDS_LOCALEDIR='"$(localedir)"' from build system
      m_info.paths.push_back(DS_LOCALEDIR);
    }
    BOOST_FOREACH (auto &&domain, m_info.domains) { addDomain(domain.name); }
  }

  const char *get(int domainId, const char *context, const char *msgId) const DS_OVERRIDE {
    if (unsigned(domainId) < m_domains.size()) {
      return m_domains.at(domainId)->get(context, msgId);
    }
    return DS_NULLPTR;
  }

  const char *get(int domainId, const char *context, const char *singleMsgId, int n) const DS_OVERRIDE {
    if (unsigned(domainId) < m_domains.size()) {
      return m_domains.at(domainId)->get(context, singleMsgId, n);
    }
    return DS_NULLPTR;
  }

  /// Return domain id for given domain.
  /// If the domain is not already loaded, it is loaded now.
  int domain(const std::string &name) const DS_OVERRIDE {
    auto domainIt = boost::find_if(m_domains, [&](const std::unique_ptr<Domain> &x) { return x->name() == name; });
    if (domainIt != m_domains.end()) {
      return domainIt - m_domains.begin();
    }
    return addDomain(name);
  }

  const char *convert(const char *msg, string_type &buffer) const DS_OVERRIDE { return msg; }

  const std::string &language() const DS_OVERRIDE { return m_info.language; }

protected:
  // std::locale is the only one who can own and delete facets
  ~MessagesImpl() {}

private:
  MessagesInfo m_info;
  mutable std::vector<std::unique_ptr<Domain>> m_domains;

  int addDomain(const std::string &name) const {
    m_domains.push_back(ds::make_unique<Domain>(m_info, name));
    auto &&id = m_domains.size() - 1;
    DS_DEBUG_LEAVE(name, id);
    return id;
  }
};

} // namespace

Messages::~Messages() = default;

Messages *makeMessages(const MessagesInfo &info) { return new MessagesImpl(info); }

} // namespace locale
} // namespace ds
