#pragma once

// Origin: https://github.com/bytemaster/boost_reflect reviosion9597e954497db673b952bd6c08515e587723647a
// License: Boost Software License

#include <stdint.h>
#include <array>
#include <boost/function_types/result_type.hpp>
#include <boost/fusion/container/vector.hpp>
#include <boost/preprocessor/facilities/empty.hpp>
#include <boost/preprocessor/seq/enum.hpp>
#include <boost/preprocessor/seq/for_each.hpp>
#include <boost/preprocessor/seq/for_each_i.hpp>
#include <boost/preprocessor/seq/seq.hpp>
#include <boost/preprocessor/stringize.hpp>
#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/type_traits/remove_const.hpp>
#include <boost/type_traits/remove_pointer.hpp>
#include <boost/type_traits/remove_reference.hpp>
#include <boost/typeof/typeof.hpp>
#include "common.h"

// Reflects members of a struct.
//
// ````cpp
// struct A {
//   std::string hello;
//   std::string world;
// };
//
// // at global scope
// DS_REFLECT( A, (hello)(world) )
//
// template<typename T>
// struct Cout {
//     const T& x;
//     Cout(const T& x) : x(x) {}
//
//     template<typename MemberPtr, MemberPtr m>
//     void operator()(const char* name) const {
//       std::cout << name << ":" << x.*m << "\n";
//     }
// };
// ds::reflector<T>::visit(Cout<T>(a));
//
// [Original documentation](http://bytemaster.github.io/mace/group__mace__reflect__typeinfo.html)

namespace ds {

/// @brief defines reflection for struct T
/// Unless this is specialized, visit() will not be defined for T.
///
/// @tparam T - the type that will be visited.
///
/// The @ref DS_REFLECT(TYPE,MEMBERS) or DS_REFLECT_DERIVED(TYPE,BASES,MEMBERS) macro is used to specialize this
/// class for your type at global scope.
///
template <typename T>
struct reflector {
  typedef boost::false_type is_defined;
  //   typedef T type;
  //   typedef boost::fusion::vector<> base_class_types;

  //   @tparam Visitor a function object of the form:
  //
  //     @code
  //      struct functor {
  //         template<typename MemberPtr, MemberPtr m>
  //         void operator()( const char* name )const;
  //      };
  //     @endcode
  //
  //   @param v a functor that will be called for each member on T
  //
  //   template <typename Visitor>
  //   static constexpr void visit(Visitor&& v);
};

/// constexpr check that all struct fields are reflected
template <typename T>
struct CheckAllFieldsVisitor {
  size_t offset;
  DS_CONSTEXPR14 CheckAllFieldsVisitor() : offset(0) {}

  DS_CONSTEXPR14 void align(size_t x) {
    if (offset % x != 0) {
      offset += x - (offset % x);
    }
  }

  template <typename MemberPtr, MemberPtr m>
  DS_CONSTEXPR14 void operator()(const char* name) {
    align(std::alignment_of<decltype(std::declval<T>().*m)>::value);
    offset += sizeof(std::declval<T>().*m);
  }

  DS_CONSTEXPR14 bool check() {
    align(std::alignment_of<T>::value);
    return offset == sizeof(T);
  }
};

template <typename T>
DS_CONSTEXPR14 bool checkAllFields() {
  CheckAllFieldsVisitor<T> visitor;
  reflector<T>::visit(visitor);
  return visitor.check();
}

#define DS_REFLECT_VISIT_BASE(r, visitor, base) ds::reflector<base>::visit(visitor);

#define DS_REFLECT_VISIT_MEMBER(r, visitor, elem) \
  visitor.template operator()<BOOST_TYPEOF(&type::elem), &type::elem>(BOOST_PP_STRINGIZE(elem));

#define DS_REFLECT_BASE_MEMBER_COUNT(r, OP, elem) OP ds::reflector<elem>::total_member_count

#define DS_REFLECT_DERIVED_IMPL_INLINE(TYPE, INHERITS, MEMBERS) \
  template <typename Visitor>                                   \
  static DS_CONSTEXPR14 void visit(Visitor&& v) {               \
    BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_VISIT_BASE, v, INHERITS)   \
    BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_VISIT_MEMBER, v, MEMBERS)  \
  }

#define DS_REFLECT_DERIVED_IMPL_EXT(TYPE, INHERITS, MEMBERS)   \
  template <typename Visitor>                                  \
  void ds::reflector<TYPE>::visit(const Visitor& v) {          \
    BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_VISIT_BASE, v, INHERITS)  \
    BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_VISIT_MEMBER, v, MEMBERS) \
  }

/**
 *  @def DS_REFLECT_DERIVED(TYPE,INHERITS,MEMBERS)
 *
 *  @brief Specializes ds::reflector for TYPE where
 *         type inherits other reflected classes
 *
 *  @param INHERITS - a sequence of base class names (basea)(baseb)(basec)
 *  @param MEMBERS - a sequence of member names.  (field1)(field2)(field3)
 */
#define DS_REFLECT_DERIVED(TYPE, INHERITS, MEMBERS)                                                            \
  namespace ds {                                                                                               \
  template <>                                                                                                  \
  struct reflector<TYPE> {                                                                                     \
    typedef boost::true_type is_defined;                                                                       \
    typedef TYPE type;                                                                                         \
    enum member_count_enum {                                                                                   \
      local_member_count = BOOST_PP_SEQ_SIZE(MEMBERS),                                                         \
      total_member_count = local_member_count BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_BASE_MEMBER_COUNT, +, INHERITS) \
    };                                                                                                         \
    DS_REFLECT_DERIVED_IMPL_INLINE(TYPE, INHERITS, MEMBERS)                                                    \
  };                                                                                                           \
  DS_CONSTEXPR14_STATIC_ASSERT(checkAllFields<TYPE>(), "Some fields are not reflected.");                      \
  } // namespace ds

/**
 *  @def DS_REFLECT(TYPE,MEMBERS)
 *  @brief Specializes ds::reflector for TYPE
 *
 *  @param MEMBERS - a sequence of member names.  (field1)(field2)(field3)
 *
 *  @see DS_REFLECT_DERIVED
 */
#define DS_REFLECT(TYPE, MEMBERS) DS_REFLECT_DERIVED(TYPE, BOOST_PP_SEQ_NIL, MEMBERS)

template <typename T, typename Enable = void>
struct IsDefaultInstance {
  static bool call(const T& x) { return x == T(); }
};

/// Checks whether value is default instance.
/// Usefull in serialization code to skip fields with default value.
template <typename T>
bool isDefaultInstance(T&& x) {
  return IsDefaultInstance<typename std::decay<T>::type>::call(x);
}

template <>
struct IsDefaultInstance<const char*> {
  static bool call(const char* x) { return (x == DS_NULLPTR) || (*x == 0); }
};

template <>
struct IsDefaultInstance<char*> {
  static bool call(const char* x) { return IsDefaultInstance<const char*>::call(x); }
};

// Specialize for all containers with `empty()` method.
// Operator == in default template expends to more code
// and requires item's `operator == `.
template <typename T>
struct IsDefaultInstance<T,
    typename std::enable_if<std::is_same<bool, decltype(std::declval<T>().empty())>::value>::type> {
  static bool call(const T& x) { return x.empty(); }
};

template <typename T>
struct IsDefaultInstanceReflected : boost::noncopyable {
  const T& x;
  bool result;
  IsDefaultInstanceReflected(const T& x) : x(x), result(true) {}

  template <typename MemberPtr, MemberPtr m>
  void operator()(const char*) {
    result = result && isDefaultInstance(x.*m);
  }
};

template <typename T>
struct IsDefaultInstance<T, typename std::enable_if<ds::reflector<T>::is_defined::value>::type> {
  static bool call(const T& x) {
    IsDefaultInstanceReflected<T> visitor(x);
    reflector<T>::visit(visitor);
    return visitor.result;
  }
};


} // namespace ds
