/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "str.h"
#include <ds/catch/catch.h>
#include <boost/make_shared.hpp>
#include <locale>
#include <stdexcept>
#include "log.h"

static const char* TAGS = "[dsStr][ds]";

namespace {
enum class Enum { A, B };
std::ostream& operator<<(std::ostream& stream, Enum x) {
  switch (x) {
    case Enum::A:
      return stream << "a";
    case Enum::B:
      return stream << "b";
  }
  return stream << "?";
}

struct X { int i; char c; unsigned char uc; std::string s;};
struct Y {
  X x;
  std::vector<X> xs;
  Enum enum_;
};

} // namespace

DS_REFLECT(X, (i)(c)(uc)(s));
DS_REFLECT(Y, (x)(xs)(enum_));

namespace {

TEST_CASE("dsStr", TAGS) {
  SECTION("all arguments are serialized and concatenated") {
    CHECK(ds::str() == "");
    CHECK(ds::str("a") == "a");
    CHECK(ds::str("a", "b") == "ab");
    CHECK(ds::str("a", 4, "b") == "a4b");
    CHECK(ds::str("a", -4, "b") == "a-4b");
  }

  SECTION("ostream operator << is used for serialization") { CHECK(ds::str("-", Enum::A, Enum::B, "-") == "-ab-"); }

  SECTION("unsigned and signed chars are serialized as numbers") {
    CHECK(ds::str(static_cast<unsigned char>('a')) == "97");
    CHECK(ds::str(static_cast<signed char>('a')) == "97");
    CHECK(ds::str(static_cast<uint8_t>('a')) == "97");
    CHECK(ds::str(static_cast<int8_t>('a')) == "97");
  }

  SECTION("chars are serialized as characters") {
    CHECK(ds::str('a') == "a");
    CHECK(ds::str(static_cast<char>('a')) == "a");
  }

  SECTION("boost::shared_ptr is serialized as value or null") {
    CHECK(ds::str(boost::make_shared<int>(7)) == "7");
    CHECK(ds::str(boost::shared_ptr<int>()) == "null");
  }

  SECTION("std:exceptions references are serialized with what()") {
    auto e = std::runtime_error("a");
    CHECK(ds::str(static_cast<const std::exception&>(e)) == "a");
    CHECK(ds::str(static_cast<std::exception&>(e)) == "a");
  }

  SECTION("std::vector<T> is serialized") {
    CHECK(ds::str(std::vector<int>()) == "[]");
    CHECK(ds::str(std::vector<int>({1})) == "[1]");
    CHECK(ds::str(std::vector<int>({1, 2})) == "[1,2]");
    CHECK(ds::str(std::vector<std::string>()) == "[]");
    CHECK(ds::str(std::vector<std::string>({"a"})) == "[a]");
    CHECK(ds::str(std::vector<std::string>({"a", "b"})) == "[a,b]");
    CHECK(ds::str(std::vector<std::vector<int>>({{1}, {}})) == "[[1],[]]");
  }

  SECTION("std::set<T> is serialized") {
    CHECK(ds::str(std::set<int>()) == "[]");
    CHECK(ds::str(std::set<int>({1})) == "[1]");
    CHECK(ds::str(std::set<int>({2, 1})) == "[1,2]");
    CHECK(ds::str(std::set<std::string>()) == "[]");
    CHECK(ds::str(std::set<std::string>({"a"})) == "[a]");
    CHECK(ds::str(std::set<std::string>({"a", "b"})) == "[a,b]");
  }

  SECTION("std::pair<T1, T2> is serialized") {
    CHECK(ds::str(std::make_pair(0, "a")) == "(0,a)");
    CHECK(ds::str(std::make_pair(std::make_pair(1, 2), "a")) == "((1,2),a)");
  }

  SECTION("struct with reflection is serialized") {
    X x = {};
    x.i = 8;
    x.c = 'f';
    x.uc = 9;
    x.s = "hi";
    Y y = {};
    y.x = x;
    y.xs.push_back(x);
    y.xs.push_back(x);
    y.xs.back().i = 0;
    y.enum_ = Enum::B;

    CHECK(ds::str(y) == R"({x:{i:8 c:f uc:9 s:hi} xs:[{i:8 c:f uc:9 s:hi},{c:f uc:9 s:hi}] enum_:b})");
  }

  SECTION("(const char *) nullptr is serialized as (nullptr)") {
    CHECK(ds::str((const char*)DS_NULLPTR) == "(nullptr)");
  }
}

} // namespace
