/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <gsl.h>
#include <array>
#include <boost/foreach.hpp>
#include <boost/shared_ptr.hpp>
#include <sstream>
#include "common.h"
#include "reflect.h"

namespace google {
namespace protobuf {
class Message; // Avoid linking to protobuf
} // namespace protobuf
} // namespace google

namespace ds {

/// Concatenate all arguments into string using std::ostream operator <<
/// Example: ds:str("int:", 7, " double:", 7.7, " char:", '7'); // == "int:7 double:7.7 char:7"
template <typename... Args>
std::string str(Args&&... args);

// implementation details

namespace detail {

template <typename T, class Enable = void>
struct Str {
  template <typename T2>
  static void f(std::ostream& ostream, T2&& x, int indent) {
    ostream << std::forward<T2>(x);
  }
};

// uint8_t and int8_t shall be serialized as numbers, not characters
template <>
struct Str<unsigned char> {
  static void f(std::ostream& ostream, unsigned char x, int indent);
};

template <>
struct Str<signed char> {
  static void f(std::ostream& ostream, signed char x, int indent);
};

template <typename T>
struct Str<boost::shared_ptr<T>> {
  static void f(std::ostream& ostream, const boost::shared_ptr<T>& x, int indent) {
    if (x) {
      Str<typename std::decay<decltype(*x)>::type>::f(ostream, *x, indent);
    } else {
      ostream << "null";
    }
  }
};

template <>
struct Str<std::exception> {
  static void f(std::ostream& ostream, const std::exception& x, int indent);
};

template <typename Range>
struct Str<Range,
    typename std::enable_if<IsRange<Range>::value && !std::is_convertible<Range, gsl::cstring_span>::value>::type> {
  static inline void f(std::ostream& ostream, const Range& xs, int indent) {
    ostream << '[';
    auto first = true;
    BOOST_FOREACH (const auto& x, xs) {
      if (!first) {
        ostream << ',';
      } else {
        first = false;
      }
      Str<typename std::decay<decltype(x)>::type>::f(ostream, x, indent + 1);
    }
    ostream << ']';
  }
};

template <typename T1, typename T2>
struct Str<std::pair<T1, T2>> {
  static inline void f(std::ostream& ostream, const std::pair<T1, T2>& x, int indent) {
    ostream << '(';
    Str<typename std::decay<T1>::type>::f(ostream, x.first, indent);
    ostream << ',';
    Str<typename std::decay<T2>::type>::f(ostream, x.second, indent);
    ostream << ')';
  }
};

// Serialize protobuf messages, avoid linking to protobuf library.
template <typename T>
struct Str<T, typename std::enable_if<std::is_base_of<google::protobuf::Message, T>::value>::type> {
  static inline void f(std::ostream& ostream, const T& x, int indent) { ostream << x.DebugString(); }
};

template <typename T>
struct StrReflect : boost::noncopyable {
  std::ostream& ostream;
  const T& x;
  mutable bool first;
  int indent;
  StrReflect(std::ostream& ostream, const T& x, int indent) : ostream(ostream), x(x), first(true), indent(indent + 1) {}

  template <typename MemberPtr, MemberPtr m>
  void operator()(const char* name) const {
    const auto& value = x.*m;
    typedef typename std::decay<decltype(value)>::type Member;

    if (isDefaultInstance(value)) {
      return;
    }
    if (first) {
      first = false;
    } else {
      ostream << ' ';
    }
    ostream << name << ':';
    Str<Member>::f(ostream, value, indent);
  }
};

template <typename T>
struct Str<T, typename std::enable_if<ds::reflector<T>::is_defined::value>::type> {
  static inline void f(std::ostream& ostream, const T& x, int indent) {
    ostream << "{";
    ds::reflector<T>::visit(StrReflect<T>(ostream, x, indent));
    ostream << "}";
  }
};

// (const char *) nullptr is serialized as (nullptr).
// Default behavior is to terminate the stream.
template <>
struct Str<const char*> {
  static inline void f(std::ostream& ostream, const char* x, int indent) {
    if (x) {
      ostream << x;
    } else {
      ostream << "(nullptr)";
    }
  }
};

inline void strRecursive(std::ostream&) {}
template <typename Arg, typename... Args>
void strRecursive(std::ostream& ostream, Arg&& arg, Args&&... args) {
  Str<typename std::decay<Arg>::type>::f(ostream, std::forward<Arg>(arg), 0);
  strRecursive(ostream, std::forward<Args>(args)...);
}

} // namespace detail

template <typename... Args>
std::string str(Args&&... args) {
  std::ostringstream ostream;
  detail::strRecursive(ostream, std::forward<Args>(args)...);
  return ostream.str();
}

} // namespace ds
