/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "system-clock.h"
#include <ds/catch/catch.h>
#include <ds/catch/fakeit.h>
#include <ds/log.h>

static const char* TAGS = "[ds]";

namespace {

TEST_CASE("dsSystemClock", TAGS) {
  auto timePoint = ds::SystemClock::TimePoint(std::chrono::milliseconds(1505731033777));

  SECTION("returns system_clock by default") {
    const auto& clock = ds::systemClock();
    auto dsNow = clock.now();
    auto stdNow = std::chrono::system_clock::now();
    CHECK(dsNow <= stdNow);
    CHECK(stdNow - dsNow <= std::chrono::seconds(5));
  }

  SECTION("can be mocked") {
    Mock<ds::SystemClock> clockMock;
    Method(clockMock, now) = timePoint;

    auto& clock = clockMock.get();
    CHECK(clock.now() == timePoint);
  }

  SECTION("toLocalMsIsoString") {
    auto string = ds::SystemClock::toLocalMsIsoString(timePoint);
    // 2017-09-18T12:37:13.777+02:00 but time zone may differ
    CHECK(string.size() == 29);

    tm tm_ = {};
    int ms;
    char tzSign;
    int tzH, tzM;
    CHECK(::sscanf(string.c_str(), "%d-%d-%dT%d:%d:%d.%d%c%d:%d", &tm_.tm_year, &tm_.tm_mon, &tm_.tm_mday, &tm_.tm_hour,
              &tm_.tm_min, &tm_.tm_sec, &ms, &tzSign, &tzH, &tzM) == 10);
    CHECK((tzSign == '+' || tzSign == '-'));
    tm_.tm_gmtoff = (tzH * 60 * 60 + tzM * 60) * (tzSign == '-' ? -1 : 1);
    CHECK(tm_.tm_year == 2017);
    CHECK(tm_.tm_mon == 9);

    // convert the rest (days, hours, etc) to duration and compare
    auto duration1 = std::chrono::hours(tm_.tm_mday) * 24 + std::chrono::hours(tm_.tm_hour) +
                     std::chrono::minutes(tm_.tm_min) + std::chrono::seconds(tm_.tm_sec) +
                     std::chrono::milliseconds(ms) - std::chrono::seconds(tm_.tm_gmtoff);
    auto duration2 = std::chrono::hours(18) * 24 + std::chrono::hours(12) + std::chrono::minutes(37) +
                     std::chrono::seconds(13) + std::chrono::milliseconds(777) - std::chrono::seconds(2 * 3600);
    CHECK(duration1.count() == duration2.count());
  }

  SECTION("toMsIsoString") {
    auto string = ds::SystemClock::toMsIsoString(timePoint);
    CHECK(string == "2017-09-18T10:37:13.777Z");
  }
}

} // namespace
