/*
    Copyright (c) 2014 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "vdc-connection.h"

#include <iostream>
#include <boost/make_shared.hpp>

#include "base.h"
#include "util.h"
#include "ds485types.h"
#include "dss.h"
#include "model/apartment.h"
#include "ds485/dsbusinterface.h"
#include "messages/vdc-messages.pb.h"
#include "logger.h"
#include "model-features.h"
#include "vdc-element-reader.h"

static ds::LogChannel dsLogChannel("vdcConnection");
namespace dss {

  VdsdSpec_t VdcHelper::getSpec(dsuid_t _vdsm, dsuid_t _device) {
    vdcapi::Message message = VdcConnection::getProperty(
        _vdsm, _device, {"hardwareModelGuid", "modelUID", "modelFeatures", "vendorGuid", "vendorId", "vendorName",
                            "oemGuid", "oemModelGuid", "configURL", "hardwareGuid", "model", "modelVersion",
                            "hardwareVersion", "name", "displayId", "deviceClass", "deviceClassVersion"});

    VdsdSpec_t ret;

    VdcElementReader rootReader(message.vdc_response_get_property().properties());
    ret.hardwareModelGuid = rootReader["hardwareModelGuid"].getValueAsString();
    ret.vendorGuid = rootReader["vendorGuid"].getValueAsString();
    ret.oemGuid = rootReader["oemGuid"].getValueAsString();
    ret.oemModelGuid = rootReader["oemModelGuid"].getValueAsString();
    ret.configURL = rootReader["configURL"].getValueAsString();
    ret.hardwareGuid = rootReader["hardwareGuid"].getValueAsString();
    ret.model = rootReader["model"].getValueAsString();
    ret.modelUID = rootReader["modelUID"].getValueAsString();
    ret.hardwareVersion = rootReader["hardwareVersion"].getValueAsString();
    ret.name = rootReader["name"].getValueAsString();
    ret.vendorId = rootReader["vendorId"].getValueAsString();
    ret.vendorName = rootReader["vendorName"].getValueAsString();
    ret.modelVersion = rootReader["modelVersion"].getValueAsString();
    ret.displayId = rootReader["displayId"].getValueAsString();
    ret.descriptionsGroup = rootReader["descriptionsGroup"].getValueAsString();
    ret.descriptionsClass = rootReader["descriptionsClass"].getValueAsString();

    auto& features = ret.modelFeatures;
    VdcElementReader featuresReader = rootReader["modelFeatures"];
    for (VdcElementReader::iterator it = featuresReader.begin(); it != featuresReader.end(); it++) {
      VdcElementReader featureReader = *it;
      if (!featureReader.getValueAsBool()) {
        continue;
      }
      auto&& featureName = featureReader.getName();
      if (auto&& feature = ds::tryToEnum<ModelFeatureId>(featureName)) {
        features.insert(*feature);
      } else {
        Logger::getInstance()->log("Ignoring feature '" + featureName + "' from device " +
                                    dsuid2str(_device));
      }
    }

    DS_INFO("Vdc device spec.", ret);
    return ret;
  }

  boost::shared_ptr<VdcSpec> VdcHelper::getVdcSpec(dsuid_t _vdsm) {
    vdcapi::Message message = VdcConnection::getProperty(_vdsm, _vdsm,
        {"capabilities", "modelVersion", "model", "hardwareVersion", "hardwareModelGuid", "implementationId",
            "modelUID", "vendorGuid", "oemGuid", "oemModelGuid", "configURL", "hardwareGuid", "name", "displayId"});

    boost::shared_ptr<VdcSpec> ret = boost::make_shared<VdcSpec>();
    vdcapi::vdc_ResponseGetProperty response = message.vdc_response_get_property();
    VdcElementReader properties(response.properties());
    auto&& capabilities = properties["capabilities"];

    ret->capabilities.metering = capabilities["metering"].getValueAsBool();
    ret->capabilities.identification = capabilities["identification"].getValueAsBool();
    ret->capabilities.dynamicDefinitions = capabilities["dynamicDefinitions"].getValueAsBool();
    ret->modelVersion = properties["modelVersion"].getValueAsString();
    ret->model = properties["model"].getValueAsString();
    ret->hardwareVersion = properties["hardwareVersion"].getValueAsString();
    ret->hardwareModelGuid = properties["hardwareModelGuid"].getValueAsString();
    ret->implementationId = properties["implementationId"].getValueAsString();
    ret->vendorGuid = properties["vendorGuid"].getValueAsString();
    ret->oemGuid = properties["oemGuid"].getValueAsString();
    ret->oemModelGuid = properties["oemModelGuid"].getValueAsString();
    ret->configURL = properties["configURL"].getValueAsString();
    ret->hardwareGuid = properties["hardwareGuid"].getValueAsString();
    ret->modelUID = properties["modelUID"].getValueAsString();
    ret->name = properties["name"].getValueAsString();
    ret->displayId = properties["displayId"].getValueAsString();

    return ret;
  }

  void VdcHelper::getIcon(dsuid_t _vdsm, dsuid_t _device, size_t *size, uint8_t **data) {
    *size = 0;
    *data = NULL;

    vdcapi::Message message = VdcConnection::getProperty(_vdsm, _device, {"deviceIcon16"});

    vdcapi::vdc_ResponseGetProperty response = message.vdc_response_get_property();
    if (response.properties_size() == 0) {
      return;
    }

    for (int i = 0; i < response.properties_size(); i++) {
      vdcapi::PropertyElement el = response.properties(i);

      if (!el.has_name() || !el.has_value()) {
        continue;
      }

      vdcapi::PropertyValue val = el.value();
      if (el.name() == "deviceIcon16" && (val.has_v_bytes())) {
        std::string icon = val.v_bytes();
        if (icon.length() > 0) {
          *data = (uint8_t *)malloc(sizeof(uint8_t) * icon.length());
          if (*data == NULL) {
            throw std::runtime_error("could not allocate memory for buffer");
          }
          memcpy(*data, icon.data(), icon.length());
          *size = icon.length();
        }
      }
    }

    return;
  }

  VdcHelper::State VdcHelper::getState(dsuid_t _vdsm, dsuid_t _device)
  {
    vdcapi::Message message = VdcConnection::getProperty(_vdsm, _device, {"binaryInputStates", "deviceStates"});

    Logger::getInstance()->log("VdcHelper::getState: message " + message.DebugString(), lsDebug);
    vdcapi::vdc_ResponseGetProperty response = message.vdc_response_get_property();
    VdcElementReader reader(message.vdc_response_get_property().properties());

    State state;
    auto& binaryInputStates = state.binaryInputStates;
    VdcElementReader biStatesReader = reader["binaryInputStates"];
    for (VdcElementReader::iterator it = biStatesReader.begin(); it != biStatesReader.end(); it++) {
      VdcElementReader biStateReader = *it;
      const std::string& biStateName = biStateReader.getName();
      {
        VdcElementReader reader = biStateReader["error"];
        if (reader.getValueAsInt() != 0) {
            Logger::getInstance()->log("VdcHelper::getStateInputValue: device:" + dsuid2str(_device) +
                " name:" + biStateName + " error:" + reader.getValueAsString(), lsWarning);
            continue;
        }
      }
      {
        VdcElementReader reader = biStateReader["extendedValue"];
        if (reader.isValid()) {
          binaryInputStates.push_back(std::pair<std::string, int> (biStateName, reader.getValueAsInt()));
          continue;
        }
      }
      {
        VdcElementReader reader = biStateReader["value"];
        if (reader.isValid()) {
          int value = static_cast<int> (reader.getValueAsBool() ? BinaryInputStateValue::Active : BinaryInputStateValue::Inactive);
          binaryInputStates.push_back(std::pair<std::string, int> (biStateName, value));
          continue;
        }
      }
    }

    std::vector<std::pair<std::string, std::string> >& deviceStates = state.deviceStates;
    VdcElementReader deviceStatesReader = reader["deviceStates"];
    for (VdcElementReader::iterator it = deviceStatesReader.begin(); it != deviceStatesReader.end(); it++) {
      VdcElementReader reader = *it;
      deviceStates.push_back(std::pair<std::string, std::string>(reader.getName(), reader["value"].getValueAsString()));
    }

    for (auto it = binaryInputStates.begin(); it != binaryInputStates.end(); ++it ) {
      Logger::getInstance()->log("VdcHelper::getState: device " + dsuid2str(_device) +
          ": binary input: " + it->first + "=" + intToString(it->second), lsDebug);
    }
    for (auto it = deviceStates.begin(); it != deviceStates.end(); ++it ) {
      Logger::getInstance()->log("VdcHelper::getState: device " + dsuid2str(_device) +
          ": state: " + it->first + "=" + it->second, lsDebug);
    }
    return state;
  }

  std::map<std::string,VdcHelper::ChannelDesc> VdcHelper::getChannelDesc(dsuid_t _vdsm, dsuid_t _device)
  {
    vdcapi::Message message = VdcConnection::getProperty(_vdsm, _device, {"channelDescriptions"});

    Logger::getInstance()->log("VdcHelper::getChannelDesc: message " + message.DebugString(), lsDebug);
    vdcapi::vdc_ResponseGetProperty response = message.vdc_response_get_property();
    VdcElementReader reader(message.vdc_response_get_property().properties());

    std::map<std::string,ChannelDesc> channelList;
    VdcElementReader channelDescReader = reader["channelDescriptions"];
    for (VdcElementReader::iterator it = channelDescReader.begin(); it != channelDescReader.end(); it++) {
      ChannelDesc sDesc;
      VdcElementReader channelReader = *it;
      std::string propertyName = channelReader.getName();
      for (VdcElementReader::iterator it = channelReader.begin(); it != channelReader.end(); it++) {
        VdcElementReader fieldProp = *it;
        const std::string& fieldName = fieldProp.getName();
        if (fieldName == "name") {
          sDesc.name = fieldProp.getValueAsString();
        } else if (fieldName == "channelType") {
          sDesc.type = fieldProp.getValueAsInt();
        } else if (fieldName == "dsIndex") {
          // new in vdc-api v3
          sDesc.index = fieldProp.getValueAsInt();
        } else if (fieldName == "channelIndex") {
          // old in vdc-api v2
          sDesc.index = fieldProp.getValueAsInt();
        }
      }
      if (sDesc.type == 255) {
        try {
          sDesc.type = strToInt(propertyName);
        } catch(std::exception&) {}
      }
      if (sDesc.index == 255) {
        sDesc.index = channelList.size();
      }

      Logger::getInstance()->log("VdcHelper::getChannelDesc: device " + dsuid2str(_device) +
          ": id " + propertyName +
          ", index " + intToString(sDesc.index) +
          ", name " + sDesc.name +
          ", type " + intToString(sDesc.type), lsDebug);

      channelList[propertyName] = sDesc;
    }
    return channelList;
  }

  VdcHelper::ClimateDesc VdcHelper::getClimateSettings(dsuid_t _vdsm, dsuid_t _device)
  {
    ClimateDesc desc;
    vdcapi::Message message = VdcConnection::getProperty(_vdsm, _device, {"outputDescription", "outputSettings"});

    Logger::getInstance()->log("VdcHelper::getCoolingMode: message " + message.DebugString(), lsDebug);
    vdcapi::vdc_ResponseGetProperty response = message.vdc_response_get_property();
    VdcElementReader reader(message.vdc_response_get_property().properties());

    VdcElementReader outputDescReader = reader["outputDescription"];
    for (VdcElementReader::iterator it = outputDescReader.begin(); it != outputDescReader.end(); it++) {
      VdcElementReader fieldProp = *it;
      const std::string& fieldName = fieldProp.getName();
      if (fieldName == "activeCoolingMode") {
        desc.activeCoolingMode = fieldProp.getValueAsBool();
      }
    }

    VdcElementReader outputSettingsReader = reader["outputSettings"];
    for (VdcElementReader::iterator it = outputSettingsReader.begin(); it != outputSettingsReader.end(); it++) {
      VdcElementReader fieldProp = *it;
      const std::string& fieldName = fieldProp.getName();
      if (fieldName == "heatingSystemType") {
        desc.heatingSystemType = fieldProp.getValueAsInt();
      }
    }

    return desc;
  }

  vdcapi::Message VdcHelper::callLearningFunction(dsuid_t vdc, bool establish, int64_t timeout, const vdcapi::PropertyElement& params)
  {
    DeviceBusInterface* deviceBusInterface = DSS::getInstance()->getApartment().getDeviceBusInterface();
    if (!deviceBusInterface) {
      throw std::runtime_error("Bus interface not available");
    }
    google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> learningCall;
    vdcapi::PropertyElement* param0 = learningCall.Add();
    param0->set_name("establish");
    param0->mutable_value()->set_v_bool(establish);
    if (timeout >= 0) {
      vdcapi::PropertyElement* param1 = learningCall.Add();
      param1->set_name("timeout");
      param1->mutable_value()->set_v_int64(timeout);
    }
    if (params.elements_size()) {
      vdcapi::PropertyElement* paramx = learningCall.Add();
      paramx->set_name("params");
      *paramx->mutable_elements() = params.elements();
    }
    return VdcConnection::genericRequest(vdc, vdc, "pair", learningCall);
  }

  vdcapi::Message VdcHelper::callFirmwareFunction(dsuid_t vdc, bool checkOnly, bool clearSettings, const vdcapi::PropertyElement& params)
  {
    DeviceBusInterface* deviceBusInterface = DSS::getInstance()->getApartment().getDeviceBusInterface();
    if (!deviceBusInterface) {
      throw std::runtime_error("Bus interface not available");
    }
    google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> firmwareCall;
    vdcapi::PropertyElement* param0 = firmwareCall.Add();
    param0->set_name("checkonly");
    param0->mutable_value()->set_v_bool(checkOnly);
    vdcapi::PropertyElement* param1 = firmwareCall.Add();
    param1->set_name("clearsettings");
    param1->mutable_value()->set_v_bool(clearSettings);
    if (params.elements_size()) {
      vdcapi::PropertyElement* paramx = firmwareCall.Add();
      paramx->set_name("params");
      *paramx->mutable_elements() = params.elements();
    }
    return VdcConnection::genericRequest(vdc, vdc, "firmwareUpdate", firmwareCall);
  }

  vdcapi::Message VdcHelper::callAuthenticate(dsuid_t vdc, const std::string& authData, const std::string& authScope, const vdcapi::PropertyElement& params)
  {
    DeviceBusInterface* deviceBusInterface = DSS::getInstance()->getApartment().getDeviceBusInterface();
    if (!deviceBusInterface) {
      throw std::runtime_error("Bus interface not available");
    }
    google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> authenticateCall;
    vdcapi::PropertyElement* param0 = authenticateCall.Add();
    param0->set_name("authData");
    param0->mutable_value()->set_v_string(authData);
    vdcapi::PropertyElement* param1 = authenticateCall.Add();
    param1->set_name("authScope");
    param1->mutable_value()->set_v_string(authScope);
    if (params.elements_size()) {
      vdcapi::PropertyElement* paramx = authenticateCall.Add();
      paramx->set_name("params");
      *paramx->mutable_elements() = params.elements();
    }
    return VdcConnection::genericRequest(vdc, vdc, "authenticate", authenticateCall);
  }

  vdcapi::Message VdcHelper::callIdentify(dsuid_t vdc)
  {
    DeviceBusInterface* deviceBusInterface = DSS::getInstance()->getApartment().getDeviceBusInterface();
    if (!deviceBusInterface) {
      throw std::runtime_error("Bus interface not available");
    }
    google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> authenticateCall;

    return VdcConnection::genericRequest(vdc, vdc, "identify", authenticateCall);
  }

  vdcapi::Message VdcConnection::genericRequest(const dsuid_t& vdcId, const dsuid_t& targetId,
    const std::string& methodName,
    const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement >& params)
  {
    char dsuid_string[DSUID_STR_LEN];
    dsuid_to_string(&targetId, dsuid_string);

    vdcapi::Message message;
    message.set_type(vdcapi::VDSM_REQUEST_GENERIC_REQUEST);
    auto genericRequest = message.mutable_vdsm_request_generic_request();
    genericRequest->set_dsuid(dsuid_string);
    genericRequest->set_methodname(methodName);
    *genericRequest->mutable_params() = params;
    uint8_t arrayOut[REQUEST_LEN];
    if (!message.SerializeToArray(arrayOut, sizeof(arrayOut))) {
      throw std::runtime_error("SerializeToArray failed");
    }

    uint8_t arrayIn[RESPONSE_LEN];
    uint16_t arrayInSize;
    DSS::getInstance()->getApartment().getBusInterface()->getStructureQueryBusInterface()->protobufMessageRequest(
        vdcId, message.ByteSize(), arrayOut, &arrayInSize, arrayIn);

    if (!message.ParseFromArray(arrayIn, arrayInSize)) {
      throw std::runtime_error("ParseFromArray failed");
    }
    if (message.type() != vdcapi::GENERIC_RESPONSE) {
      throw std::runtime_error("Invalid vdc response");
    }
    const vdcapi::GenericResponse& response = message.generic_response();
    if (response.code() != vdcapi::ERR_OK) {
      throw std::runtime_error(std::string("Vdc error code:") + intToString(response.code())
          + " message:" + response.description());
    }
    return message;
  }

  vdcapi::Message VdcConnection::setProperty(const dsuid_t& vdcId, const dsuid_t& targetId,
      const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement >& properties)
  {
    char dsuid_string[DSUID_STR_LEN];
    dsuid_to_string(&targetId, dsuid_string);

    vdcapi::Message message;
    message.set_type(vdcapi::VDSM_REQUEST_SET_PROPERTY);
    vdcapi::vdsm_RequestSetProperty* setPropertyRquest = message.mutable_vdsm_request_set_property();
    setPropertyRquest->set_dsuid(dsuid_string);
    *setPropertyRquest->mutable_properties() = properties;
    uint8_t arrayOut[REQUEST_LEN];
    if (!message.SerializeToArray(arrayOut, sizeof(arrayOut))) {
      throw std::runtime_error("SerializeToArray failed");
    }

    uint8_t arrayIn[RESPONSE_LEN];
    uint16_t arrayInSize;
    DSS::getInstance()->getApartment().getBusInterface()->getStructureQueryBusInterface()->protobufMessageRequest(
        vdcId, message.ByteSize(), arrayOut, &arrayInSize, arrayIn);

    if (!message.ParseFromArray(arrayIn, arrayInSize)) {
      throw std::runtime_error("ParseFromArray failed");
    }
    if (message.type() != vdcapi::GENERIC_RESPONSE) {
      throw std::runtime_error("Invalid vdc response");
    }
    const vdcapi::GenericResponse& response = message.generic_response();
    if (response.code() != vdcapi::ERR_OK) {
      throw std::runtime_error(std::string("Vdc error code:") + intToString(response.code())
          + " message:" + response.description());
    }

    return message;
  }

  vdcapi::Message VdcConnection::getProperty(
      const dsuid_t& vdcId, const dsuid_t& targetId, const std::vector<std::string>& keys) {
    return DSS::getInstance()->getApartment().getBusInterface()->getStructureQueryBusInterface()->getVdcProperty(
        vdcId, targetId, keys);
  }

}// namespace
