/*
    Copyright (c) 2014 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef __DSS_VDC_CONNECTION_H__
#define __DSS_VDC_CONNECTION_H__

#include "config.h"

#include <string>
#include <map>
#include <vector>
#include <boost/shared_ptr.hpp>
#include <digitalSTROM/dsuid.h>

#include "messages/vdc-messages.pb.h"
#include "model/modelconst.h"
#include "vdc.h"
#include "ds/json.h"

  namespace dss {
    typedef struct {
      std::string name;
      int dSIndex;
      int inputType;
      int inputUsage;
      int sensorFunction;
      float updateInterval;
    } DeviceBinaryInputDescriptions_t;
    } // namespace dss
    DS_REFLECT(dss::DeviceBinaryInputDescriptions_t, (name)(dSIndex)(inputType)(inputUsage)(sensorFunction)(updateInterval));

  namespace dss {
    typedef struct {
      int group;
      int sensorFunction;
    } DeviceBinaryInputSettings_t;
    } // namespace dss
    DS_REFLECT(dss::DeviceBinaryInputSettings_t, (group)(sensorFunction));

  namespace dss {
    typedef struct {
      bool value;
      int error;
    } DeviceBinaryInputStates_t;
    } // namespace dss
    DS_REFLECT(dss::DeviceBinaryInputStates_t, (value)(error));

  namespace dss {

    struct VdsdSpec_t {
      // dS-Article Data
      std::string oemGuid;          // instance id => e.g. gs1:sgtin
      std::string oemModelGuid;     // class id => GTIN of the dS-Article
      std::string vendorGuid;
      std::string vendorId;
      // Hardware Endpoint description
      std::string hardwareGuid;
      std::string hardwareModelGuid;
      // System ID's
      std::string modelUID;         // device configuration id
      std::string descriptionsGroup;
      std::string descriptionsClass;
      // End-User Info
      std::string name;
      std::string model;
      std::string displayId;        // unique id for this device, e.g. s/n, mac address
      std::string hardwareVersion;
      std::string modelVersion;
      std::string vendorName;
      // Info read from backend VDC
      std::string dSUID;
      int primaryGroup;
      std::vector<std::pair<std::string,DeviceBinaryInputDescriptions_t>> binaryInputDescriptions;
      std::vector<std::pair<std::string,DeviceBinaryInputSettings_t>> binaryInputSettings;
      std::vector<std::pair<std::string,DeviceBinaryInputStates_t>> binaryInputStates;

      // Configurator integration
      std::string configURL;

      // backendvdc specific
      bool active;
      std::set<ModelFeatureId> modelFeatures;
    };
    } // namespace dss
    DS_REFLECT(dss::VdsdSpec_t, (oemGuid)(oemModelGuid)(vendorGuid)(vendorId)(hardwareGuid)(hardwareModelGuid)(modelUID)(
                                    descriptionsGroup)(descriptionsClass)(name)(model)(displayId)(hardwareVersion)(modelVersion)(vendorName)(
                                    dSUID)(primaryGroup)(binaryInputDescriptions)(binaryInputSettings)(binaryInputStates)(configURL)(modelFeatures)(active));

  namespace dss {

  struct VdcHelper
  {
    static VdsdSpec_t getSpec(dsuid_t _vdsm, dsuid_t _device);
    static boost::shared_ptr<VdcSpec> getVdcSpec(dsuid_t _vdsm);
    static void getIcon(dsuid_t _vdsm, dsuid_t _device, size_t *size, uint8_t **data);

    struct State {
      std::vector<std::pair<std::string, int> > binaryInputStates;
      std::vector<std::pair<std::string, std::string> > deviceStates;
    };
    static State getState(dsuid_t _vdsm, dsuid_t _device);

    struct ChannelDesc {
      std::string name;
      uint8_t index;
      uint8_t type;
      ChannelDesc() : index(255), type(255) {}
    };
    static std::map<std::string,ChannelDesc> getChannelDesc(dsuid_t _vdsm, dsuid_t _device);

    struct ClimateDesc {
      int heatingSystemType;
      bool activeCoolingMode;
      ClimateDesc() : heatingSystemType(0), activeCoolingMode(false) {}
    };
    static ClimateDesc getClimateSettings(dsuid_t _vdsm, dsuid_t _device);

    static vdcapi::Message callLearningFunction(dsuid_t vdc, bool establish, int64_t timeout, const vdcapi::PropertyElement& params);
    static vdcapi::Message callFirmwareFunction(dsuid_t vdc, bool checkOnly, bool clearSettings, const vdcapi::PropertyElement& params);
    static vdcapi::Message callAuthenticate(dsuid_t vdc, const std::string& authData, const std::string& authScope, const vdcapi::PropertyElement& params);
    static vdcapi::Message callIdentify(dsuid_t vdc);

  };

  struct VdcConnection
  {
    static vdcapi::Message genericRequest(const dsuid_t& vdcId, const dsuid_t& targetId,
        const std::string& methodName,
        const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement >& params);

    static vdcapi::Message setProperty(const dsuid_t& vdcId, const dsuid_t& targetId,
        const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement >& properties);

    // Handy method to call getProperty that builds the query from list of keys
    static vdcapi::Message getProperty(
        const dsuid_t& vdcId, const dsuid_t& targetId, const std::vector<std::string>& keys);
  };

} // namespace
#endif//__DSS_VDC_CONNECTION_H__
