/*
 *  Copyright (c) 2015 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include "vdc-db.h"

#include <ds/log.h>
#include <boost/algorithm/string/predicate.hpp>
#include <fstream>

#include "ds/log.h"
#include "base.h"
#include "dss.h"
#include "foreach.h"
#include "logger.h"
#include "propertysystem.h"
#include "sqlite3_wrapper.h"

static ds::LogChannel dsLogChannel("vdcDb");

namespace dss {

VdcDb::~VdcDb(){};

namespace {

class VdcDbImpl : public VdcDb {
public:
  VdcDbImpl(DSS &dss, SQLite3::Mode mode = SQLite3::Mode::ReadOnly);

  SQLite3 &getDb() { return m_db; }
  std::vector<SpecDesc> getSpec(const std::string &gtin) DS_OVERRIDE;
  std::vector<DeviceStateSpec_t> getStatesLegacy(const std::string &gtin) DS_OVERRIDE;
  std::vector<StateDesc> getStates(const std::string &gtin) DS_OVERRIDE;
  std::vector<EventDesc> getEvents(const std::string &gtin) DS_OVERRIDE;
  std::vector<PropertyDesc> getProperties(const std::string &gtin) DS_OVERRIDE;
  std::vector<SensorDesc> getSensors(const std::string &gtin) DS_OVERRIDE;
  std::vector<ActionDesc> getActions(const std::string &gtin) DS_OVERRIDE;
  std::vector<StandardActionDesc> getStandardActions(const std::string &gtin) DS_OVERRIDE;
  Categories getCategories(const std::string &gtin) DS_OVERRIDE;
  Outputs getOutputs(const std::string &gtin) DS_OVERRIDE;
  bool hasActionInterface(const std::string &gtin) DS_OVERRIDE;

private:
  __DECL_LOG_CHANNEL__;
  SQLite3 m_db;
  static std::mutex s_mutex;
  std::lock_guard<std::mutex> m_lock;

  propertyTypeId idToPropertyType(const int typeId) const;
};

__DEFINE_LOG_CHANNEL__(VdcDbImpl, lsInfo);

// Parallel readers and writers would introduce errors returned by sqlite.
// It is possible to recover from these errors, but we don't aim for parallelism here.
// We avoid concurrency errors by locking this static mutex in each VdcDb instance.
std::mutex VdcDbImpl::s_mutex;

} // namespace

static std::string getFilePath(DSS &dss);

VdcDbImpl::VdcDbImpl(DSS &dss, SQLite3::Mode mode) : m_db(getFilePath(dss), mode), m_lock(s_mutex) {}

static std::string getFilePath(DSS &dss) { return dss.getDatabaseDirectory() + "/vdc.db"; }

namespace vdcDb {

std::unique_ptr<VdcDb> make(DSS &dss, SQLite3::Mode mode) { return std::unique_ptr<VdcDb>(new VdcDbImpl(dss, mode)); }

void recreate(DSS &dss) {
  try {
    // Check that VdcDb is usable, (re)create if not.
    //
    // TODO(someday): refactor this initial VdcDb recreating into separate class?
    // It is independent to the database fetcher and must be active even when fetcher is inactive.

    std::ifstream in(getFilePath(dss).c_str(), std::ifstream::ate | std::ifstream::binary);
    if (in.tellg() == 0) {
      throw std::runtime_error(
              "Error while accessing sqlite database file " +
              getFilePath(dss) + ": file is empty");
    }

    VdcDbImpl db(dss);
    DS_NOTICE("Database present");

    // check for broken database
    db.getDb().exec("pragma quick_check;");
  } catch (std::exception &e) {
    try {
      ::remove(getFilePath(dss).c_str());
      VdcDbImpl db(dss, SQLite3::Mode::ReadWrite);
      db.getDb().exec(readFile(dss.getDataDirectory() + "/vdc-db.sql"));
      DS_NOTICE("Database (re)created.", e);
    } catch (std::exception &e2) {
      // Opening db read-write leaves empty file.
      // Delete the file so that read only open fails.
      ::remove(getFilePath(dss).c_str());
      throw std::runtime_error(std::string() + "Database missing or corrupt, recreate failed. e2.what():"
          + e2.what() + " e.what():" + e.what());
    }
  }
}

void update(DSS &dss, const std::string &newDb) {
  std::string tmpDb = dss.getDatabaseDirectory() + "/vdc.db.tmp";
  SQLite3 db3(tmpDb, SQLite3::Mode::ReadWrite);
  db3.exec(newDb);

  int ret = ::rename(tmpDb.c_str(), getFilePath(dss).c_str());
  if (ret != 0) {
    throw std::runtime_error(std::string() + "Database update failed: " +
        std::string(strerror(errno)));
  }
  DS_NOTICE("Database updated.");
}

} // namespace vdcDb

VdcDb::propertyTypeId VdcDbImpl::idToPropertyType(const int typeId) const {
  switch (typeId) {
    case 1:
    case 2:
      return propertyTypeId::integer;
    case 3:
      return propertyTypeId::numeric;
    case 4:
    default:
      return propertyTypeId::string;
    case 5:
      return propertyTypeId::enumeration;
    case 6:
      return propertyTypeId::boolean;
  }
}

std::vector<DeviceStateSpec_t> VdcDbImpl::getStatesLegacy(const std::string &gtin) {
  std::vector<DeviceStateSpec_t> states;

  foreach (const StateDesc desc, getStates(gtin)) {
    states.push_back(DeviceStateSpec_t());
    states.back().Name = desc.id;
    states.back().Values = desc.values;
  }

  return states;
}

std::vector<VdcDb::SpecDesc> VdcDbImpl::getSpec(const std::string &gtin) {
  // name, tags, gtin
  std::string sql0("select * from callGetSpecBase where (gtin=? or gtin='')");
  SqlStatement query0 = m_db.prepare(sql0);
  SqlStatement::BindScope scope0 = query0.bind(gtin);

  std::vector<SpecDesc> specs;
  while (query0.step() != SqlStatement::StepResult::DONE) {
    specs.push_back(SpecDesc());
    specs.back().id = query0.getColumn<std::string>(0);
    specs.back().tags = query0.getColumn<std::string>(1);
    specs.back().value = query0.getColumn<std::string>(3);
  }

  return specs;
}

std::vector<VdcDb::StateDesc> VdcDbImpl::getStates(const std::string &gtin) {
  // name, value, tags, gtin
  std::string sql0("select * from callGetStatesBase where gtin=?");
  SqlStatement query0 = m_db.prepare(sql0);
  SqlStatement::BindScope scope0 = query0.bind(gtin);

  std::vector<StateDesc> states;
  std::string cur(""); // invalid state name
  while (query0.step() == SqlStatement::StepResult::ROW) {
    std::string id = query0.getColumn<std::string>(0);
    if (states.empty() || cur != id) {
      states.push_back(StateDesc());
      states.back().id = cur = id;
      states.back().tags = query0.getColumn<std::string>(2);
    }
    states.back().values.push_back(query0.getColumn<std::string>(1));
  }

  return states;
}

std::vector<VdcDb::EventDesc> VdcDbImpl::getEvents(const std::string &gtin) {
  // name, gtin
  std::string sql0("select * from callGetEventsBase where gtin=?");
  SqlStatement query0 = m_db.prepare(sql0);
  SqlStatement::BindScope scope0 = query0.bind(gtin);

  std::vector<EventDesc> events;
  while (query0.step() != SqlStatement::StepResult::DONE) {
    events.push_back(EventDesc());
    events.back().id = query0.getColumn<std::string>(0);
    events.back().categoryId = query0.getColumn<std::string>(2);
    events.back().accessLevel = static_cast<VdcDb::AccessLevel>(query0.getColumn<int>(3));
  }

  return events;
}

std::vector<VdcDb::PropertyDesc> VdcDbImpl::getProperties(const std::string &gtin) {
  // name, base_type, default_value, min_value, max_value, resolution, si_unit, tags, gtin, type_name, prop_id, enum_reference
  std::string sql0("select * from callGetPropertiesBase where gtin=?");
  SqlStatement query0 = m_db.prepare(sql0);
  SqlStatement::BindScope scope0 = query0.bind(gtin);

  std::vector<PropertyDesc> props;
  while (query0.step() != SqlStatement::StepResult::DONE) {
    props.push_back(PropertyDesc());
    props.back().id = query0.getColumn<std::string>(0);
    props.back().typeId = idToPropertyType(query0.getColumn<int>(1));
    props.back().defaultValue = query0.getColumn<std::string>(2);
    props.back().minValue = query0.getColumn<std::string>(3);
    props.back().maxValue = query0.getColumn<std::string>(4);
    props.back().resolution = query0.getColumn<std::string>(5);
    props.back().siUnit = query0.getColumn<std::string>(6);
    props.back().tags = query0.getColumn<std::string>(7);
    // TODO(someday) retrieve from db
    props.back().readonly = false;
    if (sqlite3_column_type(query0, 9) != SQLITE_NULL) {
      props.back().typePostfix = query0.getColumn<std::string>(9);
    }
    if (sqlite3_column_type(query0, 11) != SQLITE_NULL) {
      int typeId = query0.getColumn<int>(11);
      if ((props.back().typeId == propertyTypeId::enumeration) && (typeId != 0)) {
        std::string sql1 =
            "select value from device_actions_parameter_type_enum "
            "where device_action_parameter_type_id=?";
        SqlStatement query1 = m_db.prepare(sql1);
        SqlStatement::BindScope scope1 = query1.bind(typeId);
        while (query1.step() != SqlStatement::StepResult::DONE) {
          props.back().values.push_back(query1.getColumn<std::string>(0));
        }
      }
    }
  }

  return props;
}

std::vector<VdcDb::SensorDesc> VdcDbImpl::getSensors(const std::string &gtin) {
  std::vector<SensorDesc> sensors;

  try {
    // 0 name, 1 base_type, 2 default_value, 3 min_value, 4 max_value, 5 resolution, 6 si_unit, 7 tags, 8 gtin, 9 type_name, 10 sensorIndex
    std::string sql0("select * from callGetSensorsBase where gtin=?");
    SqlStatement query0 = m_db.prepare(sql0);
    SqlStatement::BindScope scope0 = query0.bind(gtin);

    while (query0.step() != SqlStatement::StepResult::DONE) {
      sensors.push_back(SensorDesc());
      sensors.back().prop.id = query0.getColumn<std::string>(0);
      sensors.back().prop.typeId = idToPropertyType(query0.getColumn<int>(1));
      sensors.back().prop.defaultValue = query0.getColumn<std::string>(2);
      sensors.back().prop.minValue = query0.getColumn<std::string>(3);
      sensors.back().prop.maxValue = query0.getColumn<std::string>(4);
      sensors.back().prop.resolution = query0.getColumn<std::string>(5);
      sensors.back().prop.siUnit = query0.getColumn<std::string>(6);
      sensors.back().prop.tags = query0.getColumn<std::string>(7);
      // 8 gtin
      sensors.back().prop.typePostfix = query0.getColumn<std::string>(9);
      sensors.back().sensorIndex = query0.getColumn<std::string>(10);
    }
  } catch (std::runtime_error& e) {
    Logger::getInstance()->log(std::string(__func__) + " database error: " + e.what(), lsError);
    return std::vector<SensorDesc> ();
  }

  return sensors;
}

std::vector<VdcDb::ActionDesc> VdcDbImpl::getActions(const std::string &gtin) {
  // command, parameterName, type_id, default_value, min_value, max_value, resolution, si_unit, tags, gtin, name, enum_reference, optional, categories_id, basetags
  std::string sql0("select * from callGetActionsBase where gtin=?");
  SqlStatement query0 = m_db.prepare(sql0);
  SqlStatement::BindScope scope0 = query0.bind(gtin);

  std::vector<ActionDesc> actions;
  std::string cur;
  while (query0.step() != SqlStatement::StepResult::DONE) {
    std::string name = query0.getColumn<std::string>(0);
    if (actions.empty() || cur != name) {
      actions.push_back(ActionDesc());
      actions.back().id = cur = name;
      if (sqlite3_column_type(query0, 13) != SQLITE_NULL) {
        actions.back().categoryId = query0.getColumn<std::string>(13);
      }
      if (sqlite3_column_type(query0, 14) != SQLITE_NULL) {
        actions.back().tags = query0.getColumn<std::string>(14);
      }
    }
    if (sqlite3_column_type(query0, 1) != SQLITE_NULL) {
      actions.back().params.push_back(PropertyDesc());
      actions.back().params.back().id = query0.getColumn<std::string>(1);
      actions.back().params.back().typeId = idToPropertyType(query0.getColumn<int>(2));
      actions.back().params.back().defaultValue = query0.getColumn<std::string>(3);
      actions.back().params.back().minValue = query0.getColumn<std::string>(4);
      actions.back().params.back().maxValue = query0.getColumn<std::string>(5);
      actions.back().params.back().resolution = query0.getColumn<std::string>(6);
      actions.back().params.back().siUnit = query0.getColumn<std::string>(7);
      actions.back().params.back().tags = query0.getColumn<std::string>(8);
      actions.back().params.back().optional = (sqlite3_column_type(query0, 12) != SQLITE_NULL)
                                              ? static_cast<bool>(query0.getColumn<int>(12)) : false;
      if (sqlite3_column_type(query0, 10) != SQLITE_NULL) {
        actions.back().params.back().typePostfix = query0.getColumn<std::string>(10);
      }
      if (sqlite3_column_type(query0, 11) != SQLITE_NULL) {
        int typeId = query0.getColumn<int>(11);
        if ((actions.back().params.back().typeId == propertyTypeId::enumeration) && (typeId != 0)) {
          std::string sql1 =
              "select value from device_actions_parameter_type_enum "
              "where device_action_parameter_type_id=?";
          SqlStatement query1 = m_db.prepare(sql1);
          SqlStatement::BindScope scope1 = query1.bind(typeId);
          while (query1.step() != SqlStatement::StepResult::DONE) {
            actions.back().params.back().values.push_back(query1.getColumn<std::string>(0));
          }
        }
      }
    }
  }

  return actions;
}

std::vector<VdcDb::StandardActionDesc> VdcDbImpl::getStandardActions(const std::string &gtin) {
  // predefName, command, name:1, value, gint, tags
  std::string sql0("select * from callGetStandardActionsBase where gtin=?");
  SqlStatement query0 = m_db.prepare(sql0);
  SqlStatement::BindScope scope0 = query0.bind(gtin);

  std::vector<StandardActionDesc> desc;
  std::string cur("");
  while (query0.step() != SqlStatement::StepResult::DONE) {
    std::string name = query0.getColumn<std::string>(0);
    if (desc.empty() || cur != name) {
      desc.push_back(StandardActionDesc());
      desc.back().id = cur = name;
      desc.back().action = query0.getColumn<std::string>(1);
      if (sqlite3_column_type(query0, 5) != SQLITE_NULL) {
        desc.back().tags = query0.getColumn<std::string>(5);
      }
    }
    if (sqlite3_column_type(query0, 2) != SQLITE_NULL) {
      auto arg = std::make_pair(query0.getColumn<std::string>(2), query0.getColumn<std::string>(3));
      desc.back().params.push_back(arg);
    }
  }

  return desc;
}

VdcDb::Categories VdcDbImpl::getCategories(const std::string &gtin) {
  // name|category_order|gtin|allways_visible|compact

  const std::string sql0{"select * from callGetCategoriesBase where gtin=?"};
  SqlStatement query0 = m_db.prepare(sql0);
  SqlStatement::BindScope scope0 = query0.bind(gtin);

  Categories categories;
  while (query0.step() != SqlStatement::StepResult::DONE) {
    categories.push_back(Category());
    categories.back().id = query0.getColumn<std::string>(0);
    categories.back().orderId = query0.getColumn<std::string>(1);
    categories.back().alwaysVisible = (sqlite3_column_type(query0, 3) != SQLITE_NULL)
                                      ? static_cast<bool>(query0.getColumn<int>(3)) : true;
  }

  return categories;
}

VdcDb::Outputs VdcDbImpl::getOutputs(const std::string& gtin) {
  const std::string sql0{"select * from callGetOutputsBase where gtin=?"};
  SqlStatement query0 = m_db.prepare(sql0);
  SqlStatement::BindScope scope0 = query0.bind(gtin);

  Outputs outputs;
  while (query0.step() != SqlStatement::StepResult::DONE) {
    outputs.push_back(OutputDesc());
    outputs.back().id = query0.getColumn<std::string>(0);
    outputs.back().typeId =  idToPropertyType(query0.getColumn<int>(1));
    outputs.back().defaultValue = query0.getColumn<std::string>(2);
    outputs.back().minValue = query0.getColumn<std::string>(3);
    outputs.back().maxValue = query0.getColumn<std::string>(4);
    outputs.back().resolution = query0.getColumn<std::string>(5);
    outputs.back().siUnit = query0.getColumn<std::string>(6);
    outputs.back().tags = query0.getColumn<std::string>(7);
    // 8 gtin
    outputs.back().typePostfix = query0.getColumn<std::string>(9);
    // 10 prop_id
    // 11 enum_reference
    outputs.back().outputType = query0.getColumn<std::string>(12);
    outputs.back().outputIndex = query0.getColumn<std::string>(13);
    outputs.back().categoryId = query0.getColumn<std::string>(14);
  }

  return outputs;
}



bool VdcDbImpl::hasActionInterface(const std::string &gtin) {
  Logger::getInstance()->log(std::string(__func__) + "check action interface for GTIN " + gtin);

  bool result = false;
  std::string sql = "SELECT name FROM device WHERE gtin=?";
  SqlStatement find = m_db.prepare(sql);
  SqlStatement::BindScope scope = find.bind(gtin);

  while (find.step() == SqlStatement::StepResult::ROW) {
    DS_REQUIRE(!gtin.empty(), "Empty GTIN may not produce a result");
    std::string name = find.getColumn<std::string>(0);
    result = true;

    Logger::getInstance()->log(std::string(__func__) + " found name for GTIN \"" + gtin + "\": " + name);
  }

  return result;
}

} // dss namespace
