/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef WEBREQUESTS_H
#define WEBREQUESTS_H

#include "web/restful.h"

#include <ds/exception.h>
#include <ds/json.h>
#include <boost/noncopyable.hpp>
#include <boost/shared_ptr.hpp>
#include <map>
#include <string>

#include "includerapidjson.h"

#include "logger.h"
#include "ds485types.h"
#include "dss.h"
#include "model/modelconst.h"
#include "foreach.h"

using rapidjson::StringBuffer;
using rapidjson::Writer;

struct mg_connection;

namespace dss {

  class RestfulRequest;
  class Session;

  class WebServerResponse {
  public:
    WebServerResponse(std::string _response)
    : m_response(_response), m_revokeCookie(false)
    { }
    std::string getResponse() const {
      return m_response;
    }
    void setRevokeSessionToken() {
      m_newSessionToken.clear();
      m_revokeCookie = true;
    }
    bool isRevokeSessionToken() const {
      return m_revokeCookie;
    }
    void setPublishSessionToken(const std::string &token) {
      m_newSessionToken = token;
    }
    bool isPublishSessionToken() const {
      return !m_newSessionToken.empty();
    }
    const std::string &getNewSessionToken() const {
      return m_newSessionToken;
    }
  private:
    std::string m_response;
    std::string m_newSessionToken;
    bool m_revokeCookie;
  };

  class JSONWriter {
  public:
    typedef enum {
      jsonObjectResult,
      jsonArrayResult,
      jsonNoneResult
    } jsonResult_t;
    JSONWriter(jsonResult_t _responseType = jsonObjectResult);
    std::string successJSON();

    // add field name with field value
    template <typename T>
    void add(const std::string& name, T&& x) {
      add(name);
      add(std::forward<T>(x));
    }

    void add(const std::string &value);
    void add(const char* _value);
    void add(int _value);
    void add(unsigned _value);
    void add(long long int _value);
    void add(unsigned long long int _value);
    void add(bool _value);
    void add(double _value);
    void add(const dsuid_t &dsuid) { add(dsuid2str(dsuid)); }

    void add(BinaryInputType x) { add(static_cast<int>(x)); }
    void add(BinaryInputId x) { add(static_cast<int>(x)); }
    void add(ButtonInputMode x) { add(static_cast<int>(x)); }
    void add(SensorType x) { add(static_cast<int>(x)); }
    void add(BasicScene x) { add(static_cast<int>(x)); }

    // serialize reflected enums as strings by default
    template <typename T, typename = decltype(dsEnum(T()))>
    void add(T x) {
      add(ds::fromEnum<std::string>(x));
    }

    template <class T>
    void addReflected(const T& x) {
      auto s = ds::toJson(x);
      m_writer.RawValue(s.data(), s.size(), rapidjson::kNullType);
    }

    template <typename T>
    void addReflected(const std::string& name, T&& x) {
      add(name);
      addReflected(std::forward<T>(x));
    }

    template <typename T>
    void add(const std::vector<T>& items) {
      startArray();
      foreach(auto&& item, items) {
        add(item);
      }
      endArray();
    }

    // Add object key. Return false if the object entry cannot be serialized and is skipped.
    bool addKey(const char* x) { add(x); return true; }
    bool addKey(const std::string& x) { add(x); return true; }
    bool addKey(ModelFeatureId x);

    template <class Key, class T, class Compare, class Alloc>
    void add(const std::map<T, Key, Compare, Alloc>& map) {
      startObject();
      foreach(auto&& entry, map) {
        if (addKey(entry.first)) {
          add(entry.second);
        }
      }
      endObject();
    }

    void addNull();
    void startArray(std::string _name);
    void startArray();
    void endArray();
    void startObject(std::string _name);
    void startObject();
    void endObject();

    struct DeferredArrayEndDeleter {
      void operator()(JSONWriter* x) { x->endArray(); }
    };
    typedef std::unique_ptr<JSONWriter, DeferredArrayEndDeleter> DeferredArrayEnd;
    DeferredArrayEnd startScopedArray(std::string name) DS_WARN_UNUSED_RESULT {
      startArray(name);
      return DeferredArrayEnd(this);
    }
    DeferredArrayEnd startScopedArray() DS_WARN_UNUSED_RESULT {
      startArray();
      return DeferredArrayEnd(this);
    }

    struct DeferredObjectEndDeleter {
      void operator()(JSONWriter* x) { x->endObject(); }
    };
    typedef std::unique_ptr<JSONWriter, DeferredObjectEndDeleter> DeferredObjectEnd;
    DeferredObjectEnd startScopedObject(std::string name) DS_WARN_UNUSED_RESULT {
      startObject(name);
      return DeferredObjectEnd(this);
    }
    DeferredObjectEnd startScopedObject() DS_WARN_UNUSED_RESULT {
      startObject();
      return DeferredObjectEnd(this);
    }

    std::string raw();
    static std::string success();
    static std::string success(std::string _message);
    static std::string failure(std::string _message);
    static std::string failure(const std::exception&);

  private:
    StringBuffer m_buffer;
    Writer<StringBuffer, rapidjson::UTF8<>, rapidjson::ASCII<> > m_writer;
    jsonResult_t m_resultType;
  };

  class WebServerRequestHandlerJSON {
  public:
    virtual WebServerResponse jsonHandleRequest(const RestfulRequest& _request,
                                                boost::shared_ptr<Session> _session,
                                                const struct mg_connection* _connection) = 0;
    virtual ~WebServerRequestHandlerJSON() {}
  }; // WebServerRequestHandlerJSON



}

#endif // WEBREQUESTS_H
