/*
    Copyright (c) 2017 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "jsonapiserver.h"

#include <boost/make_shared.hpp>

#include "businterface.h"
#include "event.h"
#include "web/websocket/message.h"
#include "web/websocket/observerbuilder.h"

namespace dss {

namespace websocket {

namespace {

static ds::LogChannel dsLogChannel("websocketJsonApiServer");

Message errorMessage(const std::string& id, const std::string& error) {
  Message msg;
  msg.type = MessageType::REPLY;
  msg.id = id;
  msg.error.message = error;
  return msg;
}

} /* namespace */

boost::shared_ptr<Connection> JsonApiServer::createConnectionData(mg_connection* con) const {
  return boost::make_shared<JsonApiWebSocketConnection>(con);
}

bool JsonApiServer::shouldSend(const Event& e, const Connection& c) const {
  try {
    const JsonApiWebSocketConnection& con = dynamic_cast<const JsonApiWebSocketConnection&>(c);
    return con.hasObserverFor(e);
  } catch (const std::bad_cast& ex) {
    DS_ERROR("Bad cast to JsonApiWebSocketConnection.");
  }
  return false;
}

void JsonApiServer::prepareAndSendMsgs(const Event& e, Connection& c) const {
  JsonApiWebSocketConnection& con = dynamic_cast<JsonApiWebSocketConnection&>(c);
  BOOST_FOREACH (const auto& id, con.observersLokingFor(e)) {
    try {
      auto jsonFragment = con.tryCreateJsonFragmentByObserver(id, e);
      if (!jsonFragment) {
        continue;
      }
      Message msg;
      msg.type = MessageType::REPLY;
      msg.id = id;
      msg.result = std::move(jsonFragment);
      con.addMsgToSend(ds::toJson(msg));
    } catch (const std::exception& e) {
      con.addMsgToSend(ds::toJson(errorMessage(id, e.what())));
    }
  }
}

int JsonApiServer::dataHandler(mg_connection* connection, int number, char* msg, size_t msgSize, void* cbdata) {
  (void)cbdata;
  enum Return { CLOSE_WEBSOCKET = 0, KEEP_OPEN_WEBSOCKET = 1 };
  if (number & MG_WEBSOCKET_OPCODE_CONNECTION_CLOSE) {
    DS_NOTICE("Closing connection on client request.");
    return CLOSE_WEBSOCKET;
  }
  if (number & MG_WEBSOCKET_OPCODE_TEXT) {
    handleRequest(std::string(msg, msg + msgSize), connection);
  }

  return KEEP_OPEN_WEBSOCKET;
}

void JsonApiServer::handleRequest(const std::string& msg, mg_connection* connection) {
  JsonApiWebSocketConnection* client =
      dynamic_cast<JsonApiWebSocketConnection*>(reinterpret_cast<Connection*>(mg_get_user_connection_data(connection)));

  auto msgReq = ds::fromJson<Message>(msg);
  switch (msgReq.type) {
    case MessageType::REQUEST:
      try {
        client->addObserver(msgReq.id, ObserverBuilder{}.build(msgReq.url, msgReq.method, msgReq.params));
      } catch (const std::exception& e) {
        const auto& msg = std::string{e.what()}.append("[").append(msgReq.url).append("]");
        client->addMsgToSend(ds::toJson(errorMessage(msgReq.id, msg)));
      }
      break;
    case MessageType::CANCEL: {
      client->removeObserver(msgReq.id);
      break;
    }
    case MessageType::EVENT: // no break
      break;
    case MessageType::REPLY: // no break
    default:
      DS_ERROR("Invalid msg type: ", msgReq.type);
  }
}

} // namespace websocket

} // namespace dss
