/*
    Copyright (c) 2017 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "simpleserver.h"

#include <memory>

#include <boost/make_shared.hpp>

#include "dss.h"
#include "security/security.h"
#include "session.h"
#include "sessionmanager.h"
#include "web/restful.h"
#include "web/webserver.h"

#include "event.h"
#include "model/modelmaintenance.h"

namespace dss {

namespace websocket {

namespace {

const ds::LogChannel dsLogChannel("websocketSimpleServer");

} // namespace

SimpleServer::~SimpleServer() {
  if (m_mgContext) {
    mg_set_websocket_handler(m_mgContext, m_url.c_str(), NULL, NULL, NULL, NULL, NULL);
  }
}

void SimpleServer::registerToCivetWebServer(mg_context* mgContext, int trustedPort) {
  if (m_mgContext != mgContext) {
    mg_set_websocket_handler(m_mgContext, m_url.c_str(), NULL, NULL, NULL, NULL, NULL);
  }
  m_trustedPort = trustedPort;
  m_mgContext = mgContext;
  mg_set_websocket_handler(
      m_mgContext, m_url.c_str(), &connectCallback, &readyCallback, &dataCallback, &closeCallback, this);
}

void SimpleServer::sendToWebSockets(const Event& e) {
  boost::mutex::scoped_lock lock(m_connectionsMtx);
  BOOST_FOREACH (auto& con, m_connections) {
    if (con->isReady() && shouldSend(e, *con)) {
      prepareAndSendMsgs(e, *con);
    }
  }
}

size_t SimpleServer::clientCount() const {
  const boost::mutex::scoped_lock lock(m_connectionsMtx);
  return m_connections.size();
}

boost::shared_ptr<Connection> SimpleServer::createConnectionData(mg_connection* con) const {
  return boost::make_shared<Connection>(con);
}

bool SimpleServer::shouldSend(const Event& e, const Connection& c) const {
  (void)c;

  if ((e.getName() == EventName::LogFileData) || (e.getName() == EventName::DebugMonitorUpdate) ||
      (e.getName() == EventName::Signal) || (e.getName() == EventName::WebSessionCleanup) ||
      (e.getName() == EventName::ExecutionDeniedDigestCheck) || (e.getName() == EventName::CheckSensorValues) ||
      (e.getName() == EventName::CheckHeatingGroups) || (e.getName() == EventName::UpdateAutoselect) ||
      (e.getName() == EventName::ReexportTimings) ||
      (e.getName() == "EventInterpreterPluginJavascript_cleanupScripts") || (e.getName() == EventName::SendMail) ||
      (e.getContext() == "identifyDevice")) {
    return false;
  }

  return true;
}

void SimpleServer::prepareAndSendMsgs(const Event& e, Connection& c) const {
  c.addMsgToSend(ModelMaintenance::WebSocketEvent::prepareJson(e));
}

int SimpleServer::connectHandler(const mg_connection* connection, void* cbdata) {
  (void)cbdata;

  struct mg_context* ctx = mg_get_context(connection);

  const struct mg_request_info* _info = mg_get_request_info(connection);

  if (!_info->local_uri || strlen(_info->remote_addr) < 1) {
    DS_DEBUG("Websocket connection refused: no origin information");
    return 1;
  }

  if (_info->local_uri != m_url) {
    // should actually never happen due to the specific local_uri based callback
    DS_DEBUG("Websocket connection refused: invalid local_uri path");
    return 1;
  }

  DSS::getInstance()->getSessionManager().getSecurity()->signOff();
  {
    std::string remote_s(_info->remote_addr);
    std::string local_uri_path(_info->local_uri);
    if (_info->query_string) {
      local_uri_path += "?" + std::string(_info->query_string);
    }
    DS_INFO("REST call from " + remote_s + ": " + local_uri_path);
  }

  RestfulRequest request("", _info->query_string ? _info->query_string : "");

  boost::shared_ptr<Session> session;
  std::string token = extractToken(mg_get_header(connection, "Cookie"));
  if (token.empty()) {
    // ... used when not going through lighttpd
    token = request.getParameter("token");
  }
  if (!token.empty()) {
    session = DSS::getInstance()->getSessionManager().getSession(token);
  }

  // lighttpd is proxying to our trusted port. We trust that it
  // properly authenticated the user and just filter the HTTP
  // Authorization header for the username
  if (((session == NULL) || (session->getUser() == NULL)) && (_info->local_port == m_trustedPort)) {
    const char* auth_header = mg_get_header(connection, "Authorization");
    std::string userName = extractAuthenticatedUser(auth_header);

    if (!userName.empty()) {
      DS_DEBUG("Logging-in from a trusted port as '" + userName + "'");
      if (!DSS::getInstance()->getSessionManager().getSecurity()->impersonate(userName)) {
        DS_INFO("Unknown user");
        return 1;
      }
      if (session == NULL) {
        std::string newToken = DSS::getInstance()->getSessionManager().registerSession();
        if (newToken.empty()) {
          DS_ERROR("Session limit reached: " + request.getUrlPath());
          return 1;
        }
        DS_DEBUG("Registered new JSON session for trusted port (" + newToken + ")");
        session = DSS::getInstance()->getSessionManager().getSession(newToken);
        session->inheritUserFromSecurity();
      }
    }
  }

  if (session != NULL) {
    session->touch();
    DSS::getInstance()->getSessionManager().getSecurity()->authenticate(session);
  } else {
    DS_DEBUG("Authentication failed, rejecting websocket connection");
    return 1;
  }

  boost::mutex::scoped_lock lock(m_connectionsMtx);
  if (m_connections.size() >= m_maxClientCount) {
    DS_DEBUG("Rejecting new websocket connection: maximum number of connections exceeded");
    return 1; // reject connection
  }

  auto&& ws = createConnectionData(const_cast<mg_connection*>(connection));

  mg_lock_context(ctx);

  ws->setState(State::ws_connected);
  mg_set_user_connection_data(const_cast<mg_connection*>(connection), (void*)ws.get());
  m_connections.push_back(ws);
  ws->startMsgSendingThread();
  mg_unlock_context(ctx);

  return 0;
}

void SimpleServer::readyHandler(mg_connection* connection, void* cbdata) {
  (void)cbdata;

  const char* text = "{ \"ok\": true }";

  Connection* client = reinterpret_cast<Connection*>(mg_get_user_connection_data(connection));
  if (client == NULL) {
    return;
  }

  mg_websocket_write(connection, MG_WEBSOCKET_OPCODE_TEXT, text, strlen(text));
  client->setState(State::ws_ready);
}

int SimpleServer::dataHandler(mg_connection* connection, int number, char* msg, size_t msgSize, void* cbdata) {
  (void)connection;
  (void)number;
  (void)msg;
  (void)msgSize;
  (void)cbdata;
  return KEEP_OPEN_WEBSOCKET;
}

void SimpleServer::closeHandler(const mg_connection* connection, void* cbdata) {
  (void)cbdata;

  mg_context* ctx = mg_get_context(connection);
  Connection* client = reinterpret_cast<Connection*>(mg_get_user_connection_data(connection));

  const boost::mutex::scoped_lock lock(m_connectionsMtx);

  std::list<boost::shared_ptr<Connection> >::iterator i = m_connections.begin();
  while (i != m_connections.end()) {
    if ((*i).get() == client) {
      mg_lock_context(ctx);
      DS_DEFER(mg_unlock_context(ctx));
      mg_set_user_connection_data(const_cast<mg_connection*>(connection), NULL);
      (*i)->setState(State::ws_disconnected);
      m_connections.erase(i);
      break;
    }
    ++i;
  }
}

int SimpleServer::connectCallback(const mg_connection* _connection, void* cbdata) {
  try {
    SimpleServer* self = reinterpret_cast<SimpleServer*>(cbdata);
    return self->connectHandler(_connection, cbdata);
  } catch (const std::exception& e) {
    DS_ERROR(e);
  }
  return 1;
}

void SimpleServer::readyCallback(mg_connection* connection, void* cbdata) {
  try {
    SimpleServer* self = reinterpret_cast<SimpleServer*>(cbdata);
    return self->readyHandler(connection, cbdata);
  } catch (const std::exception& e) {
    DS_ERROR(e);
  }
}

int SimpleServer::dataCallback(mg_connection* connection, int number, char* msg, size_t msgSize, void* cbdata) {
  try {
    SimpleServer* self = reinterpret_cast<SimpleServer*>(cbdata);
    return self->dataHandler(connection, number, msg, msgSize, cbdata);
  } catch (const std::exception& e) {
    DS_ERROR(e);
  }
  return KEEP_OPEN_WEBSOCKET;
}

void SimpleServer::closeCallback(const mg_connection* _connection, void* cbdata) {
  try {
    SimpleServer* self = reinterpret_cast<SimpleServer*>(cbdata);
    self->closeHandler(_connection, cbdata);
  } catch (const std::exception& e) {
    DS_ERROR(e);
  }
}

} // namespace websocket

} // namespace dss
