/*
    Copyright (c) 2009,2011 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#define BOOST_TEST_NO_MAIN
#define BOOST_TEST_DYN_LINK
#include <boost/test/unit_test.hpp>

#include <boost/make_shared.hpp>

#include "model/apartment.h"
#include "model/zone.h"

using namespace dss;

BOOST_AUTO_TEST_SUITE(zone)

DSUID_DEFINE(dsuid1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1);
DSUID_DEFINE(dsuid2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2);
DSUID_DEFINE(dsuid3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3);

BOOST_AUTO_TEST_CASE(testSensorSetReset) {
  Zone zone(2, NULL);
  MainZoneSensor_t ms1 = { dsuid1, SensorType::CO2Concentration, 1 };
  MainZoneSensor_t ms3 = { dsuid3, SensorType::HumidityIndoors, 3 };

  zone.setSensor(ms1);
  zone.setSensor(ms3);

  BOOST_CHECK(zone.isSensorAssigned(ms1.m_sensorType));
  BOOST_CHECK(zone.isSensorAssigned(ms3.m_sensorType));
  zone.resetSensor(ms1.m_sensorType);
  BOOST_CHECK(!zone.isSensorAssigned(ms1.m_sensorType));
  BOOST_CHECK(zone.isSensorAssigned(ms3.m_sensorType));
}

BOOST_AUTO_TEST_CASE(testRemovInvalidSensors) {
  // test: remove sensors that don't have the hosting device in the zone

  Apartment apt(NULL);

  Zone zone(7, &apt);
  auto dev1 = DeviceReference(apt.allocateDevice(dsuid1), &apt);
  zone.addDevice(dev1);
  auto dev3 = DeviceReference(apt.allocateDevice(dsuid3), &apt);
  zone.addDevice(dev3);

  MainZoneSensor_t ms1 = { dsuid1, SensorType::CO2Concentration, 1 };
  MainZoneSensor_t ms2 = { dsuid2, SensorType::HumidityIndoors, 2 };
  MainZoneSensor_t ms3 = { dsuid3, SensorType::TemperatureIndoors, 3 };

  zone.setSensor(ms1);
  zone.setSensor(ms2);
  zone.setSensor(ms3);

  // device with dsuid2 is not in zone
  BOOST_CHECK_EQUAL(zone.getAssignedSensors().size(), 3);
  BOOST_CHECK(zone.isSensorAssigned(SensorType::HumidityIndoors));
  zone.removeInvalidZoneSensors();
  BOOST_CHECK(zone.isSensorAssigned(SensorType::CO2Concentration));
  BOOST_CHECK(zone.isSensorAssigned(SensorType::TemperatureIndoors));
  BOOST_CHECK(!zone.isSensorAssigned(SensorType::HumidityIndoors));
}

BOOST_AUTO_TEST_CASE(testAssignedUnassignedSensorTypes) {
  Apartment apt(NULL);

  Zone zone(7, &apt);
  BOOST_CHECK_GE(zone.getUnassignedSensorTypes().size(), 4);
  BOOST_CHECK_EQUAL(zone.getAssignedSensors().size(), 0);
  auto maxZoneSensors = zone.getUnassignedSensorTypes().size();

  auto dev1 = DeviceReference(apt.allocateDevice(dsuid1), &apt);
  zone.addDevice(dev1);
  auto dev3 = DeviceReference(apt.allocateDevice(dsuid3), &apt);
  zone.addDevice(dev3);

  MainZoneSensor_t ms1_1 = { dsuid1, SensorType::CO2Concentration, 1 };
  MainZoneSensor_t ms1_2 = { dsuid1, SensorType::HumidityIndoors, 2 };
  MainZoneSensor_t ms3 = { dsuid3, SensorType::TemperatureIndoors, 3 };
  zone.setSensor(ms1_1);
  zone.setSensor(ms1_2);
  zone.setSensor(ms3);

  BOOST_CHECK_EQUAL(zone.getAssignedSensors().size(), 3);
  BOOST_CHECK_EQUAL(zone.getAssignedSensorTypes(*dev1.getDevice()).size(), 2);
  BOOST_CHECK_EQUAL(zone.getUnassignedSensorTypes().size(), maxZoneSensors - 3);

  zone.removeDevice(dev1);
  zone.removeInvalidZoneSensors();
  BOOST_CHECK_EQUAL(zone.getAssignedSensors().size(), 1);
  BOOST_CHECK_EQUAL(zone.getAssignedSensorTypes(*dev1.getDevice()).size(), 0);
  BOOST_CHECK_EQUAL(zone.getUnassignedSensorTypes().size(), maxZoneSensors - 1);
}

BOOST_AUTO_TEST_CASE(testParseControlMode) {
  TemperatureConfig x;
  BOOST_CHECK_NO_THROW(
      x.parseControlMode("{\
        \"emergencyValue\": 50,\
        \"ctrlKp\": 5.2,\
        \"ctrlTs\": 240.1,\
        \"ctrlTi\": 1.2,\
        \"ctrlKd\": 1.3,\
        \"ctrlImin\": 600.0,\
        \"ctrlImax\": 2400.0,\
        \"ctrlYmin\": 0.5,\
        \"ctrlYmax\": 100.4,\
        \"ctrlAntiWindUp\": true\
    }"));

  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"emergencyValue\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlTs\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlTi\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlKd\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlYmin\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseControlMode("{\"ctrlYmax\": 50.0}"));
  BOOST_CHECK_THROW(x.parseControlMode("{\"ctrlAntiWindUp\": 1}"), std::runtime_error);
}

BOOST_AUTO_TEST_CASE(testParseFollowerMode) {
  TemperatureConfig x;
  BOOST_CHECK_NO_THROW(
      x.parseFollowerMode("{\
        \"referenceZone\": 38523,\
        \"ctrlOffset\": 10\
    }"));

  BOOST_CHECK_NO_THROW(x.parseFollowerMode("{\"referenceZone\": 50.0}"));
  BOOST_CHECK_NO_THROW(x.parseFollowerMode("{\"ctrlOffset\": 50.0}"));
}

BOOST_AUTO_TEST_CASE(testParseManualMode) {
  TemperatureConfig x;
  BOOST_CHECK_NO_THROW(
      x.parseManualMode("{\
        \"controlValue\": 10\
    }"));

  BOOST_CHECK_NO_THROW(x.parseManualMode("{\"controlValue\": 50.0}"));
}

BOOST_AUTO_TEST_SUITE_END()
