/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    Author: Grzegorz Krajewski

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#define BOOST_TEST_NO_MAIN
#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>

#include <digitalSTROM/dsuid.h>

#include "propertysystem.h"
#include "structuremanipulator.h"
#include "dss.h"
#include "exception"
#include "businterface.h"
#include "web/handler/zonerequesthandler.h"
#include "ds485types.h"
#include "model/zone.h"
#include "model/group.h"
#include "model/device.h"
#include "model/cluster.h"
#include "model/dsmeter.h"
#include "model/apartment.h"
#include "model/modelconst.h"
#include "model/devicereference.h"
#include "model/modelmaintenance.h"
#include "util/ds485-bus-mockups.h"
#include "tests/util/ds485-bus-mockups.h"
#include "tests/util/dss_instance_fixture.h"

using namespace dss;

BOOST_AUTO_TEST_SUITE(WebZone)

DSUID_DEFINE(dsuid1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1);
DSUID_DEFINE(dsuid2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2);
DSUID_DEFINE(dsuid3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3);
DSUID_DEFINE(dsmeterDSID, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10);
DSUID_DEFINE(meter1DSID, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10);

BOOST_FIXTURE_TEST_CASE(testZoneUnhandledFunctionFalse, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req("zone/unhandledFunction", "id=0");
  WebServerResponse response = handler.jsonHandleRequest(req, boost::shared_ptr<Session>(), NULL);
  testOkIs(response, false);
}

BOOST_FIXTURE_TEST_CASE(testZoneGetActionsNoMandatoryParamThrow, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req("zone/getActions", "id=0");
  BOOST_CHECK_THROW(handler.jsonHandleRequest(req, boost::shared_ptr<Session>(), NULL), std::runtime_error);
}

BOOST_FIXTURE_TEST_CASE(testZoneGetActionsUnknownApplicationThrow, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req("zone/getActions", "id=0&application=666");
  BOOST_CHECK_THROW(handler.jsonHandleRequest(req, boost::shared_ptr<Session>(), NULL), std::runtime_error);
}

BOOST_FIXTURE_TEST_CASE(testZoneGetActionsBlindsTrue, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req("zone/getActions", "id=0&application=2");
  WebServerResponse response = handler.jsonHandleRequest(req, boost::shared_ptr<Session>(), NULL);
  testOkIs(response, true);
}

BOOST_FIXTURE_TEST_CASE(testZoneCallActionClusterBlindsTrue, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());
  boost::shared_ptr<Cluster> cluster16 = apartment.getCluster(GroupIDAppUserMin);
  cluster16->setApplicationType(ApplicationType::Blinds);
  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(GroupIDAppUserMin);
  dev1->setName("blinds_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req("zone/callAction", "id=0&action=app.moveUp&cluster=" + std::to_string(GroupIDAppUserMin));
  WebServerResponse response = handler.jsonHandleRequest(req, boost::shared_ptr<Session>(), NULL);
  BOOST_CHECK_EQUAL(actionInterface.getLog(), "callScene(0," + std::to_string(GroupIDAppUserMin) + "," + std::to_string(Scene1) + ")");
}

BOOST_FIXTURE_TEST_CASE(testZoneCallActionNoMandatoryParamThrow, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req("zone/callAction", "id=0");
  BOOST_CHECK_THROW(handler.jsonHandleRequest(req, boost::shared_ptr<Session>(), NULL), ds::Exception);
}

BOOST_FIXTURE_TEST_CASE(testZoneCallActionBlindsInEveryRoomExceptZone3, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds1_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);
  boost::shared_ptr<Zone> zone1 = apartment.allocateZone(1);
  boost::shared_ptr<Device> dev2 = apartment.allocateDevice(dsuid2);
  dev2->setShortAddress(2);
  dev2->setDSMeter(meter1);
  dev2->setZoneID(1);
  dev2->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev2->setName("blinds2_GR_KL220");
  DeviceSpec spec2 = {};
  spec2.activeGroup = 2;
  spec2.FunctionID = 0x2131;
  spec2.ProductID = 0x0CDC;
  spec2.VendorID = 1;
  dev2->setPartiallyFromSpec(spec2);
  DeviceReference devRef2 = DeviceReference(dev2, &apartment);
  zone1->addDevice(devRef2);
  boost::shared_ptr<Zone> zone3 = apartment.allocateZone(3);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req1("zone/callAction", "id=0&action=app.moveUp&application=2");
  WebServerResponse response1 = handler.jsonHandleRequest(req1, boost::shared_ptr<Session>(), NULL);

  BOOST_CHECK_EQUAL(actionInterface.getLog(), "callScene(0,2," + std::to_string(Scene1) + ")");
}

BOOST_FIXTURE_TEST_CASE(testZoneCallActionNoAwningsInApartment, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  // create manually awnings group due to the usage of 'dss_instance_fixture'
  boost::shared_ptr<Group> group;
  group.reset(new Group(GroupIDGlobalAppDsAwnings, apartment.getZone(0)));
  group->setName("awnings");
  group->setApplicationType(ApplicationType::Awnings);
  group->setIsValid(true);
  apartment.getZone(0)->addGroup(group);
  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds1_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);
  boost::shared_ptr<Zone> zone1 = apartment.allocateZone(1);
  boost::shared_ptr<Device> dev2 = apartment.allocateDevice(dsuid2);
  dev2->setShortAddress(2);
  dev2->setDSMeter(meter1);
  dev2->setZoneID(1);
  dev2->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev2->setName("blinds2_GR_KL220");
  DeviceSpec spec2 = {};
  spec2.activeGroup = 2;
  spec2.FunctionID = 0x2131;
  spec2.ProductID = 0x0CDC;
  spec2.VendorID = 1;
  dev2->setPartiallyFromSpec(spec2);
  DeviceReference devRef2 = DeviceReference(dev2, &apartment);
  zone1->addDevice(devRef2);
  boost::shared_ptr<Zone> zone3 = apartment.allocateZone(3);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req1("zone/callAction", "id=0&action=app.moveUp&application=2");
  WebServerResponse response1 = handler.jsonHandleRequest(req1, boost::shared_ptr<Session>(), NULL);
  std::string excpectedResponse = "callScene(0,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req2("zone/callAction", "id=0&action=app.moveIn&application=65");
  WebServerResponse response2 = handler.jsonHandleRequest(req2, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(0,65," + std::to_string(SceneOff) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req3("zone/callAction", "id=1&action=app.moveUp&application=2");
  WebServerResponse response3 = handler.jsonHandleRequest(req3, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(1,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req5("zone/callAction", "id=2&action=app.moveUp&application=2");
  WebServerResponse response5 = handler.jsonHandleRequest(req5, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(2,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req7("zone/callAction", "id=3&action=app.moveUp&application=2");
  WebServerResponse response7 = handler.jsonHandleRequest(req7, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(3,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);
}

BOOST_FIXTURE_TEST_CASE(testZoneCallActionOneLegacyAwningInSeperateZone3, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  // create manually awnings group due to the usage of 'dss_instance_fixture'
  boost::shared_ptr<Group> group;
  group.reset(new Group(GroupIDGlobalAppDsAwnings, apartment.getZone(0)));
  group->setName("awnings");
  group->setApplicationType(ApplicationType::Awnings);
  group->setIsValid(true);
  apartment.getZone(0)->addGroup(group);
  boost::shared_ptr<Zone> zone3 = apartment.allocateZone(3);
  boost::shared_ptr<Device> devAwning = apartment.allocateDevice(dsuid3);
  devAwning->setShortAddress(3);
  devAwning->setDSMeter(meter1);
  devAwning->setZoneID(3);
  devAwning->addToGroup(static_cast<int>(ApplicationType::Blinds));
  devAwning->setName("legacy_awning_GR_KL210");
  DeviceSpec specAwning = {};
  specAwning.activeGroup = 2;
  specAwning.FunctionID = 0x2131;
  specAwning.ProductID = 0x0CD2;
  specAwning.VendorID = 1;
  devAwning->setPartiallyFromSpec(specAwning);
  DeviceReference devRefAwning = DeviceReference(devAwning, &apartment);
  zone3->addDevice(devRefAwning);

  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds1_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);
  boost::shared_ptr<Zone> zone1 = apartment.allocateZone(1);
  boost::shared_ptr<Device> dev2 = apartment.allocateDevice(dsuid2);
  dev2->setShortAddress(2);
  dev2->setDSMeter(meter1);
  dev2->setZoneID(1);
  dev2->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev2->setName("blinds2_GR_KL220");
  DeviceSpec spec2 = {};
  spec2.activeGroup = 2;
  spec2.FunctionID = 0x2131;
  spec2.ProductID = 0x0CDC;
  spec2.VendorID = 1;
  dev2->setPartiallyFromSpec(spec2);
  DeviceReference devRef2 = DeviceReference(dev2, &apartment);
  zone1->addDevice(devRef2);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req2("zone/callAction", "id=0&action=app.moveIn&application=65");
  WebServerResponse response2 = handler.jsonHandleRequest(req2, boost::shared_ptr<Session>(), NULL);
  std::string excpectedResponse =
      "callScene(0,65," + std::to_string(SceneOff) + ")" +
      "callScene(65534,2," + std::to_string(SceneOff) + ")" +
      "callScene(3,2," + std::to_string(SceneOff) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);
  actionInterface.clearLog();

  RestfulRequest req4("zone/callAction", "id=0&action=app.moveUp&application=2");
  WebServerResponse response4 = handler.jsonHandleRequest(req4, boost::shared_ptr<Session>(), NULL);
  excpectedResponse =
      "callScene(65534,2," + std::to_string(Scene1) + ")" +
      "callScene(2,2," + std::to_string(Scene1) + ")" +
      "callScene(1,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);
  actionInterface.clearLog();

  RestfulRequest req3("zone/callAction", "id=1&action=app.moveUp&application=2");
  WebServerResponse response3 = handler.jsonHandleRequest(req3, boost::shared_ptr<Session>(), NULL);
  excpectedResponse = "callScene(1,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);
  actionInterface.clearLog();

  RestfulRequest req5("zone/callAction", "id=2&action=app.moveUp&application=2");
  WebServerResponse response5 = handler.jsonHandleRequest(req5, boost::shared_ptr<Session>(), NULL);
  excpectedResponse = "callScene(2,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);
  actionInterface.clearLog();

  RestfulRequest req7("zone/callAction", "id=3&action=app.moveUp&application=2");
  WebServerResponse response7 = handler.jsonHandleRequest(req7, boost::shared_ptr<Session>(), NULL);
  excpectedResponse = "callScene(3,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);
  actionInterface.clearLog();
}

BOOST_FIXTURE_TEST_CASE(testZoneCallActionOneAwningInSeperateZone3, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  // create manually awnings group due to the usage of 'dss_instance_fixture'
  boost::shared_ptr<Group> group;
  group.reset(new Group(GroupIDGlobalAppDsAwnings, apartment.getZone(0)));
  group->setName("awnings");
  group->setApplicationType(ApplicationType::Awnings);
  group->setIsValid(true);
  apartment.getZone(0)->addGroup(group);
  boost::shared_ptr<Zone> zone3 = apartment.allocateZone(3);
  boost::shared_ptr<Device> devAwning = apartment.allocateDevice(dsuid3);
  devAwning->setShortAddress(3);
  devAwning->setDSMeter(meter1);
  devAwning->setZoneID(3);

  devAwning->addToGroup(static_cast<int>(ApplicationType::Blinds));
  devAwning->setName("awning_GR_KL210");
  DeviceSpec specAwning = {};
  specAwning.activeGroup = 65;
  specAwning.FunctionID = 0x2131;
  specAwning.ProductID = 0x0CD2;
  specAwning.VendorID = 1;
  devAwning->setPartiallyFromSpec(specAwning);
  DeviceReference devRefAwning = DeviceReference(devAwning, &apartment);
  zone3->addDevice(devRefAwning);

  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds1_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);
  boost::shared_ptr<Zone> zone1 = apartment.allocateZone(1);
  boost::shared_ptr<Device> dev2 = apartment.allocateDevice(dsuid2);
  dev2->setShortAddress(2);
  dev2->setDSMeter(meter1);
  dev2->setZoneID(1);
  dev2->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev2->setName("blinds2_GR_KL220");
  DeviceSpec spec2 = {};
  spec2.activeGroup = 2;
  spec2.FunctionID = 0x2131;
  spec2.ProductID = 0x0CDC;
  spec2.VendorID = 1;
  dev2->setPartiallyFromSpec(spec2);
  DeviceReference devRef2 = DeviceReference(dev2, &apartment);
  zone1->addDevice(devRef2);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req2("zone/callAction", "id=0&action=app.moveIn&application=65");
  WebServerResponse response2 = handler.jsonHandleRequest(req2, boost::shared_ptr<Session>(), NULL);
  std::string excpectedResponse = "callScene(0,65," + std::to_string(SceneOff) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req3("zone/callAction", "id=1&action=app.moveUp&application=2");
  WebServerResponse response3 = handler.jsonHandleRequest(req3, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(1,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req5("zone/callAction", "id=2&action=app.moveUp&application=2");
  WebServerResponse response5 = handler.jsonHandleRequest(req5, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(2,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req7("zone/callAction", "id=3&action=app.moveUp&application=2");
  WebServerResponse response7 = handler.jsonHandleRequest(req7, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(3,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);
}

BOOST_FIXTURE_TEST_CASE(testZoneCallActionAwningsAndBlindsMixedInZone2, DSSInstanceFixture) {
  Apartment &apartment(DSS::getInstance()->getApartment());

  boost::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(meter1DSID);
  // create manually awnings group due to the usage of 'dss_instance_fixture'
  boost::shared_ptr<Group> group;
  group.reset(new Group(GroupIDGlobalAppDsAwnings, apartment.getZone(0)));
  group->setName("awnings");
  group->setApplicationType(ApplicationType::Awnings);
  group->setIsValid(true);
  apartment.getZone(0)->addGroup(group);
  boost::shared_ptr<Zone> zone3 = apartment.allocateZone(3);
  boost::shared_ptr<Zone> zone2 = apartment.allocateZone(2);
  boost::shared_ptr<Device> devAwning = apartment.allocateDevice(dsuid3);
  devAwning->setShortAddress(3);
  devAwning->setDSMeter(meter1);
  devAwning->setZoneID(3);
  devAwning->addToGroup(static_cast<int>(ApplicationType::Blinds));
  devAwning->setName("awning_GR_KL210");
  DeviceSpec specAwning = {};
  specAwning.activeGroup = 65;
  specAwning.FunctionID = 0x2131;
  specAwning.ProductID = 0x0CD2;
  specAwning.VendorID = 1;
  devAwning->setPartiallyFromSpec(specAwning);
  DeviceReference devRefAwning = DeviceReference(devAwning, &apartment);
  zone3->addDevice(devRefAwning);

  boost::shared_ptr<Device> dev1 = apartment.allocateDevice(dsuid1);
  dev1->setShortAddress(1);
  dev1->setDSMeter(meter1);
  dev1->setZoneID(2);
  dev1->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev1->setName("blinds1_GR_KL220");
  DeviceSpec spec1 = {};
  spec1.activeGroup = 2;
  spec1.FunctionID = 0x2131;
  spec1.ProductID = 0x0CDC;
  spec1.VendorID = 1;
  dev1->setPartiallyFromSpec(spec1);
  DeviceReference devRef1 = DeviceReference(dev1, &apartment);
  zone2->addDevice(devRef1);
  boost::shared_ptr<Zone> zone1 = apartment.allocateZone(1);
  boost::shared_ptr<Device> dev2 = apartment.allocateDevice(dsuid2);
  dev2->setShortAddress(2);
  dev2->setDSMeter(meter1);
  dev2->setZoneID(1);
  dev2->addToGroup(static_cast<int>(ApplicationType::Blinds));
  dev2->setName("blinds2_GR_KL220");
  DeviceSpec spec2 = {};
  spec2.activeGroup = 2;
  spec2.FunctionID = 0x2131;
  spec2.ProductID = 0x0CDC;
  spec2.VendorID = 1;
  dev2->setPartiallyFromSpec(spec2);
  DeviceReference devRef2 = DeviceReference(dev2, &apartment);
  zone1->addDevice(devRef2);

  DummyStructureModifyingInterface modifyingInterface;
  DummyStructureQueryBusInterface queryInterface;
  DummyActionRequestInterface actionInterface;
  DummyBusInterface busInterface(&modifyingInterface, &queryInterface, &actionInterface);

  apartment.setBusInterface(&busInterface);
  ZoneRequestHandler handler(apartment, &modifyingInterface, &queryInterface);

  RestfulRequest req2("zone/callAction", "id=0&action=app.moveIn&application=65");
  WebServerResponse response2 = handler.jsonHandleRequest(req2, boost::shared_ptr<Session>(), NULL);
  std::string excpectedResponse = "callScene(0,65," + std::to_string(SceneOff) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req3("zone/callAction", "id=1&action=app.moveUp&application=2");
  WebServerResponse response3 = handler.jsonHandleRequest(req3, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(1,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req5("zone/callAction", "id=2&action=app.moveUp&application=2");
  WebServerResponse response5 = handler.jsonHandleRequest(req5, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(2,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);

  RestfulRequest req7("zone/callAction", "id=3&action=app.moveUp&application=2");
  WebServerResponse response7 = handler.jsonHandleRequest(req7, boost::shared_ptr<Session>(), NULL);
  excpectedResponse += "callScene(3,2," + std::to_string(Scene1) + ")";
  BOOST_CHECK_EQUAL(actionInterface.getLog(), excpectedResponse);
}

//TODO: Add 'lang' param test

BOOST_AUTO_TEST_SUITE_END()
